# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
# See LICENSE.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
This library is used to communicate with Micsoft Azure AD

Class:
    AzureAd        Azure Ad related operation
    CvAzureAd        Commvault Azure Ad operation class
"""
__all__ = []
import datetime
from time import sleep
import requests


from azure.graphrbac import GraphRbacManagementClient
from azure.graphrbac.models import UserCreateParameters, PasswordProfile,\
                                GroupCreateParameters, ApplicationCreateParameters,\
                                ServicePrincipalCreateParameters
from azure.common.credentials import UserPassCredentials
import adal
from msrestazure.azure_active_directory import AdalAuthentication
from msrestazure.azure_cloud import AZURE_PUBLIC_CLOUD
from .exceptions import ADException
from .adconstants import AZUREAD_SUPPORTED_TYPE_MAPPER, AZUREAD_TYPE_CREATE_PARANAME,\
                         AZUREAD_TYPE_CREATE_KEYNAME, AZUREAD_TYPE_CREATE_ENTRIES,\
                         AZUREAD_TYPE_ODATA_MAPPER, AZUREAD_INDEX_META_MAPPER,\
                         AZUREAD_DELETED_URL, AZUREAD_TYPE_ATTRIBUTES




class AzureAd():
    """
    Azure AD class to handle regular AD operations
    """

    def __init__(self, tenant, username, password, client_id, client_pass, log):
        """ initial class for Azure AD connection
        args:
            tenant      (string)    azure ad tenant id
            username    (string)    azure global admin user name
            password    (string)    azure admin password
            client_id    (string)    azure client to handle delete objects
            client_pass    (string)    azure clinet password
            log    (CV log instance)    log instance passed from test case
        """
        self.log = log
        self.graphclient = self.conn(tenant, username, password)
        self.domainname = self._get_domain_name()
        if self.domainname is None:
            raise ADException("azure", 100, "can't find the domain name")
        self.token = self.adal_token(client_id, client_pass, tenant)
        self.graph_api_endpoint = 'https://graph.microsoft.com/beta{0}'
        self.headers = {
            'Content-Type' : 'application/json',
            'Authorization': 'Bearer {0}'.format(self.token['accessToken']),
            'Accept' : 'application/json'}

    def conn(self, tenant, username, password):
        """ connect azure ad with user credential
        args:
            tenant      (string)    azure ad tenant id
            username    (string)    azure global admin user name
            password    (string)    azure admin password
        return:
            graphclient    (obj)    azure graph client
        """
        conn_credntials = UserPassCredentials(username,
                                              password,
                                              resource="https://graph.windows.net")
        graphclient = GraphRbacManagementClient(conn_credntials, tenant)
        self.log.info(f"MS graph client is setuped")
        return graphclient

    def adal_token(self, client_id, client_pass, tenant):
        """ connect to azure ad with client and cilent pass to generate token
        args:
            client_id    (string)    azure client to handle delete objects
            client_pass    (string)    azure clinet password
        return:
            token    (obj)    adal connection token for app
        """
        context = adal.AuthenticationContext(f'https://login.microsoftonline.com/{tenant}')
        token = context.acquire_token_with_client_credentials("https://graph.microsoft.com/",
                                                              client_id,
                                                              client_pass)
        self.log.debug(f"Azure connection token is setup")
        return token

    def type_operation(self, type_, operation, **kwargs):
        """ common type operation to avoid duplicate code
        args:
            type_    (string)    object type, select from the following type
                                    user, group, reg_app, ent_app
            operation(string)    supported operation from the following ops
                                    list, create, delete, find, update
            kwargs    (dict)    additional parameter to operation on azure ad obj
        return:
            operation_result    (objs)    list of azure ad objects
        """
        obj_type_name = AZUREAD_SUPPORTED_TYPE_MAPPER[type_]
        obj_type = getattr(self.graphclient, obj_type_name)

        operation_ins = None
        operation_result = None

        if operation == "list":
            operation_ins = getattr(obj_type, "list")()
            operation_result = []
            for obj_ in operation_ins:
                operation_result.append(obj_)
        elif operation == "create":
            operation_ins = getattr(obj_type, "create")
            operation_result = self._operation_create(obj_type,
                                                      operation_ins,
                                                      type_,
                                                      **kwargs)
        elif operation == "delete":
            self.log.debug(f"start to delete {type_} objects")
            operation_ins = getattr(obj_type, "delete")
            objs = self._find_objs(type_, **kwargs)
            self.log.debug(f"will delete {len(objs)} {type_} objects")
            for obj_ in objs:
                operation_ins(obj_.object_id)
                self.log.debug("{} object {} with name {} is deleted".\
                               format(type_, obj_.object_id, obj_.display_name))
            operation_result = self.type_operation(type_, "list")
        elif operation == "find":
            operation_result = self._find_objs(type_, **kwargs)
        elif operation == "update":
            if type_ in ['user', "ent_app"]:
                operation_ins = getattr(obj_type, "update")
            else:
                operation_ins = obj_type
            operation_result = self._operation_update(operation_ins, type_, **kwargs)
        return operation_result

    def _operation_create(self, obj_type, ins_, type_, **kwargs):
        """ create object
        args:
            obj_type (obj)    auzre object operation  obj create function
            ins_    (obj)    azure object opeeration obj
            type_    (str)    object type, select from the following type
                                    user, group, reg_app, ent_app
            kwargs    (dict)    additional parameter to create azure ad obj
        return:
            operation_result    (obj)    new object created or found obj
        """
        time_ = datetime.datetime.now()
        timestamp = int(time_.timestamp())
        new_define = {}
        for entry in AZUREAD_TYPE_CREATE_ENTRIES[type_]:
            if entry in kwargs:
                new_define[AZUREAD_TYPE_CREATE_ENTRIES[type_][entry]] = kwargs[entry]
            else:
                new_define[AZUREAD_TYPE_CREATE_ENTRIES[type_][entry]] = \
                f"Script_{type_}_{timestamp}"
        new_define = self._type_create_decoration(type_, new_define)
        type_define_ins = globals()[AZUREAD_TYPE_CREATE_PARANAME[type_]]
        type_define = type_define_ins(**new_define)
        try:
            type_obj = obj_type.get(new_define[AZUREAD_TYPE_CREATE_KEYNAME[type_]])
            self.log.debug(f"found  exist {type_} object name\
                {new_define[AZUREAD_TYPE_CREATE_KEYNAME[type_]]}")
        except:
            type_obj = ins_(type_define)
        operation_result = type_obj
        return operation_result

    def _operation_update(self, ins_, type_, **kwargs):
        """ udpate attribute in the object
        args:
             ins_    (obj)    azure object opeeration obj
            type_    (str)    object type, select from the following type
                                    user, group, reg_app, ent_app
            kwargs    (dict)    additional parameter to create azure ad obj
        return:
            operation_result    (obj)    updated objects
        """
        if "find" in kwargs:
            objs = self._find_objs(type_, **kwargs['find'])
        else:
            raise ADException("azuread", 401, kwargs)

        if type_ in ['group', 'reg_app']:
            self.log.info(f"{type_} do not support update attribute, will update owner")
            operation_result = objs
        else:
            if "update" in kwargs:
                updatekey = list(kwargs['update'].keys())
                # only support single attribute update
                if updatekey[0] in AZUREAD_TYPE_ATTRIBUTES[type_]:
                    if len(objs) == 1:
                        ins_(objs[0].object_id, kwargs['update'])
                    else:
                        for _ in objs:
                            operation_result = ins_(_.object_id, kwargs['update'])
                    operation_result = objs
                else:
                    raise ADException('azuread', 403, "type is {}, update is {}".\
                                      format(type_, kwargs['update']))
            else:
                raise ADException('azuread', 402, kwargs)

        return operation_result

    def user_ops(self, operation="list", **kwargs):
        """User related operations
        args:
            operation    (string)    supported operation from the following ops
                                    list, create, delete, find, update,
                                    default option is list operation
            kwargs    (dict)    additional parameter to operation on azure ad obj
        return:
            users    (objs)    azure user objects
        """
        users = self.type_operation("user", operation, **kwargs)
        return users

    def group_ops(self, operation="list", **kwargs):
        """group related operations
        args:
            operation(string)    supported operation from the following ops
                                    list, create, delete, find, update
                                    default option is list operation
            kwargs    (dict)    additional parameter to operation on azure ad obj
        return:
            groups    (objs)    azure groups objects
        """
        groups = self.type_operation("group", operation, **kwargs)
        return groups

    def reg_app_ops(self, operation="list", **kwargs):
        """registred application related operation
        args:
            operation(string)    supported operation from the following ops
                                    list, create, delete, find, update
                                    default option is list operation
            kwargs    (dict)    additional parameter to operation on azure ad obj
        return:
            reg_apps    (objs)    azure reg_app objects
        """
        reg_apps = self.type_operation("reg_app", operation, **kwargs)
        return reg_apps

    def ent_app_ops(self, operation="list", **kwargs):
        """ enterprise app related operations
        args:
            operation(string)    supported operation from the following ops
                                    list, create, delete, find, update
                                    default option is list operation
            kwargs    (dict)    additional parameter to operation on azure ad obj
        return:
            ent_apps    (objs)    azure ent_apps objects
        """
        ent_apps = self.type_operation("ent_app", operation, **kwargs)
        return ent_apps

    def _type_create_decoration(self, type_, new_define):
        """ process specail case for object creation maping
        args:
            type_    (string)    object type, select from the following type
                                    user, group, reg_app, ent_app
            new_define    (obj)    auzre ad creattion obj based on input
        return:
            new_define    (obj)    azure ad creation obj with correct value
        """
        if type_ == "user":
            new_define['user_principal_name'] = new_define['user_principal_name']\
                                                +f"@{self.domainname}"
            new_define['account_enabled'] = True
            new_define['password_profile'] = PasswordProfile(password='MyStr0ngP4ssword',
                                                             force_change_password_next_login=True)
        elif type_ == "reg_app":
            if "identifier_uris" not in new_define:
                new_define['identifier_uris'] = []
            elif new_define['identifier_uris'].startswith("Script"):
                new_define['identifier_uris'] = []
            new_define['available_to_other_tenants'] = True
        elif type_ == "ent_app":
            new_reg_app = self.reg_app_ops(operation="create",
                                           **{"displayname": new_define['displayname']})
            new_define['app_id'] = new_reg_app.app_id
            new_define["account_enabled"] = True
            # have to remove the displayname before create script principal object
            del(new_define['displayname'])
        return new_define

    def _find_objs(self, type_, **kwargs):
        """ find the objects based on give name or pattern
        args:
            type_    (string)    object type, select from the following type
                                    user, group, reg_app, ent_app
            kwargs    (dict)    additional parameter to operation on azure ad obj
        return:
            matchobjs    (objs)    founded objects in azure ad
        """
        matchobjs = []
        if "args" in kwargs:
            displaykey = None
            displaypattern = kwargs['args']
        else:
            displaykey = list(kwargs)[0]
            displaypattern = kwargs[displaykey]
        self.log.debug("find {} object with the display key {} and pattern {}".\
                       format(type_, displaykey, displaypattern))

        if isinstance(displaypattern, list) and len(displaypattern) == 1:
            operation = "match"
            displayname = displaypattern[0]
        else:
            operation = "find"
            displayname = displaypattern
        self.log.debug(f"will use method {operation} to get objects")

        for _ in self.type_operation(type_, "list"):
            if displaykey is None:
                attr_ins = _.display_name
            else:
                if type_ == "ent_app":
                    attr_ins = _.display_name
                else:
                    attr_ins = getattr(_, AZUREAD_TYPE_CREATE_ENTRIES[type_][displaykey])

                if operation == "match":
                    if attr_ins.startswith(displayname):
                        matchobjs.append(_)
                else:
                    if attr_ins.find(displayname) >= 0:
                        matchobjs.append(_)
        self.log.debug("found {} objects,here are the objects list: {}".\
                       format(len(matchobjs), [_.object_id for _ in matchobjs]))
        return matchobjs

    def _get_domain_name(self):
        """ get domain name from the tenant
        args:
        return:
            domainname    (string)    domain name in auzre ad
        """
        for _ in self.graphclient.domains.list():
            if _.is_default:
                domainname = _.name
                break
            else:
                domainname = None
        return domainname

    def group_objs_create(self, types=None, prestring="Script"):
        """ create group objects based on the types
        args:
            type_    (string)    object type, select from the following type
                                    user, group, reg_app, ent_app
            prestring (string)    the string put before the object name
        return:
            group_objs    (dict)    dict for each type. a list of type objects
        """
        group_objs = {}
        if types is None:
            types = list(AZUREAD_SUPPORTED_TYPE_MAPPER.keys())

        for type_ in types:
            ins_type_name = getattr(self, f"{type_}_ops")
            type_objects = ins_type_name()
            self.log.info(f"There are total {len(type_objects)} {type_} in Azure directory")
            newobj_name = f"{prestring}_{type_}"
            if type_ == "user":
                new_obj = ins_type_name("create",
                                        **{"username": newobj_name,
                                           "displayname": newobj_name})
            elif type_ == "group":
                new_obj = ins_type_name("create", **{"displayname": newobj_name})
            else:
                new_obj = ins_type_name("create", **{"displayname": newobj_name})
            self.log.debug(f"a new {type_} object created {new_obj}")
            group_objs[type_] = new_obj
        return group_objs

    def group_objs_check(self, types=None):
        """ return objects list
        args:
            type_    (string)    object type, select from the following type
                                    user, group, reg_app, ent_app
        return:
            group_objs    (dict)    dict for each type. a list of type objects
        """
        group_objs = {}
        if types is None:
            types = list(AZUREAD_SUPPORTED_TYPE_MAPPER.keys())

        for type_ in types:
            ins_type_name = getattr(self, f"{type_}_ops")
            type_objects = ins_type_name()
            group_objs[type_] = type_objects

        return group_objs

    def azuread_obj_mapper(self, obj_):
        """ covnert the auzre objects to a readable objects
        args:
            obj_     (obj)    auzre ad object
        return:
            obj_dict    (dict)    return dict of azure ad object
        """
        obj_dict = {}
        type_name = obj_.additional_properties['odata.type']
        type_ = AZUREAD_TYPE_ODATA_MAPPER[type_name]
        obj_dict['type'] = type_
        obj_dict['name'] = obj_.display_name
        obj_dict['id'] = obj_.object_id
        self.log.debug("will convert the Azure ad object :\n{} \nto readable obj:\n {}".\
                       format(obj_, obj_dict))
        obj_dict['adv_attribute'] = {}
        for _ in AZUREAD_TYPE_ATTRIBUTES[type_]:
            try:
                obj_dict['adv_attribute'][_] = getattr(obj_, _)
            except:
                self.log.debug(f"{type_} object {obj_.object_id} do not have {_} attribute")
        return obj_dict

    def get_delete_objs(self, type_):
        """ get delete items by type
        args:
            type_    (string)    object type, select from the following type
                                    user, group, reg_app, ent_app
        return:
            objs_     (obj)    azure ad objects
        """
        objs_ = []
        for _ in self.deleted_obj_list(type_):
            objs_.append(_)
        return objs_

    def deleted_obj_list(self, type_, objectid=None):
        """ get the delete objects list
        args:
            type_    (string)    object type, select from the following type
                                    user, group, reg_app, ent_app
            objectid    (string)    azure ad object id string
        return:
            objs     (obj)    azure ad objects
        """
        if objectid is not None:
            url = self.graph_api_endpoint.format(f"/directory/deletedItems/{objectid}")
        else:
            deleted_obj_mapper = {"user" : '/directory/deletedItems/microsoft.graph.user',
                                  "group" : "/directory/deletedItems/microsoft.graph.group",
                                  "reg_app" : \
                                  "/directory/deletedItems/microsoft.graph.application",
                                  "ent_app" : \
                                  "/directory/deletedItems/microsoft.graph.serviceprincipal"}
            url = self.graph_api_endpoint.format(deleted_obj_mapper[type_])
        result = requests.get(url, headers=self.headers)
        if result.status_code == 200:
            objs = self._graph_requests_result_process(result)
        elif result.status_code == 400:
            objs = None
        elif result.status_code == 404:
            objs = None
        else:
            raise ADException("azuread", 501, f" {url} reutrn {result.text}")
        return objs

    def _graph_requests_result_process(self, result):
        """ process requests result
        args:
            result    (obj)    request response result
        return:
            return_result    (list)    a list of result in json format
        """
        if "value" in result.json():
            return_result = result.json()['value']
        elif isinstance(result.json(), dict):
            return_result = [result.json()]
        self.log.debug(f"here is the request result {result}")
        return return_result

    def deleted_obj_delete(self, objectid=None):
        """ hard delete deleted object from recycle bin
        args:
             objectid    (string)    azure ad object id string
        """
        if objectid is None:
            raise ADException("azuread", 503)
        url = self.graph_api_endpoint.format(f"/directory/deletedItems/{objectid}")
        result = requests.delete(url, headers=self.headers)
        if result.status_code != 204:
            raise ADException("azuread", 502, url)

    def deleted_obj_clean(self, types):
        """ clean all deleted objects
        args:
            types    (list)    object type, select from the following type
                                    user, group, reg_app, ent_app
        """
        for type_ in types:
            del_objs = self.deleted_obj_list(type_)
            while len(del_objs) > 0:
                self.log.info(f"There are {len(del_objs)} found in {type_}")
                for _ in del_objs:
                    self.log.debug(f"{_['id']}  with name {_['displayName']} will be hard deleted")
                    self.deleted_obj_delete(objectid=_['id'])
                del_objs = self.deleted_obj_list(type_)

class CvAzureAd():
    """ azure ad class for CV software"""

    def __init__(self, aad_ins, sc_ins):
        """ initial class
        args:
            aad_ins     ins    AzureAd instance
            sc_ins        ins     subclient instnace from test case
        """
        self.log = aad_ins.log
        self.aad_ins = aad_ins
        self.subclient = sc_ins


    def cv_restore_option_builder(self, cv_objs, type_, overwrite=True):
        """ create cv restore option file
        args:
            cv_objs    list    azuread objects stored in cv backup
            types    (list)    object type, select from the following type
                                    user, group, reg_app, ent_app
            overwrite    Boolean    restore overwrite option
        return:
            restore_options    dict    restore option to create restore job
        """
        restore_items = []
        self.log.debug(f"This is the object will be restored {cv_objs}")
        for _ in cv_objs:
            item_ = {"type" : AZUREAD_INDEX_META_MAPPER[type_],
                     "isFolder" : False, "id": _['guid']}
            restore_items.append(item_)
        restore_options = {"fs_options" : {"overwrite" : overwrite},
                           "restore_option" : {"azureADOption" : {"restoreAllMatching": False,
                                                                  "restoreMembership" : True,
                                                                  "newUserDefaultPassword": "",
                                                                  "items": restore_items}}}
        return restore_options

    def cv_obj_delete_restore(self, sourceobjs, types=None, harddelete=False):
        """ process azure ad object, delete it first delete then run the restore job
        args:
            sourceobjs    list    auzre ad objects of source
            types        list    list of types from the following type
                                    user, group, reg_app, ent_app
            harddelete    Boolean    if the object type support hard delete or not
        return:
            obj_result    list    azure ad objects after rstore job
        """
        if types is None:
            types = list(AZUREAD_SUPPORTED_TYPE_MAPPER.keys())
        obj_result = {}

        for type_ in types:
            obj_ = self.aad_ins.azuread_obj_mapper(sourceobjs[type_])

            find_result = self.aad_ins.type_operation(type_, "find",
                                                      **{"displayname" : obj_['name']})
            if len(find_result) == 1:
                self.log.debug("found one matching obj with name {} and id is {}".\
                               format(obj_['name'], obj_['id']))
            # deleted objects
            self.aad_ins.type_operation(type_, "delete",
                                        **{"displayname": obj_['name']})
            # check if the object move to recycle bin,wait 10 seconds to check delete obj
            self.log.debug("Wait for 60 seconds to check delete item in recycle bine")
            sleep(60)
            result = self.aad_ins.deleted_obj_list(type_, obj_['id'])
            if result is not None and len(result) == 1:
                deleted_state = True
                self.log.debug("{} object {} found in recycle bin with value like {}".\
                               format(type_, obj_['id'], result[0]))
                if harddelete:
                    self.log.debug("hard delete is set, will permanent delete it")
                    self.aad_ins.deleted_obj_delete(obj_['id'])
                    self.log.debug(f"{type_} object {obj_['id']} is deleted")
            else:
                deleted_state = False
                self.log.debug("{} object {} NOT found. it is hard deleted already".\
                               format(type_, obj_['id']))

            find_result = self.aad_ins.type_operation(type_, "find",
                                                      **{"displayname" : obj_['name']})
            if len(find_result) == 0:
                self.log.debug(f"{type_} object {obj_['name']} is removed")

            job_ins = self.cv_obj_restore(obj_)
            self.log.debug("Wait for 30 seconds to check objects after restore")
            sleep(30)
            obj_match = self.cv_obj_compare(obj_)
            if deleted_state:
                obj_result_new = obj_match
                if harddelete:
                    obj_result_new = not obj_match
            else:
                obj_result_new = not obj_match
            obj_result[type_] = (job_ins, obj_result_new)
        return obj_result

    def cv_obj_change(self, sourceobjs, types=None, changevalue=None):
        """ process auzre object, change one attribute, then restor it
        args:
            sourceobjs    list    auzre ad objects of source
            types        list    list of types from the following type
                                    user, group, reg_app, ent_app
            changevlue    dict    type based change dict, will pass to change option
        return:
            changevalue    dict    type based change dict, use for compare
        """
        if types is None:
            types = list(AZUREAD_SUPPORTED_TYPE_MAPPER.keys())
        self.log.debug(f'Here is the {types} we used to update attribute')

        if changevalue is None:
            changevalue = {}
            for _ in types:
                self.log.debug(f"start to get default change value for {_}")
                change_name = AZUREAD_TYPE_ATTRIBUTES[_][1]
                change_value = f"{change_name}_changed_value"
                self.log.debug("here is the default change atttribute {} and value {}".\
                               format(change_name, change_value))
                changevalue[_] = {change_name : change_value}
        self.log.debug(f"Here is the change value {changevalue}")
        for type_ in types:
            for _ in changevalue[type_]:
                change_name = _
                change_value = changevalue[type_][change_name]
                self.log.debug("here is the attribute {} and old value {}".\
                               format(change_name, change_value))
            obj_ = sourceobjs[type_]
            self.log.debug(f'{type_} object with id {obj_.object_id}')
            updateinfo = {"find" : {"displayname": obj_.display_name},
                          "update" : changevalue[type_]}

            new_obj = self.aad_ins.type_operation(type_, "update", **updateinfo)
            if type_ in ['user', "ent_app"]:
                self.log.debug("here is the {} object {} attribute value {}".\
                               format(type_,
                                      new_obj[0].object_id,
                                      getattr(new_obj[0], change_name)))
            else:
                self.log.debug("check the owner of the object")
        return changevalue

    def cv_obj_change_browse_check(self, sourceobjs, changevalue, job_ins):
        """ process azure objects, check browse result and see if we get update value
        args:
            sourceobjs    list    auzre ad objects of source
            changevlue    dict    type based change dict, will pass to change option
            job_ins        ins    the latest backup job instance
        """
        types = list(sourceobjs.keys())
        browse_time = job_ins.summary['lastUpdateTime']
        jobinfo = {}
        jobinfo['totalobjs'] = job_ins.summary['totalNumOfFiles']
        jobinfo['totalsize'] = job_ins.summary['sizeOfApplication']
        self.log.info(f"here is the backup job info: {jobinfo}")
        self.log.debug(f"Backup job {job_ins.job_id} with browse time {browse_time}")
        self.log.debug(f"Here is the value we want to change {changevalue}")
        for type_ in types:
            obj_ = self.aad_ins.azuread_obj_mapper(sourceobjs[type_])
            browse_option = {
                "to_time" : browse_time,
                "folder" : type_,
                "show_deleted" : False,
                "search" : obj_['name'],
                "adv_attributes" : True}
            count, result = self.subclient.browse(browse_option)
            self.log.debug(f"browse found {count} objects. Here is the result {result}")
            adv_value = result[0]['adv_attribute']
            self.log.debug(f"{type_} object adv attributes are {adv_value}")

    def cv_obj_change_restore(self, sourceobjs, job_ins):
        """ run retore job from base and restore to old way
        args:
            sourceobjs    list    auzre ad objects of source
            job_ins        ins    the resetore job instance
        """
        types = list(sourceobjs.keys())
        browse_time = job_ins.summary['lastUpdateTime']
        jobinfo = {}
        jobinfo['totalobjs'] = job_ins.summary['totalNumOfFiles']
        jobinfo['totalsize'] = job_ins.summary['sizeOfApplication']
        self.log.info(f"here is the backup job info: {jobinfo}")
        self.log.debug(f"Backup job {job_ins.job_id} with browse time {browse_time}")
        for type_ in types:
            browse_option = {
                "to_time" : browse_time,
                "folder" : type_,
                "show_deleted" : False}
            count, result = self.subclient.browse(browse_option)
        self.log.debug(f"found {count} objects. Here is the result {result}")

    def cv_obj_restore(self, obj_, matchobj=True):
        """ run a restore job with more options
        args:
            obj_    obj    azure ad object
            matchobj     Boolean    decide if search by object id or name
        return:
            restore_job    ins    restore job instance
        """
        self.log.debug(f"start to restore {obj_}")
        count, aad_ids = self.subclient.browse(**{"folder" : obj_['type'],
                                                  "search" : obj_['name'],
                                                  "show_deleted": False})
        self.log.debug(f"restore {count} ojbects and the objcts list is {aad_ids}")
        if matchobj:
            match_obj = []
            for _ in aad_ids:
                if _['azureid'] == obj_['id']:
                    match_obj.append(_)
                    break
            aad_ids = match_obj
        restore_option = self.cv_restore_option_builder(aad_ids, obj_['type'])
        restore_job = self.subclient.restore_in_place(**restore_option)
        self.log.info(f"restore job {restore_job.job_id} started")
        jobresult = restore_job.wait_for_completion()
        if not jobresult:
            raise ADException("cvaad", 201,
                              f"restore job have some issue {restore_job.summary}.")
        self.log.info(f"restore job {restore_job.job_id} is completed with {jobresult}")
        return restore_job

    def cv_obj_compare(self, source_obj):
        """ find the matching objects and compare with source obj
        args:
            source_obj    list    azure ad object
        return:
            obj_match    Boolean    return compare result state
        """
        find_result = self.aad_ins.type_operation(source_obj['type'], "find",
                                                  **{"displayname" : source_obj['name']})
        obj_match = False
        if len(find_result) > 0:
            self.log.debug(f"There are {len(find_result)} result found")
            for _ in find_result:
                find_obj = self.aad_ins.azuread_obj_mapper(_)
                if find_obj['id'] == source_obj['id']:
                    self.log.debug("{} object name {} with id {} is restored".\
                                   format(source_obj['type'],
                                          find_obj['name'],
                                          find_obj['id']))
                    obj_match = True
                    break
                else:
                    self.log.debug("""
the restore id is not match , original id is {} and restored id is {}""".\
                                   format(source_obj['id'], find_obj['id']))
        return obj_match

    def cv_obj_browse_check(self, source_objs, job_ins):
        """ check backup job result and source obj
        args:
            source_obj    list    azure ad objects list of source data
            job_ins        ins    the latest backup job instance
        """
        types = list(source_objs.keys())

        browse_time = job_ins.summary['lastUpdateTime']
        jobinfo = {}
        jobinfo['totalobjs'] = job_ins.summary['totalNumOfFiles']
        jobinfo['totalsize'] = job_ins.summary['sizeOfApplication']
        self.log.info(f"here is the backup job info: {jobinfo}")
        self.log.debug(f"Backup job {job_ins.job_id} with browse time {browse_time}")
        for type_ in types:
            browse_option = {
                "to_time" : browse_time,
                "folder" : type_,
                "show_deleted" : False}
            count, result = self.subclient.browse(browse_option)
            if count == len(source_objs[type_]):
                self.log.debug(f" Browse return {count}, it match objects in azure AD {result}")
            else:
                raise ADException("cvaad", 301,
                                  "browse return {} and objects in ad is {}".\
                                  format(count, len(source_objs[type_])))
