# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright  Commvault Systems, Inc.
# See LICENSE.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""Main module for communicating with cvpysdk for all commvault related operations.

CvConnector class is defined in this file.

CvConnector: Performs Commvault related operations using cvpysdk

CvConnector:
    __init__()                                  --  initializes cv connector object

    client()                                    --  fetches client object

    instance()                                  --  fetches instance object

    backupset()                                 --  fetches bacupset object

    add_salesforce_client()                     --  adds salesforce client if it is not exist

    add_salesforce_instance()                   --  adds salesforce instance if it is not exist

    add_salesforce_backupset()                  --  adds salesforce backupset if it is not exist

    update_salesforce_backupset()               --  updates salesforce backupset properties

    update_salesforce_subclient()               --  updates salesforce subclient properties

    run_backup()                                --  runs backup with provided backup type

    run_restore_to_file_system()                --  runs restore to file system

    run_restore_to_database()                   --  runs restore to database

    run_restore_to_salesforce_from_database()   --  runs restore to salesforce from database

    run_restore_to_salesforce_from_media()      --  runs restore to salesforce from media

    check_job_status()                          --  checks for job status and waits for completed

"""

import time
from AutomationUtils import logger


class CvConnector(object):
    """Class for performing Commvault operations"""

    def __init__(self, commcell, tc_inputs):
        """Initializes the CvConnector object.

            Args:
                commcell   (object)  --  commcell object
                tc_inputs  (dict)  --  testcase inputs

        """
        self.tc_inputs = tc_inputs
        self.commcell = commcell
        self.log = logger.get_log()
        self.client_name = self.tc_inputs.get('client_name', 'Salesforce_Automation')
        self.instance_name = self.tc_inputs.get('instance_name','ORG1')
        self.backupset_name = self.tc_inputs['salesforce_options'].get('salesforce_user_name')
        self.subclient_name = self.tc_inputs.get("subclient_name", "default")
        self.access_node = self.tc_inputs.get("access_node", "")
        self.cvlt_options = {
            "storage_policy": self.tc_inputs.get('storage_policy',''),
            "download_cache_path": self.tc_inputs.get('download_cache_path', '/tmp'),
            "streams": self.tc_inputs.get('streams', 2)
        }
        self.salesforce_options = self.tc_inputs.get('salesforce_options', {})
        self.db_options = self.tc_inputs.get('db_options', {})
        self._client = None
        self._instance = None
        self._backupset = None
        self._subclient = None
        self.add_salesforce_client()
        self.agent = self.client.agents.get("Cloud Apps")
        self.add_salesforce_instance()
        self.add_salesforce_backupset()

    @property
    def backupset(self):
        """Returns backupset commcell object."""
        try:
            if not self._backupset:
                if not self.backupset_name or not self.instance.backupsets.has_backupset(self.backupset_name):
                    raise Exception("Backupset does not exist.")
                self._backupset = self.instance.backupsets.get(self.backupset_name)
            return self._backupset
        except Exception as excp:
            raise Exception('Exception {0} while getting backupset object'.format(str(excp)))

    @property
    def instance(self):
        """Returns instance commcell object."""
        try:
            if not self._instance:
                if not self.instance_name or not self.agent.instances.has_instance(self.instance_name):
                    raise Exception("instance does not exist.")
                self._instance = self.agent.instances.get(self.instance_name)
            return self._instance
        except Exception as excp:
            raise Exception('Exception {0} while getting instance object'.format(str(excp)))

    @property
    def subclient(self):
        """Returns subclient commcell object"""
        try:
            if not self._subclient:
                self._subclient = self.backupset.subclients.get(self.subclient_name)
            return self._subclient
        except Exception as excp:
            raise Exception('Exception {0} while getting subclient object'.format(str(excp)))

    @property
    def client(self):
        """Treats the client as a read-only attribute."""
        try:
            if not self._client:
                if not self.client_name or not self.commcell.clients.has_client(self.client_name):
                    raise Exception("Client does not exist.")
                self._client = self.commcell.clients.get(self.client_name)
            return self._client
        except Exception as excp:
            raise Exception("Exception {0} while getting client object".format(str(excp)))

    def add_salesforce_client(self):
        """
        Add Salesforce Client to commcell
        """

        try:

            if self.commcell.clients.has_client(self.client_name):
                self.log.info('Salesforce client %s already exists.', self.client_name)
                self._client = self.commcell.clients.get(self.client_name)
            else:
                self.log.info('Creating Salesforce client %s', self.client_name)
                self._client = self.commcell.clients.add_salesforce_client(self.client_name,
                                                                           self.access_node,
                                                                           self.salesforce_options, self.db_options,
                                                                           instance_name=self.instance_name,
                                                                           **self.cvlt_options)
                self.log.info('Created Salesforce client %s', self.client_name)

        except Exception as excp:
            self.log.exception("An error occurred while creating salesforce client")
            raise excp

    def add_salesforce_instance(self):
        """
        Add salesforce instance to commcell
        """
        try:
            if self.agent.instances.has_instance(self.instance_name):
                self.log.info('Salesforce instance %s already exists under Salesforce client %s.',
                              self.instance_name, self.client_name)
                self._instance = self.agent.instances.get(self.instance_name)
            else:
                self.log.info('Creating Salesforce instance %s under Salesforce client %s',
                              self.instance_name, self.client_name)
                self._instance = self.agent.instances.add_salesforce_instance(self.instance_name,
                                                                              self.access_node,
                                                                              self.salesforce_options,
                                                                              self.db_options, **self.cvlt_options)
                self.log.info('Created Salesforce instance %s under Salesforce client %s',
                              self.instance_name, self.client_name)
        except Exception as excp:
            self.log.exception("An error occurred while creating Salesforce instance")
            raise excp

    def add_salesforce_backupset(self):
        """
        Add salesforce backupset to commcell
        """
        try:

            if self.instance.backupsets.has_backupset(self.backupset_name):
                self.log.info('Salesforce backupset %s already exists under Salesforce client %s and instance %s.',
                              self.backupset_name, self.client_name, self.instance_name)
                self._backupset = self.instance.backupsets.get(self.backupset_name)
            else:
                self.log.info('Creating Salesforce backupset %s under salesforce instance %s and Salesforce client %s',
                              self.backupset_name, self.instance_name, self.client_name)
                self._backupset = self.instance.backupsets.add_salesforce_backupset(self.salesforce_options,
                                                                                    self.db_options,
                                                                                    **self.cvlt_options)
                self.log.info('Created Salesforce backupset %s under salesforce instance %s and Salesforce client %s',
                              self.backupset_name, self.instance_name, self.client_name)
        except Exception as excp:
            self.log.exception("An error occurred while creating Salesforce backupset")
            raise excp

    def update_salesforce_backupset(self, properties):
        """updates salesforce backupset
            Args:
                properties       (dict)   --  backupset properties to update
                Example:
                     properties =   {
                                        'mutual_auth_path' : mutual auth certificate path,
                                        'download_cache_path' : download cache path
                                    }
        """
        try:
            if 'mutual_auth_path' in properties:
                self.backupset.mutual_auth_path = properties['mutual_auth_path']
                self.log.info("Mutual auth certificate is enabled on backupset")

        except Exception as excp:
            self.log.exception("An error occurred while updating salesforce backupset")
            raise excp

    def update_salesforce_subclient(self, properties):
        """updates salesforce subclient
            Args:
                properties       (dict)   --  subclient properties to update
                Example:
                     properties =   {
                                        'files' : False,
                                        'metadata': True,
                                        'archived_deleted': True
                                    }

         """
        try:
            if 'files' in properties:
                if properties['files']:
                    self.subclient.enable_files()
                    self.log.info("Backup files enabled on subclient")
                else:
                    self.subclient.disable_files()
                    self.log.info("Backup files disabled on subclient")
            if 'metadata' in properties:
                if properties['metadata']:
                    self.subclient.enable_metadata()
                    self.log.info("Backup metadata enabled on subclient")
                else:
                    self.subclient.disable_metadata()
                    self.log.info("Backup metadata disabled on subclient")
            if 'archived_deleted' in properties:
                if properties['archived_deleted']:
                    self.subclient.enable_archived_deleted()
                    self.log.info("Backup archive deleted data enabled on subclient")
                else:
                    self.subclient.disable_archived_deleted()
                    self.log.info("Backup archive deleted data disabled on subclient")

        except Exception as excp:
            self.log.exception("An error occurred while updating salesforce subclient")
            raise excp

    def run_backup(self, backup_type):
        """Initiates backup job and waits for completion
            Args:
                backup_type         (str)   --  specify backup type like Full or Incremental
        """

        self.log.info("*" * 10 + " Starting Subclient {0} Backup ".format(backup_type) + "*" * 10)
        job = self.subclient.backup(backup_type)
        self.check_job_status(job)
        if backup_type == "Full":
            time.sleep(10)
            job = self.subclient.find_latest_job(include_finished=False)
            self.check_job_status(job)
        return job

    def run_restore_to_file_system(self, restore_options, salesforce_restore_options=None):
        """Initiates restore to file system job and waits for completion
            Args:
                restore_options               (dict)    --  provide restore options like destination info,
                                                                list of objects to restore
                                                                Example:
                                                                    restore_options = {
                                                                        'object_list': /Objects/Account,
                                                                        'dest_client_name': destination client name
                                                                    }
                salesforce_restore_options     (dict)    --  provide salesforce specific options
                                                                Example:
                                                                    salesforce_restore_options = {
                                                                        'destination_path': staging path location,
                                                                        'dependent_level': restore dependent children
                                                                        level
                                                                        'streams': number of streams
                                                                        'copy_precedence': copy from which restore needs
                                                                         to be done
                                                                        'from_time': start time for browse range
                                                                        'to_time': end time for browse range
                                                                    }
            Returns:
                job                          (Object)  -- job object
        """
        try:
            self.log.info("*" * 10 + " Starting {0} Restore to FileSystem *" * 10)
            job = self.subclient.restore_to_file_system(
                restore_options.get('object_list', '//'),
                restore_options.get('dest_client_name', None),
                salesforce_restore_options)
            self.check_job_status(job)
            return job
        except Exception as excp:
            self.log.exception("An error occurred while running restore to file system")
            raise excp

    def run_restore_to_database(self, restore_options, salesforce_restore_options=None):
        """Initiates restore to database job and waits for completion
            Args:
                restore_options               (dict)    --  provide restore options like destination info,
                                                                list of objects to restore
                                                                Example:
                                                                    restore_options = {
                                                                        'object_list': /Objects/Account,
                                                                        'dest_client_name': destination client name
                                                                    }

                salesforce_restore_options    (dict)   --  provide salesforce specific options
                                                                Example:
                                                                    salesforce_restore_options = {
                                                                        'destination_path': staging path location,
                                                                        'db_type': SQLSERVER or POSTGRESQL,
                                                                        'db_host_name': database hostname,
                                                                        'db_instance': database instance name,
                                                                        'db_name': database name,
                                                                        'db_port': port of the database,
                                                                        'db_user_name': database user name,
                                                                        'db_user_password': database user password,
                                                                        'dependent_level': restore dependent children
                                                                         level,
                                                                        'streams': number of streams,
                                                                        'copy_precedence': copy from which restore needs
                                                                         to be done,
                                                                        'from_time': start time for browse range,
                                                                        'to_time': end time for browse range
                                                                    }
            Returns:
                job                           (Object)  -- job object
        """
        try:
            self.log.info("*" * 10 + " Starting {0} Restore to Database *" * 10)
            job = self.subclient.restore_to_database(
                restore_options.get('object_list', '/Objects/'),
                restore_options.get('dest_client_name', None),
                salesforce_restore_options)
            self.check_job_status(job)
            return job
        except Exception as excp:
            self.log.exception("An error occurred while running restore to database")
            raise excp

    def run_restore_to_salesforce_from_database(self, restore_options, salesforce_restore_options=None):
        """Initiates restore to salesforce from database job and waits for completion
            Args:
                restore_options               (dict)    --  provide restore options like destination info,
                                                             list of objects to restore
                                                              Example:
                                                                restore_options = {
                                                                    'object_list': /Objects/Account,
                                                                    'dest_client_name': destination client name,
                                                                    'dest_instance_name': destination instance name,
                                                                    'dest_backupset_name': destination backupset name
                                                                }

                salesforce_restore_options    (dict)    --  provide salesforce specific options
                                                                Example:
                                                                    salesforce_restore_options = {
                                                                        'destination_path': staging path location,
                                                                        'db_type': SQLSERVER or POSTGRESQL,
                                                                        'db_host_name': database hostname,
                                                                        'db_instance': database instance name,
                                                                        'db_name': database name,
                                                                        'db_port': port of the database,
                                                                        'db_user_name': database user name,
                                                                        'db_user_password': database user password,
                                                                        'dependent_level': restore dependent children
                                                                         level,
                                                                        'streams': number of streams,
                                                                        'copy_precedence': copy from which restore needs
                                                                         to be done,
                                                                        'from_time': start time for browse range,
                                                                        'to_time': end time for browse range
                                                                    }
            Returns:
                job                         (Object)  -- job object
        """
        try:
            self.log.info("*" * 10 + " Starting restore to salesforce from database " + "*" * 10)
            job = self.subclient.restore_to_salesforce_from_database(
                restore_options.get('object_list', '/Objects/Contact'),
                restore_options.get('dest_client_name', None),
                restore_options.get('dest_instance_name', None),
                restore_options.get('dest_backupset_name', None),
                salesforce_restore_options)
            self.check_job_status(job)
            return job
        except Exception as excp:
            self.log.exception("An error occurred while running restore to salesforce from database")
            raise excp

    def run_restore_to_salesforce_from_media(self, restore_options, salesforce_restore_options=None):
        """Initiates restore to salesforce from media job and waits for completion
            Args:
                restore_options                (dict)    --  provide restore options like destination info,
                                                             list of objects to restore
                                                                Example:
                                                                    restore_options = {
                                                                     'object_list': /Objects/Account,
                                                                     'dest_client_name': destination client name,
                                                                     'dest_instance_name': destination instance name,
                                                                     'dest_backupset_name': destination backupset name
                                                                    }

                salesforce_restore_options    (dict)    --  provide salesforce specific options
                                                            Example:
                                                                 salesforce_restore_options = {
                                                                    'destination_path': staging path location,
                                                                    'db_type': SQLSERVER or POSTGRESQL,
                                                                    'db_host_name': database hostname,
                                                                    'db_instance': database instance name,
                                                                    'db_name': database name,
                                                                    'db_port': port of the database,
                                                                    'db_user_name': database user name,
                                                                    'db_user_password': database user password,
                                                                    'dependent_level': restore dependent children
                                                                    level,
                                                                    'streams': number of streams,
                                                                    'copy_precedence': copy from which restore needs
                                                                     to be done,
                                                                    'from_time': start time for browse range,
                                                                    'to_time': end time for browse range
                                                                }
            Returns:
                job                         (Object)    -- job object
        """
        try:
            self.log.info("*" * 10 + " Starting {0} Restore to Salesforce from Media *" * 10)
            job = self.subclient.restore_to_salesforce_from_media(
                restore_options.get('object_list', '/Objects/Contact'),
                restore_options.get('dest_client_name', None),
                restore_options.get('dest_instance_name', None),
                restore_options.get('dest_backupset_name', None),
                salesforce_restore_options)
            self.check_job_status(job)
            return job
        except Exception as excp:
            self.log.exception("An error occurred while running restore to salesforce from media")
            raise excp

    def check_job_status(self, job):
        """Checks the status of job until it is finished and
            raises exception on pending, failure etc.
            Args:
                job         (object)    --  job object
        """
        self.log.info("{1} job started with id: {0}".format(str(job.job_id), job.job_type))
        if not job.wait_for_completion():
            raise Exception(
                "Failed to run job with error: {0}".format(
                    job.delay_reason
                )
            )
        self.log.info('{0} job {1} completed successfully'.format(job.job_type, str(job.job_id)))
