# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright  Commvault Systems, Inc.
# See LICENSE.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""Helper file for performing salesforce operations

SalesforceConnector: Connector class to perform salesforce operations

SalesforceConnector:
    __init__()                      --  initializes salesforce connector object

    sf_login()                      --  connects to salesforce

    sf_mutual_auth_login()          --  connects to salesforce using mutual auth certificate user

    sf_relogin()                    --  reconnects to salesforce if session is invalid

    sf_list_object()                --  lists all the available objects on salesforce

    sf_bulk_cmd()                   --  prepares bulk command

    sf_bulk_query()                 --  query records using bulk

    sf_soap_query()                 --  query records using soap

    sf_mutual_auth_soap_query()     --  query records using soap with mutual auth user

    sf_bulk_update()                --  update records using bulk

    sf_bulk_insert()                --  insert records using bulk

    sf_bulk_delete()                --  delete records using bulk

    generate_data_dictionary()      --  generate data for update/insert/delete

"""

from datetime import datetime
import uuid

from AutomationUtils import logger
from simple_salesforce import Salesforce


class SalesforceConnector(object):
    """
    Class to connect to Salesforce
    """

    def __init__(self, tc_inputs):
        """
        Constructor for the class
        Args:
            tc_inputs       (dict)     --  testcase inputs

        """
        self.tc_inputs = tc_inputs

        if 'mutual_auth_path' in self.tc_inputs:
            self.sf_user = self.tc_inputs['salesforce_options'].get('salesforce_non_massl_user_name')
            self.sf_password = self.tc_inputs['salesforce_options'].get('salesforce_non_massl_user_password')
            self.sf_token = self.tc_inputs['salesforce_options'].get('salesforce_non_massl_user_token')
            self.sf_mutual_auth_user = self.tc_inputs['salesforce_options'].get('salesforce_user_name')
            self.sf_mutual_auth_password = self.tc_inputs['salesforce_options'].get('salesforce_user_password')
            self.sf_mutual_auth_token = self.tc_inputs['salesforce_options'].get('salesforce_user_token')
            self.sandbox = self.tc_inputs['salesforce_options'].get('sandbox', True)
        else:
            self.sf_user = self.tc_inputs['salesforce_options'].get('salesforce_user_name')
            self.sf_password = self.tc_inputs['salesforce_options'].get('salesforce_user_password')
            self.sf_token = self.tc_inputs['salesforce_options'].get('salesforce_user_token')
        self.sandbox = self.tc_inputs['salesforce_options'].get('sandbox', True)
        self._sfconn = None
        self._sfconn_mutual = None
        self._bulk_type = "sf.bulk"
        self._bulk_cmd = None
        self.sf_objects = []
        self.log = logger.get_log()
        self.sf_login()
        self.sf_list_objects()

    def sf_login(self):
        """
        Connects to salesforce
        """
        try:
            self.log.info('sfuser: ' + self.sf_user)
            self.log.info('sfpassword: ********** ')
            self.log.info('sftoken: **********')
            self.log.info('sandbox: ' + str(self.sandbox))

            if self.sandbox:
                self._sfconn = Salesforce(
                    username=self.sf_user,
                    password=self.sf_password,
                    security_token=self.sf_token,
                    sandbox=self.sandbox)
            else:
                self._sfconn = Salesforce(
                    username=self.sf_user,
                    password=self.sf_password,
                    security_token=self.sf_token)

        except Exception as exp:
            raise Exception("Failed to login to salesforce: {0} ".format(str(exp)))

    def sf_mutual_auth_login(self):
        """
        Connects to salesforce using mutual auth certificate user
        """
        try:
            self.log.info('sf mutual auth user: ' + self.sf_mutual_auth_user)
            self.log.info('sf mutual auth user password: ********** ')
            self.log.info('sf mutual auth user token: **********')
            self.log.info('sandbox: ' + str(self.sandbox))

            if self.sandbox:
                self._sfconn_mutual = Salesforce(
                    username=self.sf_mutual_auth_user,
                    password=self.sf_mutual_auth_password,
                    security_token=self.sf_mutual_auth_token,
                    sandbox=self.sandbox)
            else:
                self._sfconn_mutual = Salesforce(
                    username=self.sf_mutual_auth_user,
                    password=self.sf_mutual_auth_password,
                    security_token=self.sf_mutual_auth_token)

        except Exception as exp:
            raise Exception("Failed to login to salesforce using mutual auth user: {0} ".format(str(exp)))

    def sf_relogin(self):
        """
        Relogin to salesforce if connection is invalid
        """
        try:
            self.sf_soap_query("select id from Organization")
        except Exception as exp:
            self.log.info("Exception {0} while querying records. Re-logging..".format(str(exp)))
            self.sf_login()

    def sf_list_objects(self):
        """
        Lists all the available objects in salesforce
        """
        try:
            for x in self._sfconn.describe()["sobjects"]:
                self.sf_objects.append(x["name"])
        except Exception as exp:
            raise Exception("Failed to get list of objects from salesforce: {0} ".format(str(exp)))

    def sf_bulk_cmd(self, sf_object):
        """
        Prepares bulk command
        Args:
            sf_object       (str)       --  Salesforce object name
        """
        try:
            self._bulk_cmd = self._sfconn.bulk.Automation
            self._bulk_cmd.object_name = sf_object
        except Exception as exp:
            raise Exception("sf_bulk_cmd error: {0} ".format(str(exp)))

    def sf_bulk_query(self, sf_object, query):
        """
        Runs bulk query
        Args:
            sf_object       (str)       --  Salesforce object name
            query           (str)       --  soql query to run
        Return:
            result          (dict)      --  query response
        """
        try:
            self.sf_bulk_cmd(sf_object)
            result = self._bulk_cmd.query(query)
            return result
        except Exception as exp:
            raise Exception("sf_bulk_query error: {0} ".format(str(exp)))

    def sf_soap_query(self, query):
        """
        Runs soap query
        Args:
            query           (str)       --  soql query to run
        Return:
            result          (dict)      --  query response
        """
        try:
            result = self._sfconn.query(query)
            return result
        except Exception as exp:
            raise Exception("sf_soap_query: " + str(exp))

    def sf_mutual_auth_soap_query(self, query):
        """
        Runs soap query using mutual auth cert user
        Args:
            query           (str)       --  soql query to run
        Return:
            result          (dict)      --  query response

        """
        try:
            result = self._sfconn_mutual.query(query)
            return result
        except Exception as exp:
            self.log.info("Exception message is {0}".format(str(exp)))
            return str(exp)

    def sf_bulk_insert(self, sf_object, fname="ABD", rec_count=1000):
        """
        Runs bulk insert operation
        Args:
            sf_object       (str)       --  salesforce object name
            fname           (str)       --  column value
            rec_count       (int)       --  number of records to insert
        """
        try:
            data = self.generate_data_dictionary(sf_object, fname, rec_count)
            self.sf_bulk_cmd(sf_object)
            result = self._bulk_cmd.insert(data)
            self.log.info("Bulk insert job succeeded ")

        except Exception as exp:
            raise Exception("Bulk insert failed: {0} ".format(str(exp)))

    def sf_bulk_delete(self, sf_object, where_condition, hard_delete=False):
        """
        Runs bulk delete operation
        Args:
            sf_object           (str)       --  salesforce object name
            where_condition     (str)       --  filter the records for deletion based on condition
            hard_delete         (bool)         --  hard delete the records even from recyle bin
        """
        try:
            query = "select id from " + sf_object + " where " + where_condition
            result = self.sf_bulk_query(sf_object, query)
            if len(result):
                delete_data = [
                    {
                        'Id': record['Id']
                    } for record in result
                ]
            else:
                self.log.info("no matched records")
                return False
            self.sf_bulk_cmd(sf_object)
            if hard_delete:
                result = self._bulk_cmd.hard_delete(delete_data)
            else:
                result = self._bulk_cmd.delete(delete_data)
            self.log.info("Bulk delete job succeeded ")

        except Exception as exp:
            raise Exception("sf_bulk_delete error:{0} ".format(str(exp)))

    def sf_bulk_update(self, sf_object, field_name, where_condition):
        """
        Runs bulk update operation
        Args:
            sf_object           (str)       --  salesforce object name
            field_name          (str)       --  column name to update
            where_condition     (str)       --  filter the records for updation based on condition
        """
        try:
            query = "select id from " + sf_object + " where " + where_condition
            result = self.sf_bulk_query(sf_object, query)
            update_data = []
            if len(result):
                if sf_object == "Contact":
                    update_data = [
                        {
                            'Id': record['Id'],
                            '%s' % field_name: 'automation%s_%s@cvlt.com' % (
                                str(datetime.today().strftime('%Y-%m-%d')), str(uuid.uuid4().hex[:7])),
                        } for record in result
                    ]
            else:
                self.log.info("no matched records")
                return False
            self.sf_bulk_cmd(sf_object)
            result = self._bulk_cmd.update(update_data)
            self.log.info("Update succeeded ")
        except Exception as exp:
            raise Exception("sf_bulk_update error: {0} ".format(str(exp)))

    def generate_data_dictionary(self, sf_object, fname="ABD", rec_count=1000):
        """
        Prepares data dictionary for insert/update/delete operations
        Args:
            sf_object           (str)       --  salesforce object name
            fname          (str)       --  column name to update
            rec_count           (int)       --  number of records to generate
        """
        try:
            data = []
            if sf_object == "Contact":
                data = [
                    {
                        'FirstName': '%s_FN%s_%s' % (fname, i, datetime.now().replace(microsecond=0)),
                        'LastName': '%s_LN%s_%s' % (fname, i, datetime.now().replace(microsecond=0)),
                        'Phone': '555-555-5555',
                    } for i in range(rec_count)
                ]
            if sf_object == "grandparent1__c":
                data = [
                    {
                        'Name': '%s%s_%s' % (fname, i, datetime.now().replace(microsecond=0)),
                    } for i in range(rec_count)
                ]

            return data
        except Exception as exp:
            raise Exception("generate data error: {0} ".format(str(exp)))
