# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright  Commvault Systems, Inc.
# See LICENSE.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""Helper file for integrating both salesforce and commvault operations

SalesforceHelper is the only class defined in this file

SalesforceHelper: Helper class to integrate salesforce and commvault operations

SalesforceHelper:
    __init__()                          --  initializes salesforce object

    create_test_data()                  --  creates test data in salesforce

    create_test_data_for_incremental()  --  creates incremental test data in salesforce

    update_test_data()                  --  updates test data in salesforce

    delete_test_data()                  --  deletes test data in salesforce

    query_test_data()                   --  queries test data in salesforce

    validate_mutual_auth()              --  validate whether the given use is mutual auth user or not

    validate_test_data_in_cloud()       --  verifies testdata in salesforce after restore

"""
from .cv_connector import CvConnector
from .salesforce_connector import SalesforceConnector
from AutomationUtils import logger
from time import sleep


class SalesforceHelper(object):
    """
    Class for integrating salesforce and commvault operations
    """

    def __init__(self, commcell, tc_inputs):
        """Constructor for the class
        Args:
            commcell        (object)     -- commcell object
            tc_inputs       (dict)       --  testcase inputs

        """
        self.tc_inputs = tc_inputs
        self.log = logger.get_log()
        self._batch_count = 2
        self._object_name = "Account"
        self._column_data = "ABD"
        self.cvconnector = CvConnector(commcell, self.tc_inputs)
        self.salesforceconnector = SalesforceConnector(self.tc_inputs)

    @property
    def object_name(self):
        """getter for object_name"""
        return self._object_name

    @object_name.setter
    def object_name(self, value):
        """
        setter for object_name
        Args:
            value   (str)  --  object name
        """
        self._object_name = value

    @property
    def column_data(self):
        """getter for object_name"""
        return self._column_data

    @column_data.setter
    def column_data(self, value):
        """
        setter for object_name
        Args:
            value  (str)  --  column value
        """
        self._column_data = value

    @property
    def batch_count(self):
        """getter for object_name"""
        return self._batch_count

    @batch_count.setter
    def batch_count(self, value):
        """
        setter for object_name
        Args:
            value   (int)  --  batch count
        """
        self._batch_count = value

    def create_test_data(self, rec_count=5000):
        """
        inserts test data in salesforce based on record count and batch count
        Args:
             rec_count      (int)
        """
        try:
            self.salesforceconnector.sf_relogin()
            query = "select count() from {0}".format(self.object_name)
            before_insert = self.salesforceconnector.sf_soap_query(query)
            for _ in range(1, self.batch_count):
                self.salesforceconnector.sf_bulk_insert(self.object_name, self.column_data, rec_count)
            after_insert = self.salesforceconnector.sf_soap_query(query)
            if before_insert['totalSize'] == after_insert['totalSize']:
                raise Exception("Bulk insert job succeeded but no new records got added.")
        except Exception as exp:
            raise Exception('create_test_data: {0}'.format(str(exp)))

    def create_test_data_for_incremental(self, column_name, where_condition, rec_count=1000):
        """
        inserts/updates/deletes test data in salesforce based on record count and filter condition
        Args:
            column_name         (str)       -- column name to update
            where_condition     (str)       -- filter query to delete/update the records
            rec_count           (int)       -- record count
        """
        try:
            self.salesforceconnector.sf_relogin()
            for _ in range(1, self.batch_count):
                self.salesforceconnector.sf_bulk_insert(self.object_name, "INC_" + self.column_data, rec_count)
                self.salesforceconnector.sf_bulk_delete(self.object_name, where_condition, False)
                self.salesforceconnector.sf_bulk_update(self.object_name, column_name, where_condition)
            sleep(60)
        except Exception as exp:
            raise Exception('create_test_data_for_incremental: {0}'.format(str(exp)))

    def update_test_data(self, column_name, where_condition):
        """
        updates test data in salesforce based on column name and filter condition
        Args:
            column_name         (str)       -- column name to update
            where_condition     (str)       -- filter query to update the records
        """
        try:
            self.salesforceconnector.sf_relogin()
            for _ in range(1, self.batch_count):
                self.salesforceconnector.sf_bulk_update(self.object_name, column_name, where_condition)
        except Exception as exp:
            raise Exception('update_test_data: {0}'.format(str(exp)))

    def delete_test_data(self, where_condition, hard_delete=False):
        """
        deletes test data in salesforce based filter condition
        Args:
            where_condition     (str)       -- filter query to update the records
            hard_delete         (bool)      -- True/False to hard delete data from recyclebin
        """
        try:
            self.salesforceconnector.sf_relogin()
            query = "select count() from {0}".format(self.object_name)
            before_delete = self.salesforceconnector.sf_soap_query(query)
            for _ in range(1, self.batch_count):
                self.salesforceconnector.sf_bulk_delete(self.object_name, where_condition, hard_delete)
            after_delete = self.salesforceconnector.sf_soap_query(query)
            if before_delete['totalSize'] == after_delete['totalSize']:
                raise Exception("Bulk delete job succeeded but no records got deleted.")
        except Exception as exp:
            raise Exception('delete_test_data: {0}'.format(str(exp)))

    def query_test_data(self, query):
        """
        queries test data from salesforce
        Args:
            query       (str)       -- SOQL query to fetch the records from an object
        """
        try:
            self.salesforceconnector.sf_relogin()
            query_data = self.salesforceconnector.sf_bulk_query(self.object_name, query)
            for record in query_data:
                del record['attributes']
            return query_data
        except Exception as exp:
            raise Exception('query_test_data: {0}'.format(str(exp)))

    def validate_mutual_auth_user(self):
        """
        validates the given user for whether mutual auth certificate enabled or not
        """
        try:
            self.salesforceconnector.sf_mutual_auth_login()
            query = "select count() from Organization"
            response = self.salesforceconnector.sf_mutual_auth_soap_query(query)
            if response.find("MUTUAL_AUTHENTICATION_FAILED") >= 0:
                self.log.info("Provided salesforce user is a mutual auth user.")
            else:
                raise Exception("Provided salesforce user is not a mutual auth user")
        except Exception as exp:
            raise Exception('validate_mutual_auth_user: {0}'.format(str(exp)))

    def validate_test_data_in_cloud(self, query, backup_data):
        """
        validates the restored data with backup data
        Args:
            query           (str)       --  SOQL query to fetch restored records
            backup_data     (dict)      --  backup records dictionary

        """
        try:
            restore_data = self.query_test_data(query)
            if len(backup_data) == len(restore_data):
                self.log.info("Record count before backup is: {0}. Record count after restore is: {1} "
                              .format(len(backup_data), len(restore_data)))
                for record in backup_data:
                    if record in restore_data:
                        continue
                    else:
                        self.log.error("Validation failed. Data before backup is: {0}. "
                                       "Data after restore is: {1}".format(backup_data, restore_data))
                        raise Exception("Restore validation to cloud failed. Please check logs")
                self.log.info("Validation succeeded. Data before backup is: {0}. Data after restore is: {1} "
                              .format(backup_data, restore_data))
            else:
                self.log.error("Validation failed. Record count before backup is: {0}. Record count after "
                               "restore is: {1}".format(len(backup_data), len(restore_data)))
                raise Exception("Restore validation to cloud failed. Please check logs")

        except Exception as exp:
            raise Exception('validate_test_data_in_cloud: {0}'.format(str(exp)))
