# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
# See LICENSE.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""Helper file for performing operations on Azure Cosmos Database (SQL API)
using Azure SDK for python
CosmosSQLAPI is the only class defined in this file
===========================================================================

CosmosSQLAPI:

__init__            --  Initializes the CosmosSQLAPI instance using URI and key

create_database()   --  Creates CosmosDB database

get_db_client()     --  Returns the object of database class to perform operations

                        on CosmosDB database

delete_database()   --  Deletes the given database

create_container()  --  Creates CosmosDB container under given database

get_cnt_client()    --  Returns the object of container class to invoke operations

                        on CosmosDB container

populate_container() -- Populates the given container with test data

delete_container()   -- Deletes the container under given database

validate_container() -- Validates the items in the container after restore
"""

import time
from azure.cosmos import cosmos_client
from azure.cosmos import PartitionKey
from AutomationUtils import logger


class CosmosSQLAPI:
    """Helper class to interact with Azure CosmosDB SQL API
        using Azure SDK for python"""
    def __init__(self, account_uri, account_key):
        """Initializes the CosmosDB account URL, key and Instantiates Cosmos client class
        Args:
        account_uri:    (str): The URI of CosmosDB account to connect to

        account_key:    (str): The primary master key of the CosmosDB account
        """
        self.account_uri = account_uri
        self.account_key = account_key
        self.client = cosmos_client.CosmosClient(self.account_uri, self.account_key)
        self.log = logger.get_log()

    def create_database(self, database_name, throughput=0):
        """Creates CosmosDB database
        Args:
            database_name:  (str)   --  Name of database

            throughput:     (int)   --  Throughput to be assigned
                default- 0

        Raises:
            Exception:  If database could not be created due to CosmosDB issue
                        Account URI or key is not valid
        """
        try:
            databases = self.client.list_databases()
            for database in databases:
                if database_name in database.get('id'):
                    self.log.info('Database with name %s exists, deleting and recreating it',
                                  database_name)
                    self.delete_database(database_name)
            if throughput == 0:
                self.client.create_database(id=database_name)
            else:
                self.client.create_database(id=database_name, offer_throughput=throughput)
            self.log.info('Database created successfully')

        except Exception as exp:
            self.log.exception('Creating database failed with error: %s', exp)

    def get_db_client(self, database_name):
        """Returns the object of database class to perform operations on CosmosDB database
        Args:
            database_name:  (str)   --  Name of database

        Raises:
            Exception:
                If database with given name does not exist
        """
        try:
            return self.client.get_database_client(database_name)
        except Exception as exp:
            self.log.exception("Database doesn't exist or not able to return object: %s", exp)
            raise exp

    def delete_database(self, database_name):
        """Deletes the database
        Args:
            database_name:  (str)   --  Name of database

        Raises:
            Exception:
                If database with given name does not exist
                Database could not be deleted due to CosmosDB issue
        """
        try:
            self.client.delete_database(database_name)
            self.log.info("database deleted successfully")
        except Exception as exp:
            self.log.exception("delete database failed with error: %s", exp)
            raise exp

    def create_container(self, database_name, container_name, partition_key, throughput=0):
        """Creates a CosmosDB container with given attributes
        Args:
            database_name:  (str)   --  Name of database

            container_name  (str)   --  Name of container

            partition_key   (str)   -- Name of the partition key column

            throughput      (int)   --  Throughput to be assigned

        Raises:
            Exception:
                If database doesnt exist
                invalid throughput value is provided
                container could not be created due to CosmosDB issue
        """
        try:
            containers = self.get_db_client(database_name).list_containers()
            for container in containers:
                if container_name in container.get('id'):
                    self.log.info('Database with name %s exists, '
                                  'deleting and recreating it', container_name)
                    self.delete_container(container_name, database_name)
            if throughput > 0:
                self.get_db_client(database_name).create_container(
                    id=container_name, partition_key=PartitionKey(path="/"+partition_key),
                    offer_throughput=throughput)
            else:
                self.get_db_client(database_name).create_container(
                    id=container_name,
                    partition_key=PartitionKey(path="/"+partition_key))
            self.log.info("Container created successfully")
        except Exception as exp:
            self.log.exception("Create container failed with error: %s", exp)
            raise exp

    def get_cnt_client(self, database_name, container_name):
        """Returns the object of container class to perform operations on CosmosDB container
        Args:
            database_name:  (str)   --  Name of database

            container_name  (str)   --   Name of container

        Returns:
            The object of the container class for given container

        Raises:
            Exception:
                If database does not exist
                If container does not exist
        """
        try:
            return self.get_db_client(
                database_name).get_container_client(container_name)
        except Exception as exp:
            self.log.exception("Container doesnt exist or not able to return object:%s", exp)
            raise exp

    def populate_container(self, database_name, container_name, partition_key,
                           num_items, start=1):
        """Populates the given container with test data
        Args:
            database_name:  (str)   --  Name of database

            container_name  (str)   --  Name of container

            partition_key   (str)   --  The name of the partition key column

            num_items       (int)   --  The number of items to populate

            start           (int)   --  The item/row number to start populating
                default is 1

        Raises:
            Exception:
                Non existing database or container name is passed
                Container could not be populate due to CosmosDB issue
        """
        try:
            container_client = self.get_cnt_client(database_name, container_name)
            for i in range(start, start+num_items):
                container_client.upsert_item({
                    partition_key: i,
                    'item': 'item' + str(i),
                    'timestamp': time.time(),
                    'description': 'CV AUTOMATION TESTCASE',
                })
            self.log.info("Container populated successfully")
        except Exception as exp:
            self.log.exception("Error when populating container:%s", exp)
            raise exp

    def delete_container(self, database_name, container_name):
        """Deletes the container under given database
        Args:
            database_name:  (str)   --  Name of database

            container_name  (str)   --  Name of container
        Raises:
            Exception:
                Non existing database or container name is provided
                Container could not be deleted due to CosmosDB issue
        """
        try:
            self.get_db_client(database_name).delete_container(container_name)
            self.log.info("Container deleted successfully")
        except Exception as exp:
            self.log.exception("delete container failed with error: %s", exp)
            raise exp

    def validate_container(self, database_name, container_name, partition_key,
                           expected_count, start=1):
        """Validates the items in the container after restore
        Args:
            database_name:  (str)   --  Name of database

            container_name  (str)   --  Name of container

            partition_key   (str)   --  The name of the partition key column

            expected_count  (int)   --  The count of items on source

            start           (int)   --  The item/row number to start validating
                default is 1

        Raises:
            Exception:
                Non existing database or container name is passed
                If items in container could not be read due to CosmosDB issue
        """
        try:
            container_client = self.get_cnt_client(database_name, container_name)
            for item in container_client.query_items(
                    query=f'SELECT * FROM {container_name} c order by c.{partition_key}',
                    enable_cross_partition_query=True):
                if(item.get(partition_key) == start and item.get('item') == 'item'+str(start)
                        and item.get('description') == 'CV AUTOMATION TESTCASE'):
                    start += 1
                else:
                    raise Exception("Container validation failed, did not find expected items")
            if start != expected_count+1:
                raise Exception("Item count is not matching after restore")
            self.log.info("Container validated successfully")
        except Exception as exp:
            self.log.exception("Validating container failed with error: %s", exp)
            raise exp
