# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
# See LICENSE.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""Main Module for invoking OneDrive APIs.

OneDrive is the only class defined in this file.

OneDrive: Class for performing OneDrive related operations

OneDrive:

        __init__(self, cc_object)           --  Initializes the OneDrive object

        discover(self)                      --  Method to discover users on azure AD

        create_files()                      --  This method creates and
        upload files on user's OneDrive

        get_file_properties()               --  Method to fetch file properties from
        OneDrive Automation folder and store them to local DB

        get_folder_id_from_graph()         --  Method to fetch folder id of
        the automation folder

        _get_files_list()                   --  Method to get the list of file metadata

        compare_file_properties()           --  Method to compare file properties

        _compare_files(self, user_id)       --  Compares the file properties of
        a single user

        delete_folder()                     --  Method to delete automation folder
        from user OneDrive

        _process_response()                 --  Process the Graph API response and
        returns the minimal dict

        request()                           --  Method to make Graph API requests

"""

import os
import json
import time
from requests.exceptions import HTTPError
from oauthlib.oauth2.rfc6749.errors import TokenExpiredError
from AutomationUtils.machine import Machine
from .wiki import Wiki
from .exception import CVCloudException
from . import constants
from .csdb_helper import CSDBHelper


class OneDrive:
    """Class for performing OneDrive related operations."""

    def __init__(self, cc_object):
        """Initializes the oneDrive object.

                Args:

                    cc_object  (Object)  --  instance of cloud_connector module


                Returns:

                    object  --  instance of OneDrive class

        """

        self.tc_object = cc_object.tc_object
        self.log = self.tc_object.log
        self.api_name = self.__class__.__name__
        self.auth_object = cc_object.one_drive_auth
        self.auth_object.authenticate_client()
        self.dbo = cc_object.dbo
        self.sqlite = cc_object.sqlite
        self.wiki = Wiki(self.log)
        self.machine = Machine()
        self.headers = {
            'Content-Type': 'application/json',
            'Host': 'graph.microsoft.com'
        }
        self.csdb = CSDBHelper(self.tc_object)

    def discover(self):
        """Method to discover users on azure AD"""

        try:
            users_endpoint = (f'{constants.MS_GRAPH_ENDPOINT}' 
                              f'{constants.MS_GET_USERS}' 
                              f'?$top={str(constants.USER_PAGE_SIZE)}')
            self.log.info('users_endpoint:%s', users_endpoint)

            response = self.request(url=users_endpoint, headers=self.headers)
            self.log.info('response:%s', response)
            self.log.info('response length: %s', len(response.get('value')))
            self.dbo.save_into_table('users', response.get('value'), '_list')

            while '@odata.nextLink' in response:
                self.log.info('@odata.nextLink:%s', response.get('@odata.nextLink'))
                response = self.request(url=response.get('@odata.nextLink'), headers=self.headers)
                self.dbo.save_into_table('users', response.get('value'), '_list')
                self.log.info('response:%s', response)
        except Exception as excp:
            self.log.exception('Exception while discovering users')
            raise CVCloudException(self.api_name, '101', str(excp))

    def create_files(self, unicode_data=False,
                     no_of_docs=constants.NO_OF_DOCS_TO_CREATE,
                     pdf=False,
                     new_folder=True,
                     user=None):
        """This method creates and upload files on user's OneDrive

                Args:

                    unicode_data  (boolean)  --  True if unicode data files
                    have to be created

                    no_of_docs  (int)  --  Number of files to create and upload

                    pdf  (boolean)  --  True if pdf files have to be created

                    new_folder  (boolean)  --  True if a new folder has to be
                    created on user's GDrive

                    user  (str)   --  SMTP address of the user

        """

        if user:
            sc_content = [{'SMTPAddress': user}]
        else:
            sc_content = self.tc_object.subclient.content
        for content in sc_content:
            user_id = content.get('SMTPAddress')
            self.wiki.create_docx(
                no_of_docs=no_of_docs,
                unicode_data=unicode_data,
                pdf=pdf,
                ca_type='onedrive')

            try:
                self.log.info(
                    'creating documents into users OneDrive: "%s"', user_id)
                if new_folder:
                    self.log.info('creating folder on users onedrive')
                    folder_endpoint = f'{constants.MS_GRAPH_ENDPOINT}{constants.MS_GET_USERS}' \
                                      f'/{user_id}{constants.MS_CREATE_FOLDER}'
                    self.log.info('create folder endpoint: %s', folder_endpoint)
                    data = {
                        "name": "AutomationFolder",
                        "folder": {},
                        "@microsoft.graph.conflictBehavior": "replace"
                    }

                    resp = self.request(
                        method='POST',
                        url=folder_endpoint,
                        data=json.dumps(data),
                        headers=self.headers)
                    self.log.info('folder creation response: %s', resp)
                    self.dbo.save_into_table(user_id, resp, constants.FOLDER_TABLE)
                else:
                    self.log.info('uploading files to existing folder')
                self.log.info('Purging the latest files db before saving current properties')
                self.dbo.dbo.purge_table(f'{user_id}{constants.LATEST_FILES}')
                for root, directories, files in os.walk(
                        str(constants.ONEDRIVE_DOCUMENT_DIRECTORY)):
                    for file_name in files:
                        # Join the two strings in order to form the full filepath.
                        file_path = os.path.join(root, file_name)
                        self.log.info('uploading file: %s', file_path)
                        self.log.info('filename: %s', file_name)
                        file_endpoint = (
                            f'{constants.MS_GRAPH_ENDPOINT}' 
                            f'{constants.MS_GET_USERS}' 
                            f'{constants.MS_UPLOAD_FILE.format(user_id, constants.ONEDRIVE_FOLDER, file_name)}'
                        )
                        self.log.info('Upload file endpoint: %s', file_endpoint)
                        headers = self.headers.copy()
                        headers['Content-Type'] = 'text/plain'
                        resp = self.request(
                            url=file_endpoint, method='PUT', data=open(
                                file_path, mode='rb'), headers=headers)
                        self.log.info('file creation response: %s', resp)
                        processed_response = self._process_response({'value': [resp]})
                        self.dbo.save_into_table(user_id, processed_response, constants.LATEST_FILES)

            except Exception as excp:
                self.log.exception('exception occurred while creating folder on OneDrive')
                raise CVCloudException(self.api_name, '103', str(excp))

    def get_file_properties(self, user=None, folder=constants.ONEDRIVE_FOLDER):
        """
        Method to fetch file properties from OneDrive Automation folder and store them to local DB

            Agrs:

                user  (str)   --  SMTP address of the user

        """

        if user:
            sc_content = [{'SMTPAddress': user}]
        else:
            sc_content = self.tc_object.subclient.content

        try:
            for content in sc_content:
                user_id = content.get('SMTPAddress')

                folder_id = self.get_folder_id_from_graph(user_id, folder)
                self.log.info('Fetching items from Automation folder')
                self._get_files_list(user_id, folder_id, save_to_db=True)

        except Exception as excp:
            self.log.exception('Exception while get file properties')
            raise CVCloudException(self.api_name, '104', str(excp))

    def get_folder_id_from_graph(self, user_id, folder=constants.ONEDRIVE_FOLDER):
        """This method fetches the folder id of the automation folder on users OneDrive

                Args:

                    user_id (str)   --  User SMTP address

                Returns:

                    folder_id   (str)   --  ID of the automation folder

        """
        try:
            get_root_children_endpoint = (
                f'{constants.MS_GRAPH_ENDPOINT}'
                f'{constants.MS_GET_USERS}' 
                f'{constants.MS_GET_ROOT_CHILDREN.format(user_id)}'
            )
            self.log.info('Get root children endpoint: %s', get_root_children_endpoint)
            self.log.info('Folder to search for: %s', folder)

            folder_id = None

            while not folder_id:

                response = self.request(url=get_root_children_endpoint, headers=self.headers)
                for element in response.get('value'):
                    if element.get('name') == folder:
                        folder_id = element.get('id')
                        self.log.info('Got folder ID from Graph: %s', folder_id)
                        self.dbo.dbo.purge_table(f'{user_id}{constants.FOLDER_TABLE}')
                        self.dbo.save_into_table(user_id, element, constants.FOLDER_TABLE)
                        break

                if not folder_id:
                    if '@odata.nextLink' in response:
                        get_root_children_endpoint = response['@odata.nextLink']
                    else:
                        self.log.error('Folder does not exist on OneDrive')
                        break

            return folder_id

        except Exception as excp:
            self.log.exception(
                'Exception occurred while fetching automation folder id from MS Graph.')
            raise CVCloudException(self.api_name, '107', str(excp))

    def _get_files_list(self, user_id, folder_id, save_to_db=True):
        """This method gets the list of file metadata from users OneDrive

                Args:

                    user_id (str)   --  SMTP address of the user

                    folder_id   (str)   --  ID of the folder from file properties have to be fetched

                    save_to_db  (bool)  --  If true, file properties will be saved to local db
                    If false, files properties will be returned

                Returns:

                    file_properties (list)  --  list of file properties dicts

        """
        final_response = []
        get_folder_children_endpoint = (
            f'{constants.MS_GRAPH_ENDPOINT}' 
            f'{constants.MS_GET_USERS}'
            f'{constants.MS_GET_FOLDER_CHILDREN.format(user_id, folder_id)}'
        )
        self.log.info('Get folder children endpoint: %s', get_folder_children_endpoint)

        response = self.request(url=get_folder_children_endpoint, headers=self.headers)
        self.log.info('response:%s', response)
        processed_response = self._process_response(response)

        if save_to_db:
            self.log.info('Purging the db before saving current properties')
            self.dbo.dbo.purge_table(f'{user_id}{constants.ONEDRIVE_TABLE}')
            self.dbo.save_into_table(
                user_id,
                processed_response,
                data_type=constants.ONEDRIVE_TABLE)
        else:
            final_response = processed_response

        while '@odata.nextLink' in response:
            self.log.info('@odata.nextLink:%s', response.get('@odata.nextLink'))
            response = self.request(url=response.get('@odata.nextLink'), headers=self.headers)
            self.log.info('response:%s', response)
            processed_response = self._process_response(response)

            if save_to_db:
                self.dbo.save_into_table(
                    user_id,
                    processed_response,
                    data_type=constants.ONEDRIVE_TABLE)
            else:
                final_response.extend(processed_response)

        return final_response

    def compare_file_properties(self, oop=False, to_disk=False, incremental=False):
        """Method to compare file properties before backup and after restore."""

        sc_content = self.tc_object.subclient.content
        try:
            if not oop:
                for content in sc_content:
                    user_id = content.get('SMTPAddress')
                    self._compare_files(user_id=user_id, incremental=incremental)
            else:
                if to_disk:
                    # This is OOP restore to disk
                    self.log.info('This is oop restore to disk comparison')
                    job_list = self.dbo.get_content('job_list_disk')[0]['job_list']
                    user_id = job_list[3]
                    self.download_files(user=user_id)
                    source_path = os.path.join(constants.ONEDRIVE_DOWNLOAD_DIRECTORY, user_id)
                    destination_path = constants.DESTINATION_TO_DISK_AFTER_RESTORE % (
                        constants.DESTINATION_TO_DISK,
                        user_id,
                        constants.ONEDRIVE_FOLDER)
                    self.log.info('destination path on proxy client: %s', destination_path)
                    proxy_client = Machine(
                        self.tc_object.instance.proxy_client,
                        self.tc_object.commcell)

                    diff = self.machine.compare_folders(destination_machine=proxy_client,
                                                        source_path=source_path,
                                                        destination_path=destination_path)

                    if diff:
                        self.log.error('Following files are not matching after restore', diff)
                        raise CVCloudException(self.api_name, '111')
                    self.log.info('File hashes are matching after disk restore')
                    proxy_client.remove_directory(destination_path)
                else:
                    # This is OOP to other user account
                    job_list = self.dbo.get_content('job_list_oop')[0]['job_list']
                    user_id = job_list[2]
                    self.get_file_properties(user=user_id)
                    self._compare_files(user_id=user_id)

        except Exception as excp:
            self.log.exception('Exception in compare_files operation on OneDrive')
            raise CVCloudException(self.api_name, '107', str(excp))

    def _compare_files(self, user_id, incremental=False):
        """Method to compare files on OneDrive with file metadata in local db

                Args:
                    user_id  (str)   --  SMTP address of the user

                Raises exception on comparison failure
        """
        if incremental:
            user_table = self.dbo.get_content(f'{user_id}{constants.LATEST_FILES}')
        else:
            user_table = self.dbo.get_content(f'{user_id}{constants.ONEDRIVE_TABLE}')

        folder_id = self.get_folder_id_from_graph(user_id)

        if not folder_id:
            raise CVCloudException(self.api_name, '105')

        files_list = self._get_files_list(user_id, folder_id, save_to_db=False)

        self.log.info('Number of Backed up documents: %s', len(user_table))
        self.log.info('Number of restored documents: %s', len(files_list))

        if len(user_table) != len(files_list):
            self.log.error(
                'Number of restored documents is different than the number of backed up documents.')
            raise CVCloudException(self.api_name, '108')

        for file_metadata in files_list:
            self.log.info('file metadata: %s', file_metadata)
            match = False
            for file in user_table:
                self.log.info('file from user table: %s', file)
                if 'quickXorHash' in file:
                    if file.get('quickXorHash') == file_metadata.get('quickXorHash') and file.get(
                            'parentPath') == file_metadata.get('parentPath'):
                        self.log.info('file properties are matching')
                        match = True
                        break
                elif 'childCount' in file:
                    if file.get('childCount') == file_metadata.get('childCount') and file.get(
                            'parentPath') == file_metadata.get('parentPath'):
                        self.log.info('folder properties are matching')
                        match = True
                        break
            if not match:
                self.log.exception('File properties are not matching after restore')
                raise CVCloudException(self.api_name, '109')

    def delete_folder(self, user_id=None, folder_name=constants.ONEDRIVE_FOLDER):
        """Method to delete the automation folder from user OneDrive.

                Args:

                    user_id (str)   --  User SMTP Address

                    folder_name  (str)  --  Name of the folder to delete
                        Default: ONEDRIVE_FOLDER value from constants

        """

        try:
            if user_id:
                sc_content = [{'SMTPAddress': user_id}]
            else:
                sc_content = self.tc_object.subclient.content

            for content in sc_content:
                user_id = content.get('SMTPAddress')
                folder_id = self.get_folder_id_from_graph(user_id, folder_name)
                if folder_id:
                    delete_folder_endpoint = (
                            f'{constants.MS_GRAPH_ENDPOINT}' 
                            f'{constants.MS_GET_USERS}' 
                            f'{constants.MS_DELETE_ITEM.format(user_id, folder_id)}'
                    )
                    self.log.info('Delete folder endpoint: %s', delete_folder_endpoint)
                    resp = self.request(
                        url=delete_folder_endpoint,
                        method='DELETE',
                        headers=self.headers)
                    self.log.info('Delete folder response: %s', resp)
        except Exception as excp:
            self.log.exception('Exception while deleting folder on OneDrive.')
            raise CVCloudException(self.api_name, '107', str(excp))

    def download_files(self, user=None):
        """
        Method to download all the files from OneDrive Automation folder.
        Folder id is fetched from local DB file.

            Args:

                user    (str)   --  SMTP Address of the user

        """
        if user:
            sc_content = [{'SMTPAddress': user}]
        else:
            sc_content = self.tc_object.subclient.content
        try:
            for content in sc_content:
                user_id = content.get('SMTPAddress')

                db_content = self.dbo.get_content(f'{user_id}{constants.ONEDRIVE_TABLE}')

                download_directory = os.path.join(constants.ONEDRIVE_DOWNLOAD_DIRECTORY, user_id)

                os.makedirs(download_directory, exist_ok=True)
                self.log.info('Downloading the files')

                for files in db_content:
                    self.log.info('file: %s', files)
                    self.download_single_file(
                        user_id=user_id,
                        file_path=files['parentPath'] + '/' + files['name'],
                        download_directory=download_directory
                    )

        except Exception as excp:
            self.log.exception('Exception in download_files operation on OneDrive')
            raise CVCloudException(self.api_name, '107', str(excp))

    def download_single_file(self, user_id, file_path, download_directory):
        """Method to download a single file from OneDrive based on the file path.
        File is downloaded in one drive download folder defined in constants file.

                Args:

                    user_id    (str)   --  SMTP Address of the user

                    file_path (str)  --  File path relative to OneDrive root folder

                    download_directory (str)  --  Directory name to download files into

        """
        try:
            download_file_endpoint = (
                f'{constants.MS_GRAPH_ENDPOINT}' 
                f'{constants.MS_GET_USERS}/{user_id}{file_path}'
            )

            self.log.info('Download file endpoint: %s', download_file_endpoint)
            resp = self.request(url=download_file_endpoint, headers=self.headers)
            download_url = resp['@microsoft.graph.downloadUrl']
            self.log.info('Download file url: %s', download_url)
            resp = self.request(url=download_url)
            self.log.info('Download file response: %s', resp)
            file_name = file_path.split('/')[-1]
            download_file_path = os.path.join(download_directory, file_name)
            self.log.info('Download file location: %s', download_file_path)
            with open(download_file_path, 'wb') as fp:
                fp.write(resp.content)

        except Exception as excp:
            self.log.exception('Exception in download_single_file operation on OneDrive')
            raise CVCloudException(self.api_name, '107', str(excp))

    def _process_response(self, response):
        """Method to process the response got from MS Graph API.
        This method returns name, etag, ctag and size of the file

            Args:

                    response    (dict)  --  Raw response got from MS Graph API Call

            Returns:

                processed_response  (list)  --  List of minimal dicts containing
                name, etag, ctag, size, id of the files

        """
        try:
            processed_response = []
            if 'value' in response:
                for element in response.get('value'):
                    temp_element = {
                        'eTag': element.get('eTag'),
                        'id': element.get('id'),
                        'name': element.get('name'),
                        'cTag': element.get('cTag'),
                        'size': element.get('size'),
                        'parentPath': element.get('parentReference').get('path')
                    }

                    if 'file' in element:
                        temp_element['quickXorHash'] = element.get(
                            'file').get('hashes').get('quickXorHash')

                    if 'folder' in element:
                        temp_element['childCount'] = element.get('folder').get('childCount')

                    processed_response.append(temp_element)
                return processed_response
            else:
                self.log.error('Invalid input response for the method')
                raise CVCloudException(self.api_name, '106')
        except Exception as excp:
            self.log.exception('Exception while processing graph response.')
            raise CVCloudException(self.api_name, '107', str(excp))

    def request(self, url=None, method='GET', data=None, headers=None):
        """Method to make a request to Graph URL via OAUTH request method

                Args:

                    url (str)   --  API Endpoint to make request to

                    method (str)    --  API call method

                    Valid Values:
                        GET

                        POST

                        PUT

                        DELETE

                        PATCH

                    data (dict) --  Data to be sent with POST/PUT requests

                    headers (dict)  --  Request headers.

                    Default:

                    self._headers

                    Access token is automatically appended by
                    OAuth2Session class request method used in this module.

        """

        try:
            if not self.auth_object.oauth.authorized:
                self.auth_object.authenticate_client()

            time_out = 0

            while time_out <= 3:

                try:

                    if method in ['GET', 'DELETE']:
                        resp = self.auth_object.oauth.request(
                            method=method, url=url, headers=headers)
                    elif method in ['POST', 'PUT', 'PATCH']:
                        self.log.info('This is a %s method', method)
                        self.log.info('data: %s', data)
                        resp = self.auth_object.oauth.request(
                            method=method, url=url, data=data, headers=headers)

                    else:
                        self.log.error('Method %s not supported', method)
                        raise CVCloudException(self.api_name, '102')
                    self.log.info('status code: %s', resp.status_code)
                    if not resp.ok:
                        self.log.error('error: %s', resp.json())
                        resp.raise_for_status()
                    else:
                        self.log.info('Got success response code: %s', resp.status_code)
                        try:
                            return resp.json()
                        except Exception:
                            return resp
                except HTTPError:
                    # If the error is a rate limit, connection error or backend error,
                    # wait and try again.
                    time_out += 1
                    self.log.error('Encountered HTTP Error: {0}'.format(resp.status_code))
                    # Handle the error accordingly [400, 404, 429, 500, 503]

                    if resp.status_code == 401:
                        self.log.error('Request unauthorized.Trying to authenticate client..')
                        self.auth_object.authenticate_client()
                        continue
                    if resp.status_code == 404 and method == 'DELETE':
                        self.log.info('Item does not exist on OneDrive')
                        break
                    elif resp.status_code == 429:
                        self.log.info(
                            'Got throttled from MS server. '
                            'Waiting for few seconds before trying again')
                        time.sleep(time_out * 10)
                        # This is backend error from MS Graph so we need to retry after waiting
                        continue
                    elif resp.status_code == 500:
                        self.log.info(
                            'This is internal server error from MS Graph. '
                            'Waiting for few seconds before trying again')
                        time.sleep(time_out * 10)
                        # This is backend error from MS Graph so we need to retry after waiting
                        continue
                    elif resp.status_code == 501:
                        self.log.info('The requested feature isn’t implemented.')
                        raise
                    elif resp.status_code == 503:
                        self.log.info(
                            'Service unavailable currently. '
                            'Waiting as per Retry-After header and try again.')
                        t_value = resp.headers.get('Retry-After')
                        time.sleep(t_value)
                        continue
                    elif resp.status_code == 504:
                        self.log.info('Gateway timeout error. Retry after sometime.')
                        time.sleep(time_out * 10)
                        # This is backend error from MS Graph so we need to retry after waiting
                        continue
                    elif resp.status_code == 507:
                        self.log.info(
                            'Insufficient Storage. The maximum storage quota has been reached.')
                        raise
                    elif resp.status_code == 509:
                        self.log.info('Bandwidth Limit Exceeded')
                        raise
                except TokenExpiredError:
                    self.log.warning('Token expired. fetching a new token and trying again.')
                    self.auth_object.authenticate_client()
                    continue
            if time_out > 3:
                self.log.error('Maximum retries for Grapah request exceeded.')
                raise CVCloudException(self.api_name, '110')
        except Exception as excp:
            self.log.exception('Exception while processing graph response.')
            raise CVCloudException(self.api_name, '107', str(excp))

    def delete_single_file(self, file_name, user_id=None):
        """Method to delete a single file from OneDrive based on the file path.

                Args:

                    user_id    (str)   --  SMTP Address of the user

                    file_name (str)  --  File name to Delete from OneDrive account

        """
        try:
            if user_id:
                sc_content = [{'SMTPAddress': user_id}]
            else:
                sc_content = self.tc_object.subclient.content

            for content in sc_content:
                user_id = content.get('SMTPAddress')

                self.get_file_properties()
                file_content = self.dbo.get_content(f'{user_id}{constants.ONEDRIVE_TABLE}')

                file_id = None
                for i in file_content:
                    if i['name'] == file_name:
                        file_id = i['id']
                        break

                if file_id:
                    delete_file_endpoint = (
                        f'{constants.MS_GRAPH_ENDPOINT}'
                        f'{constants.MS_GET_USERS}'
                        f'{constants.MS_DELETE_ITEM.format(user_id, file_id)}'
                    )
                    self.log.info('Delete file endpoint: %s', delete_file_endpoint)
                    resp = self.request(
                        url=delete_file_endpoint,
                        method='DELETE',
                        headers=self.headers)
                    self.log.info('Delete file response: %s', resp)

        except Exception as excp:
            self.log.exception('Exception while deleting file on OneDrive.')
            raise CVCloudException(self.api_name, '107', str(excp))

    def upload_files_to_onedrive(self, file_location, no_of_files):
        """Method to generate files and upload them to one_drive.

                Args:

                    file_location    (str)   --  file location on the machine to generate files
                                                    (Example: file_location - 'C:\Temp')

                    no_of_files (str)  --  Number of files to generate in the file location to upload

                """
        self.machine.generate_test_data(file_path=file_location, dirs=1, files=no_of_files)
        user_id = None
        sc_content = self.tc_object.subclient.content

        for content in sc_content:
            user_id = content.get('SMTPAddress')
        try:
            folder_endpoint = f'{constants.MS_GRAPH_ENDPOINT}{constants.MS_GET_USERS}' \
                              f'/{user_id}{constants.MS_CREATE_FOLDER}'

            data = {
                "name": "AutomationFolder",
                "folder": {},
                "@microsoft.graph.conflictBehavior": "replace"
            }

            resp = self.request(method='POST', url=folder_endpoint, data=json.dumps(data), headers=self.headers)
            self.log.info('folder creation response: %s', resp)

            for root, dirs, files in os.walk(file_location):
                if os.path.basename(root) == 'regular':
                    for file in files:
                        file_path = os.path.join(root, file)
                        file_endpoint = (f'{constants.MS_GRAPH_ENDPOINT}'
                                         f'{constants.MS_GET_USERS}'
                                         f'{constants.MS_UPLOAD_FILE.format(user_id, "AutomationFolder", file)}'
                                         )
                        self.headers = self.headers.copy()
                        self.headers['Content-Type'] = 'text/plain'
                        resp = self.request(
                            url=file_endpoint,
                            method='PUT',
                            data=open(file_path, mode='rb'),
                            headers=self.headers)
                        self.log.info('Files uploading response: %s', resp)
        except Exception as excp:
            self.log.exception('exception occurred while creating files on OneDrive')
            raise CVCloudException(self.api_name, '103', str(excp))

    def verify_finalize_phase(self, job_id):
        """Method to verify onedrive backup finalize phase.

            Args:
                job_id   (str)   --  id of a backup job

        """
        user_id = None
        sc_content = self.tc_object.subclient.content
        for content in sc_content:
            user_id = content.get('SMTPAddress')

        try:
            job_info = self.sqlite.get_job_info_local_db(job_id=job_id)
            job_info = job_info[0]
            if job_info:
                if job_info[0] == int(job_id):
                    self.log.info('Job id from the local db: [%s]', job_info[0])

                    if job_info[1] == 'ok':
                        self.log.info('Job Status from the local db: [%s]', job_info[1])
                    else:
                        self.log.info('Job Status from the local db: [%s]', job_info[1])
                        raise Exception('Finalize phase failed')

                    change_link = job_info[2]
                    token = change_link.rsplit('=', 1)[1]

                    file_endpoint = (f'{constants.MS_GRAPH_ENDPOINT}'
                                     f'{constants.MS_GET_USERS}'
                                     f'{constants.ONEDRIVE_DELTA_QUERY.format(user_id, token)}'
                                     )

                    resp = self.request(method='GET', url=file_endpoint, headers=self.headers)
                    if resp['value']:
                        raise Exception('Failed to Update the Change link token')
                    else:
                        self.log.info('Successfully updated change link token')

                else:
                    raise Exception('Error in getting Job info from Local DB')

            path = self.machine.get_registry_value(f'{constants.REG_KEY_BASE}', f'{constants.PATH_KEY}')
            temp_sqlite_file = (
                f'{path}'
                f'{constants.ONEDRIVE_JOB_DIRECTORY_PATH.format(self.tc_object.subclient.subclient_id, f"Temp_{user_id}")}'
            )
            if os.path.exists(temp_sqlite_file):
                raise Exception("Temporary local db failed to move")
            else:
                self.log.info("Temporary local db successfully moved")

        except Exception as excp:
            self.log.exception('Exception during Finalize phase verification')
            raise CVCloudException(self.api_name, '112', str(excp))

    def discover_groups(self):
        """Method to discover users on azure AD"""

        try:
            groups_endpoint = (f'{constants.MS_GRAPH_ENDPOINT}' 
                               f'{constants.MS_GET_GROUPS}' 
                               f'?$top={str(constants.USER_PAGE_SIZE)}')
            self.log.info('groups_endpoint:%s', groups_endpoint)

            response = self.request(url=groups_endpoint, headers=self.headers)
            self.log.info('response:%s', response)
            self.log.info('response length: %s', len(response.get('value')))
            self.dbo.save_into_table('groups', response.get('value'), '_list')

            while '@odata.nextLink' in response:
                self.log.info('@odata.nextLink:%s', response.get('@odata.nextLink'))
                response = self.request(url=response.get('@odata.nextLink'), headers=self.headers)
                self.dbo.save_into_table('groups', response.get('value'), '_list')
                self.log.info('response:%s', response)
                self.log.info('response length: %s', len(response.get('value')))
        except Exception as excp:
            self.log.exception('Exception while discovering groups')
            raise CVCloudException(self.api_name, '101', str(excp))

    def discover_group_members(self, group_name):
        """Method to discover members of a group on azure AD"""
        try:
            group_id_endpoint = (f'{constants.MS_GRAPH_ENDPOINT}'
                                 f'{constants.MS_GET_GROUPS}'
                                 f'?$filter=startsWith(displayName,\'{group_name}\')')
            self.log.info('groups_endpoint:%s', group_id_endpoint)
            response = self.request(url=group_id_endpoint, headers=self.headers)
            self.log.info(':%s', response)
            group_id = response['value'][0]['id']
            self.log.info(f'group id: {group_id}')

            members_endpoint = (f'{constants.MS_GRAPH_ENDPOINT}'
                                f'{constants.MS_GET_GROUPS}'
                                f'{constants.MS_GET_GROUP_MEMBERS.format(group_id)}')
            self.log.info('group_members_endpoint:%s', members_endpoint)

            response = self.request(url=members_endpoint, headers=self.headers)
            self.log.info('response:%s', response)
            self.log.info('response length: %s', len(response.get('value')))
            self.dbo.save_into_table(f'{group_name.lower()}_members', response.get('value'), '_list')

            while '@odata.nextLink' in response:
                self.log.info('@odata.nextLink:%s', response.get('@odata.nextLink'))
                response = self.request(url=response.get('@odata.nextLink'), headers=self.headers)
                self.dbo.save_into_table('groups', response.get('value'), '_list')
                self.log.info('response:%s', response)
                self.log.info('response length: %s', len(response.get('value')))
        except Exception as excp:
            self.log.exception('Exception while discovering group members')
            raise CVCloudException(self.api_name, '101', str(excp))

    def create_user(self, user_name):
        """Method to discover users on azure AD"""

        try:
            users_endpoint = (f'{constants.MS_GRAPH_ENDPOINT}'
                              f'{constants.MS_GET_USERS}')
            data = constants.ONEDRIVE_NEW_USER
            data['userPrincipalName'] = f'{user_name}@{constants.ONEDRIVE_TENANT_DOMAIN}'
            data['mailNickname'] = f'{user_name}'
            self.log.info('users_endpoint:%s', users_endpoint)

            response = self.request(url=users_endpoint,
                                    method='POST',
                                    data=json.dumps(data),
                                    headers=self.headers)
            self.log.info('response:%s', response)
            self.dbo.save_into_table('new_users', response, '_list')
            self.log.info(f'Successfully created user: {user_name}')
        except Exception as excp:
            self.log.exception('Exception while creating user')
            raise CVCloudException(self.api_name, '101', str(excp))

    def update_user_email_id(self, user_name, new_email_id):
        """
        Method to update email id of user on Azure AD

        Args:
            user_name (str):    User whose email id has to be changed
            new_email_id (str): New user name to be updated in email

        """
        try:
            users_endpoint = (f'{constants.MS_GRAPH_ENDPOINT}'
                              f'{constants.MS_GET_USERS}/'
                              f'{user_name}@{constants.ONEDRIVE_TENANT_DOMAIN}')
            data = constants.ONEDRIVE_UPDATE_USER
            data['userPrincipalName'] = f'{new_email_id}@{constants.ONEDRIVE_TENANT_DOMAIN}'
            data['mailNickname'] = f'{new_email_id}'
            self.log.info('users_endpoint:%s', users_endpoint)

            response = self.request(url=users_endpoint,
                                    method='PATCH',
                                    data=json.dumps(data),
                                    headers=self.headers)
            self.log.info('response:%s', response)
            self.log.info(f'Successfully updated user: {user_name}')
        except Exception as excp:
            self.log.exception('Exception while updating user')
            raise CVCloudException(self.api_name, '101', str(excp))

    def delete_user(self, user_name):
        """
        Deletes the given user from the Azure AD

        Args:
            user_name (str):    User to be deleted

        """
        try:
            user_endpoint = (f'{constants.MS_GRAPH_ENDPOINT}'
                             f'{constants.MS_GET_USERS}/'
                             f'{user_name}@{constants.ONEDRIVE_TENANT_DOMAIN}')

            response = self.request(url=user_endpoint, method='DELETE', headers=self.headers)
            self.log.info('response:%s', response)
            self.log.info(f'Successfully deleted user: {user_name}')

        except Exception as excp:
            self.log.exception('Exception while deleting user')
            raise CVCloudException(self.api_name, '101', str(excp))

    def create_group(self, group_name):
        """
        Creates the given group

        Args:
            group_name (str):   Group to be created

        """
        try:
            group_endpoint = (f'{constants.MS_GRAPH_ENDPOINT}'
                              f'{constants.MS_GET_GROUPS}')
            data = constants.ONEDRIVE_NEW_GROUP
            data['displayName'] = group_name
            data['mailNickname'] = f'{group_name}'
            self.log.info('group_endpoint:%s', group_endpoint)

            response = self.request(url=group_endpoint,
                                    method='POST',
                                    data=json.dumps(data),
                                    headers=self.headers)
            self.log.info('response:%s', response)
            self.dbo.save_into_table('new_groups', response, '_list')
            self.log.info(f'Successfully created group: {group_name}')
        except Exception as excp:
            self.log.exception('Exception while creating group')
            raise CVCloudException(self.api_name, '101', str(excp))

    def delete_group(self, group_name):
        """
        Deletes the given group

        Args:
            group_name (str):   Group to be deleted

        """
        try:
            group_id = None
            details = self.dbo.get_content('new_groups_list')
            for group in details:
                if group['mailNickname'] == group_name:
                    group_id = group['id']
            group_endpoint = (f'{constants.MS_GRAPH_ENDPOINT}'
                              f'{constants.MS_GET_GROUPS}/{group_id}')

            response = self.request(url=group_endpoint, method='DELETE', headers=self.headers)
            self.log.info('response:%s', response)
            self.log.info(f'Successfully deleted group: {group_name}')

        except Exception as excp:
            self.log.exception('Exception while deleting user')
            raise CVCloudException(self.api_name, '101', str(excp))

    def add_group_member(self, member, group_name):
        """
        Adds the given users to the group in Azure AD

        Args:
            member (str):       User to be added to the group
            group_name (str):   Group to which users have to be added

        """
        try:
            group_id = None
            details = self.dbo.get_content('new_groups_list')
            for group in details:
                if group['mailNickname'] == group_name:
                    group_id = group['id']
            group_endpoint = (f'{constants.MS_GRAPH_ENDPOINT}'
                              f'{constants.MS_GET_GROUPS}'
                              f'{constants.MS_GET_GROUP_MEMBERS.format(group_id)}/$ref')

            user_id = None
            details = self.dbo.get_content('new_users_list')
            for user in details:
                if user['userPrincipalName'].split('@')[0] == member:
                    user_id = user["id"]
            data = dict()
            data["@odata.id"] = f'{constants.MS_GRAPH_ENDPOINT}{constants.MS_GET_USERS}/{user_id}'

            response = self.request(url=group_endpoint,
                                    method='POST',
                                    data=json.dumps(data),
                                    headers=self.headers)
            self.log.info('response:%s', response)
            self.log.info(f'Successfully added member {member} to group {group_name}')

        except Exception as excp:
            self.log.exception('Exception while adding group member')
            raise CVCloudException(self.api_name, '101', str(excp))

    def remove_group_member(self, member, group_name):
        """
        Removes the given members from group in Azure AD

        Args:
            member (str):       User to be removed from the group
            group_name (str):   Group from which users have to be removed

        """
        try:
            group_id = None
            details = self.dbo.get_content('new_groups_list')
            for group in details:
                if group['mailNickname'] == group_name:
                    group_id = group['id']

            user_id = None
            details = self.dbo.get_content('new_users_list')
            for user in details:
                if user['userPrincipalName'].split('@')[0] == member:
                    user_id = user["id"]

            group_endpoint = (f'{constants.MS_GRAPH_ENDPOINT}'
                              f'{constants.MS_GET_GROUPS}'
                              f'{constants.MS_GET_GROUP_MEMBERS.format(group_id)}'
                              f'/{user_id}/$ref')

            response = self.request(url=group_endpoint, method='DELETE', headers=self.headers)
            self.log.info('response:%s', response)
            self.log.info(f'Successfully removed member {member} from group {group_name}')

        except Exception as excp:
            self.log.exception('Exception while removing group member')
            raise CVCloudException(self.api_name, '101', str(excp))

    def launch_autodiscovery_via_cli(self, app_name):
        """
        Launch Auto-Discovery on the proxy via command line

        Args:
            app_name (str):     Name of the Office 365 Client for which
                                Autodiscovery is to be launched

        """
        proxy = Machine(machine_name=self.tc_object.instance.proxy_client,
                        commcell_object=self.tc_object.commcell)
        path = proxy.get_registry_value(f'{constants.REG_KEY_BASE}', f'{constants.BASE_KEY}')
        self.log.info(f'Path: {path}')
        instance = proxy.instance
        self.log.info(f'Instance: {instance}')
        client_id, subclient_id = self.csdb.get_subclient_id(app_name)
        client_name, machine_name = self.csdb.get_access_node_details(app_name)

        command = (
            'start-process '
            '-filepath "cvCloudDiscoverV2.exe" '
            f'-workingDirectory "{path}" '
            f'-ArgumentList "-j 0 -a 2:{subclient_id} -c {machine_name} -m {machine_name} '
            f'-t 2 -r 0 -i 0 -d 12 -ci {client_id} -refresh 0 -cn {client_name} -vm {instance}"'
        )
        self.log.info(f'Launch Auto-discovery command: {command}')

        proxy.execute_command(command)
        self.log.info('Auto-discovery launched')

    def _get_all_files_in_folder(self, user_id, folder_path):
        """
        Gets all the files and folders in the given folder

        Args:
            user_id (str):      Email id of the user
            folder_path (str):  Path to the folder from the OneDrive root

        Returns:
            folder_hierarchy (dict):    Dictionary containing folder hierarchy
                                        Format:-
                                        { folder_name : [file_list],
                                                        [{folder1: [...], [...]},
                                                         {folder2: [...], [...]}]
                                        }

        """
        try:
            file_names = list()
            folder_names = list()

            get_folder_children_endpoint = (
                f'{constants.MS_GRAPH_ENDPOINT}'
                f'{constants.MS_GET_USERS}'
                f'{constants.MS_GET_FOLDER_CHILDREN_PATH.format(user_id, folder_path)}'
            )
            self.log.info('Get folder children endpoint: %s', get_folder_children_endpoint)

            response = self.request(url=get_folder_children_endpoint, headers=self.headers)
            for element in response.get('value'):
                if 'file' in element:
                    file_names.append(element.get('name'))
                if 'folder' in element:
                    folder_names.append(element.get('name'))
            while '@odata.nextLink' in response:
                response = self.request(url=response.get('@odata.nextLink'), headers=self.headers)
                for element in response.get('value'):
                    if 'file' in element:
                        file_names.append(element.get('name'))
                    if 'folder' in element:
                        folder_names.append(element.get('name'))

            folders = list()
            for folder in folder_names:
                folder_dict = self._get_all_files_in_folder(user_id, f'{folder_path}/{folder}')
                folders.append(folder_dict)

            folder_hierarchy = {folder_path.split('/')[-1]: [file_names, folders]}
            return folder_hierarchy

        except Exception as excp:
            self.log.exception(
                'Exception occurred while fetching file names from MS Graph.')
            raise CVCloudException(self.api_name, '107', str(excp))

    def get_all_onedrive_items(self, user_id):
        """
        Gets all files and folders in OneDrive
        Not restricted to AutomationFolder

        Args:
            user_id:    email id of user

        Returns:
            folder_hierarchy (dict):    Dictionary containing folder hierarchy
                                        Format:-
                                        { folder_name : [file_list],
                                                        [{folder1: [...], [...]},
                                                         {folder2: [...], [...]}]
                                        }

        """
        try:
            file_names = list()
            folder_names = list()

            get_root_children_endpoint = (
                f'{constants.MS_GRAPH_ENDPOINT}'
                f'{constants.MS_GET_USERS}'
                f'{constants.MS_GET_ROOT_CHILDREN.format(user_id)}'
            )
            self.log.info('Get root children endpoint: %s', get_root_children_endpoint)

            response = self.request(url=get_root_children_endpoint, headers=self.headers)
            for element in response.get('value'):
                if 'file' in element:
                    file_names.append(element.get('name'))
                if 'folder' in element:
                    folder_names.append(element.get('name'))
            while '@odata.nextLink' in response:
                response = self.request(url=response.get('@odata.nextLink'), headers=self.headers)
                for element in response.get('value'):
                    if 'file' in element:
                        file_names.append(element.get('name'))
                    if 'folder' in element:
                        folder_names.append(element.get('name'))

            folders = list()
            for folder in folder_names:
                folder_dict = self._get_all_files_in_folder(user_id, folder)
                folders.append(folder_dict)

            onedrive_folder_hierarchy = {'root': [file_names, folders]}
            self.log.info(f'OneDrive Folder Hierarchy: {onedrive_folder_hierarchy}')
            return onedrive_folder_hierarchy

        except Exception as excp:
            self.log.exception(
                'Exception occurred while fetching file names from MS Graph.')
            raise CVCloudException(self.api_name, '107', str(excp))
