# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
# See LICENSE.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""Helper file for performing Azure Blob Storage operations

AzureHelper is the only class defined in this file.

AzureHelper: Helper class to perform Azure Blob Storage operations

AzureHelper :
    __init__()                           --   initializes Azure helper object

    create_session_azure()               --   creates a session with Azure blob storage

    create_container_azure()             --   creates a container in Azure blob storage

    upload_file_azure()                  --   uploads the specified file to Azure blob storage

    download_file_azure()                --   downloads the specified file from Azure blob storage

    delete_file_azure()                  --   deletes the specified file from Azure blob storage

    create_dir_download_file_azure()     --   checks if folder exists locally. If not,
    creates a directory and downloads a file to it

    delete_contents_azure()              --   deletes all the contents under the specified paths
    from Azure blob storage

    delete_container_azure()             --   deletes the specified container
    from Azure blob storage

    download_container_azure()           --   downloads the specified container with all
    it's contents from Azure blob storage

    download_contents_azure()            --   downloads all the contents under the specified paths
    from the Azure blob storage

    recur()                              --   recursively traverses through the subfolders
    and downloads the blobs

"""

import os
import socket
from pathlib import Path
import azure.common.exceptions
from azure.storage.blob import BlockBlobService
from AutomationUtils import constants
from AutomationUtils.machine import Machine


class AzureHelper(object):
    """Helper class to perform azure blob storage operations"""

    def __init__(self, testcase_object):
        """Initializes azure helper object"""
        self.automation_directory = constants.AUTOMATION_DIRECTORY
        self.log = testcase_object.log
        self.machine = Machine(socket.gethostname())

    def create_session_azure(self, account_name, account_key):
        """creates a session with Azure Blob Storage
        Args :
            account_name       (str)      --       name of the Azure blob account

            account_key        (str)      --       key of the Azure blob account

        Returns :
            (obj)                         --       Returns the azure blob session object

        Raises :
            Exception :
                if the account name or account key is invalid

        """
        try:
            session = BlockBlobService(
                account_name=account_name, account_key=account_key)
        except Exception as error:
            self.log.error(
                "The specified account name or account key is invalid")
            self.log.error(error)
            raise error
        return session

    def create_container_azure(self, session, container_name):
        """creates a container to Azure cloud
        Args :
            session           (obj)     --       Azure blob session object

            container_name    (str)     --       name of the new bucket which is to be created

        Returns :
            None

        Raises :
            Exception :
                if the container name is invalid or the container with specified name already exists

        """
        try:
            session.create_container(container_name)
        except Exception as error:
            self.log.error(
                "container name is invalid or a container already exists with that name")
            self.log.error(error)
            raise error

    def upload_file_azure(self,
                          session,
                          container_name,
                          blobname,
                          localfilename):
        """uploads a file to Azure cloud
        Args :
            session           (obj)    --       Azure blob session object

            container_name    (str)    --       name of the bucket to which the file has to be
            uploaded

            blobname          (str)    --       name of the file to be created on the cloud

            localfilename     (str)    --       name of the file in the current directory on the
            local machine

        Returns :
            None

        Raises :
            AzureMissingResourceHttpError :
                If the specified container or the file on the local machine doesn't exist

        """
        try:
            session.create_blob_from_path(
                container_name, blobname, localfilename)
        except azure.common.AzureMissingResourceHttpError as error:
            self.log.error(
                "The specified container or the file on the local machine doesn't exist")
            self.log.error(error)
            raise error

    def download_file_azure(self,
                            session,
                            container_name,
                            blobname,
                            localfilename):
        """downloads a file from Azure cloud
        Args :
            session                (obj)        --     Azure blob session object

            container_name         (str)        --     name of the azure blob container

            blobname               (str)        --     name of the blob that has to be downloaded

            localfilename          (str)        --     name of the file on local machine
            where the content has to be downloaded

        Returns :
            None

        Raises :
            AzureMissingResourceHttpError :
                If the specified container or the file does not exist on the cloud

        """
        try:
            session.get_blob_to_path(container_name, blobname, localfilename)
        except azure.common.AzureMissingResourceHttpError as error:
            self.log.error(
                "The specified container or the file does not exist on the cloud")
            self.log.error(error)
            raise error

    def delete_file_azure(self, session, container_name, blobname):
        """deletes a file from Azure cloud
        Args :
            session                (obj)        --     Azure blob session object

            container_name         (str)        --     name of the azure blob container

            blobname               (str)        --     name of the blob that has to be deleted

        Returns :
            None

        Raises :
            AzureMissingResourceHttpError :
                If the specified container or the file does not exist on the cloud

        """
        try:
            session.delete_blob(container_name, blobname)
        except azure.common.AzureMissingResourceHttpError as error:
            self.log.error(
                "The specified container or the file does not exist on the cloud")
            self.log.error(error)
            raise error

    def create_dir_download_file_azure(self, container_name, blobname, session):
        """checks if folder exists locally. If not, creates a directory and downloads a file to it
        Args :
            container_name (str)     --       name of the container

            blobname       (str)     --       name of the azure blob

            session        (obj)     --       azure session object

        Returns :
            None

        Raises :
            AzureMissingResourceHttpError :
                If the specified container or the file does not exist on the cloud

        """
        head, tail = os.path.split("{}".format(blobname))
        try:
            if self.machine.check_directory_exists(self.machine.join_path(os.getcwd(), head)):
                # download the files to this directory
                session.get_blob_to_path(container_name, blobname,
                                         self.machine.join_path(os.getcwd(), head, tail))
            else:
                # create the directory and download the file to it
                self.machine.create_directory(self.machine.join_path(os.getcwd(), head))
                session.get_blob_to_path(container_name, blobname,
                                         self.machine.join_path(os.getcwd(), head, tail))
        except azure.common.AzureMissingResourceHttpError as error:
            self.log.error(
                "The specified container does not exist on the cloud")
            self.log.error(error)
            raise error

    def delete_contents_azure(self, session, content):
        """deletes the specified contents from Azure cloud
        Args :
            session        (obj)     --       Azure blob session object

            content        (list)    --       Part of the subclient content which has to be
            deleted from the cloud

        Returns :
            None

        Raises :
            AzureMissingResourceHttpError :
                if the specified content does not exist on the cloud

        """
        for item in content:
            windows_path = Path(item)
            container_name = Path(item).parts[1]
            if (len(windows_path.parts)) == 2:
                session.delete_container(container_name)
                continue
            path_to_file = ("/".join(item.strip("/").split('/')[1:]))
            generator = session.list_blobs(container_name)
            blob_list = []
            for i in generator:
                blob_list.append(i.name)
            try:
                if path_to_file in blob_list:
                    session.delete_blob(container_name, path_to_file)
                else:
                    generator = session.list_blobs(
                        container_name, self.machine.join_path(path_to_file, '/'), delimiter='/')
                    for blob in generator:
                        session.delete_blob(container_name, blob.name)
            except azure.common.AzureMissingResourceHttpError as error:
                self.log.error(
                    "The specified contents does not exist on the cloud")
                self.log.error(error)
                raise error

    def delete_container_azure(self, session, container_name):
        """deletes a container from Azure cloud
        Args :
            session           (obj)     --       Azure blob session object

            container_name    (str)     --       name of the container which has to be deleted

        Returns :
            None

        Raises :
            AzureMissingResourceHttpError :
                if the specified container does not exist on the cloud

        """
        try:
            session.delete_container(container_name)
        except azure.common.AzureMissingResourceHttpError as error:
            self.log.error(
                "The specified container does not exist on the cloud")
            self.log.error(error)
            raise error

    def download_container_azure(self, session, container_name):
        """downloads a container from Azure cloud
        Args :
            session        (obj)   --   Azure blob session object

            container_name (str)   --   Name of the azure blob container which has to be downloaded

        Returns :
            None

        Raises :
            AzureMissingResourceHttpError :
                If the specified container does not exist on the cloud

        """
        generator = session.list_blobs(container_name)
        os.mkdir(container_name)
        try:
            os.chdir(container_name)
            # code below lists all the blobs in the container and
            # downloads them one after another
            for blob in generator:
                # check if the path contains a folder structure, create the folder structure
                if "/" in "{}".format(blob.name):
                    # extract the folder path and check if that folder exists locally,
                    # and if not create it
                    self.create_dir_download_file_azure(container_name, blob.name,
                                                        session)
                else:
                    try:
                        session.get_blob_to_path(
                            container_name, blob.name, blob.name)
                    except azure.common.AzureMissingResourceHttpError as error:
                        self.log.error(
                            "The specified container does not exist on the cloud")
                        self.log.error(error)
                        raise error
        finally:
            os.chdir(self.automation_directory)

    def download_contents_azure(self, session, content, dir_name, oop_flag):
        """downloads the specified contents from Azure cloud
        Args :
            session       (obj)      --      Azure blob session object

            content       (list)     --      Part of the subclient content which has to be
            downloaded from the cloud

            dir_name      (str)      --      Name of the folder to which the specified contents
            are to be downloaded

            oop_flag      (bool)     --      flag to determine if it's an in place restore
            or an out of place restore

        Returns :
            None

        Raises :
            AzureMissingResourceHttpError :
                If the specified content does not exist on the cloud

        """
        if oop_flag is True:
            container_name = content.replace("/", "")
            self.download_container_azure(session, container_name)
            os.chdir(str(self.automation_directory))
        else:
            os.mkdir(dir_name)
            for item in content:
                try:
                    os.chdir(dir_name)
                    path_to_file = ("/".join(item.strip("/").split('/')[1:]))
                    container_name = Path(item).parts[1]
                    generator = session.list_blobs(container_name)
                    blob_list = []
                    for j in generator:
                        blob_list.append(j.name)
                    if path_to_file == "":
                        self.download_container_azure(session, container_name)
                        os.chdir(self.automation_directory)
                    elif path_to_file in blob_list:
                        if os.path.exists(container_name) is False:
                            os.mkdir(container_name)
                        os.chdir(container_name)
                        self.create_dir_download_file_azure(container_name, path_to_file,
                                                            session)
                    else:
                        generator = session.list_blobs(container_name,
                                                       self.machine.join_path(path_to_file, '/'),
                                                       delimiter='/')
                        if os.path.exists(container_name) is False:
                            os.mkdir(container_name)
                        os.chdir(container_name)
                        # code below lists all the blobs in the container and downloads
                        # them one after another
                        for blob in generator:
                            # check if the path contains a folder structure,
                            # create the folder structure
                            if "/" in "{}".format(blob.name):
                                # extract the folder path and check if that folder exists locally,
                                # and if not create it
                                head, tail = os.path.split(
                                    "{}".format(blob.name))
                                if tail:
                                    self.create_dir_download_file_azure(container_name, blob.name,
                                                                        session)
                                else:
                                    self.recur(
                                        session, container_name, blob.name)
                            else:
                                try:
                                    session.get_blob_to_path(
                                        container_name, blob.name, blob.name)
                                except azure.common.AzureMissingResourceHttpError as error:
                                    self.log.error(
                                        "The specified contents does not exist on the cloud")
                                    self.log.error(error)
                                    raise error
                finally:
                    os.chdir(self.automation_directory)

    def recur(self, session, container_name, blobname):
        """recursively traverses through the subfolders and downloads the blobs
        Args :
            session        (obj)       --     Azure Storage session object

            container_name (str)       --     name of the azure container

            blobname       (str)       --     name of the azure blob

        Returns :
            None

        Raises :
            AzureMissingResourceHttpError :
                If the specified container or object does not exist on the cloud

        """
        generator = session.list_blobs(container_name, blobname, delimiter='/')
        for i in generator:
            head, tail = os.path.split("{}".format(i.name))
            if tail:
                self.create_dir_download_file_azure(
                    container_name, i.name, session)
            else:
                self.recur(session, container_name, i.name)
