# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
# See LICENSE.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""Helper file for performing Amazon S3 operations

S3Helper is the only class defined in this file.

S3Helper: Helper class to perform operations on Amazon S3

S3Helper :
    __init__()                           --   initializes Amazon S3 helper object

    create_session_s3()                  --   creates a new session with Amazon S3 cloud
    with the provided credentials

    create_container_s3()                --   creates a bucket in Amazon S3 cloud

    delete_container_s3()                --   deletes the specified bucket from Amazon S3 cloud

    upload_file_s3()                     --   uploads the specified file to Amazon S3 cloud

    create_dir_download_file_s3()        --   checks if the folder exists locally. If not,
    creates a directory and downloads a file to it

    delete_contents_s3()                 --   deletes all the contents under the specified
    paths from Amazon S3 cloud

    download_container_s3()              --   downloads a specified container with all
    it's contents from the Amazon S3 cloud

    download_contents_s3()               --   downloads all the contents under the
    specified paths from the Amazon S3 cloud

    recur()                              --   recursively traverses through the subfolders
    and downloads the objects

    delete_file_s3()                     --   deletes a specified file from Amazon S3 cloud

"""

import os
import socket
from pathlib import Path
from boto3.session import Session
from AutomationUtils import constants
from AutomationUtils.machine import Machine


class S3Helper(object):
    """Helper class to perform operations on Amazon S3"""

    def __init__(self, testcase_object):
        """Initializes Amazon S3 helper object"""
        self.automation_directory = constants.AUTOMATION_DIRECTORY
        self.log = testcase_object.log
        self.machine = Machine(socket.gethostname())

    def create_session_s3(self, access_key, secret_access_key, region):
        """creates a session with S3 cloud
        Args :
            access_key         (str)    --  Access key for connecting to Amazon S3 account

            secret_access_key  (str)    --  Secret access key for connecting to Amazon S3 account

            region             (str)    --  Region where the S3 buckets are residing

        Returns :
            (obj)                       --  Amazon S3 session object

        """
        session = Session(aws_access_key_id=access_key, aws_secret_access_key=secret_access_key,
                          region_name=region)
        s3_session = session.resource("s3")
        return s3_session

    def create_container_s3(self, session, bucket_name):
        """creates a container to S3 cloud
        Args :
            session        (obj)     --       Amazon S3 session object

            bucket_name    (str)     --       name of the new bucket which has to be created

        Returns :
            None

        Raises :
            Exception :
                if the bucket name is invalid or the bucket with specified name already exists

        """
        try:
            response = session.create_bucket(Bucket=bucket_name)
            self.log.info(response)
        except Exception as error:
            self.log.error(
                "The bucket name is invalid or the bucket with specified name already exists")
            self.log.error(error)
            raise error

    def delete_container_s3(self, session, bucket_name):
        """deletes a container from S3 cloud
        Args :
            session        (obj)     --       Amazon S3 session object

            bucket_name    (str)     --       name of the bucket which has to be deleted

        Returns :
            None

        Raises :
            Exception :
                If the specified bucket does not exist

        """
        bucket = session.Bucket(bucket_name)
        try:
            for key in bucket.objects.all():
                key.delete()
            bucket.delete()
        except Exception as error:
            self.log.error("The specified bucket doesn't exist")
            self.log.error(error)
            raise error

    def upload_file_s3(self, session, file_path, bucket_name, file_name):
        """uploads a file to S3 cloud
        Args :
            session        (obj)     --       Amazon S3 session object

            file_path      (str)     --       path to the file on the local machine

            bucket_name    (str)     --       name of the bucket

            file_name      (str)     --       name of the file to be created on the cloud

        Returns :
            None

        Raises :
            Exception :
                If the specified bucket or the file on the local machine doesn't exist

        """
        try:
            response = session.meta.client.upload_file(file_path, bucket_name, file_name)
            self.log.info(response)
        except Exception as error:
            self.log.error("The specified bucket or the file on the local machine doesn't exist")
            self.log.error(error)
            raise error

    def create_dir_download_file_s3(self, container_name, obj_key, session):
        """checks if folder exists locally. If not, creates a directory and downloads a file to it
        Args :
            container_name (str)     --       name of the container

            obj_key        (str)     --       name of the s3 object

            session        (obj)     --       s3 session object

        Returns :
            None

        Raises :
            Exception:
                if the specified content does not exist on the cloud

        """
        head, tail = os.path.split("{}".format(obj_key))
        try:
            if self.machine.check_directory_exists(self.machine.join_path(os.getcwd(), head)):
                # download the files to this directory
                session.meta.client.download_file(container_name, obj_key,
                                                  self.machine.join_path(os.getcwd(), head, tail))
            else:
                # create the directory and download the file to it
                self.machine.create_directory(self.machine.join_path(os.getcwd(), head))
                session.meta.client.download_file(container_name, obj_key,
                                                  self.machine.join_path(os.getcwd(), head, tail))
        except Exception as error:
            self.log.error("The specified container does not exist on the cloud")
            self.log.error(error)
            raise error

    def delete_contents_s3(self, session, content):
        """deletes the specified contents from S3 cloud
        Args :
            session        (obj)     --       Amazon S3 session object

            content        (list)    --       Part of the subclient content which has to be
            deleted from the cloud

        Returns :
            None

        Raises :
            Exception :
                if the specified content does not exist on the cloud

        """
        for item in content:
            windows_path = Path(item)
            bucket_name = Path(item).parts[1]
            if (len(windows_path.parts)) == 2:
                self.delete_container_s3(session, bucket_name)
                continue
            last = os.path.basename(item)
            bucket = session.Bucket(bucket_name)
            path_to_file = ("/".join(item.strip("/").split('/')[1:]))
            generator = bucket.objects.all()
            obj_list = []
            for i in generator:
                obj_list.append(i.key)
            if last in obj_list:
                obj = session.Object(bucket_name, path_to_file)
                obj.delete()
            else:
                generator = bucket.objects.filter(Prefix=path_to_file)
                for j in generator:
                    obj = session.Object(bucket_name, j.key)
                    obj.delete()

    def download_container_s3(self, session, container_name):
        """downloads a container from S3 cloud
        Args :
            session        (obj)   --   Amazon S3 session object

            container_name (str)   --   Name of the amazon s3 bucket which has to be downloaded

        Returns :
            None

        Raises :
            Exception :
                If the specified container does not exist on the cloud

        """
        bucket = session.Bucket(container_name)
        generator = bucket.objects.all()
        os.mkdir(container_name)
        try:
            os.chdir(container_name)
            # code below lists all the blobs in the container and downloads them one after another
            for obj in generator:
                # check if the path contains a folder structure, create the folder structure
                if "/" in "{}".format(obj.key):
                    # extract the folder path and check if that folder exists locally,
                    # and if not create it
                    self.create_dir_download_file_s3(container_name, obj.key, session)
                else:
                    try:
                        session.meta.client.download_file(container_name, obj.key, obj.key)
                    except Exception as error:
                        self.log.error("The specified container does not exist on the cloud")
                        self.log.error(error)
                        raise error
        finally:
            os.chdir(self.automation_directory)

    def download_contents_s3(self, session, content, dir_name, oop_flag):
        """downloads the specified contents from S3 cloud
        Args :
            session       (obj)      --      Amazon S3 session object

            content       (list)     --      Part of the subclient content which has to
            be downloaded from the cloud

            dir_name      (str)      --      Name of the folder where the specified
            contents are to be downloaded

            oop_flag      (bool)     --      Flag to determine if it's an inplace restore
            or an out of place restore

        Returns :
            None

        Raises :
            Exception :
                If the specified content does not exist on the cloud

        """
        if oop_flag is True:
            container_name = content.replace("/", "")
            self.download_container_s3(session, container_name)
            os.chdir(self.automation_directory)
        else:
            os.mkdir(dir_name)
            for item in content:
                try:
                    os.chdir(dir_name)
                    container_name = Path(item).parts[1]
                    path_to_file = ("/".join(item.strip("/").split('/')[1:]))
                    bucket = session.Bucket(container_name)
                    generator = bucket.objects.all()
                    obj_list = []
                    for j in generator:
                        obj_list.append(j.key)
                    if path_to_file == "":
                        self.download_container_s3(session, container_name)
                        os.chdir(self.automation_directory)
                    elif path_to_file in obj_list:
                        if os.path.exists(container_name) is False:
                            os.mkdir(container_name)
                        os.chdir(container_name)
                        self.create_dir_download_file_s3(container_name, path_to_file,
                                                         session)
                    else:
                        generator_folder = bucket.objects.filter(Prefix=path_to_file)
                        if os.path.exists(container_name) is False:
                            os.mkdir(container_name)
                        os.chdir(container_name)
                        # code below lists all the blobs in the container and
                        # downloads them one after another
                        for obj in generator_folder:
                            # check if the path contains a folder structure,
                            # create the folder structure
                            if "/" in "{}".format(obj.key):
                                # extract the folder path and check if that folder exists locally,
                                # and if not create it
                                head, tail = os.path.split("{}".format(obj.key))
                                if tail:
                                    self.create_dir_download_file_s3(container_name, obj.key,
                                                                     session)
                                else:
                                    self.recur(session, container_name, obj.key)
                            else:
                                try:
                                    session.meta.client.download_file(container_name, obj.key,
                                                                      obj.key)
                                except Exception as error:
                                    self.log.error(
                                        "The specified content does not exist on the cloud")
                                    self.log.error(error)
                                    raise error
                finally:
                    os.chdir(self.automation_directory)

    def recur(self, session, container_name, objct):
        """recursively traverses through the subfolders and downloads the objects
        Args :
            session        (obj)       --     S3 session object

            container_name (str)       --     name of the S3 container

            objct          (str)       --     name of the object

        Returns :
            None

        Raises :
            Exception :
                If the specified container or object does not exist on the cloud

        """
        bucket = session.Bucket(container_name)
        generator_folder = bucket.objects.filter(Prefix=objct)
        for i in generator_folder:
            head, tail = os.path.split("{}".format(i.key))
            if tail:
                self.create_dir_download_file_s3(container_name, i.key, session)
            else:
                self.recur(session, container_name, i.key)

    def delete_file_s3(self, session, bucket_name, filetobedeleted):
        """deletes a file from S3 cloud
        Args :
            session            (obj)       --        Amazon S3 session object

            bucket_name        (str)       --        Name of the bucket

            filetobedeleted    (str)       --        Name of the file to be deleted

        Returns :
            None

        Raises :
            Exception :
                If the specified bucket or the file does not exist on the cloud

        """
        try:
            response = session.meta.client.delete_object(Bucket=bucket_name, Key=filetobedeleted)
            self.log.info(response)
        except Exception as error:
            self.log.error("The specified bucket or the file does not exist on the cloud")
            self.log.error(error)
            raise error
