# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
# See LICENSE.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
    Module for invoking DB queries

ExchangeCSDBHelper:
    Class for performing all operations on CSDB pertaining to Exchange Mailbox Agent


ExchangeCSDBHelper
=======

    __init__(exmbclient_object)             --  initialize object of ExchangeCSDBHelper class associated with
    the Exchange Mailbox

    get_backup_time_size_from_csdb()        --  Method to get the last played backup time and size

    get_public_folder_guid()                --  Get the GUID assigned to a public folder

    get_mailbox_assoc_policy()              -- Get the Association Policies assigned to a mailbox associated with
                                                the Exchange Client

	get_assoc_mailbox_count()				--	Get the number of amilboxes associated with a client

	get_mailbox_guid()                      --  Get the GUIDs of the mailboxes

"""

from __future__ import unicode_literals
from collections import OrderedDict
from xmltodict import parse
from AutomationUtils.options_selector import OptionsSelector
from .operations import CvOperation
from . import constants


class ExchangeCSDBHelper:
    """Class for initializing input variables and object creation from different modules"""

    def __init__(self, ex_object):
        """Initializes the input variables,logging and creates object from other modules.

                Args:
                    tc_object   --  instance of testcase class

                Returns:
                    object  --  instance of ExchangeCSDBHelper class"""
        self.tc_object = ex_object.tc_object
        self._utility = OptionsSelector(self.tc_object.commcell)
        self.log = self.tc_object.log
        self.log.info('logger initialized for Exchange DB Helper')
        self.csdb = self.tc_object.csdb

    def get_backup_time_size_from_csdb(self):
        """
            Method to get the lastBackupTime and the totalBackupSize of an
            Excahnge Client's Backupset from CSDB

            Arguments:
                 None
            Returns:
                prop_dict       (dict)--    Dictionary of retrieved properties
            Example:
                  {
                    'lastBackupTime': value,
                    'totalBackupSize' : value
                    }
        """
        query = constants.GET_BACKUP_PROP % self.tc_object.backupset.backupset_id
        query_result = self._utility.exec_commserv_query(query=query)
        self.tc_object.log.info(query_result)
        query_result = OrderedDict(
            parse(
                query_result[0][0],
                process_namespaces=True
            )
        )
        res_dict = dict()
        res_dict['applicationSize'] = query_result['Indexing_DbStats']['apps']['stat']['@applicationSize']
        res_dict['lastPlayedBackupTime'] = query_result['Indexing_DbStats']['@lastPlayedBkpTime']
        return res_dict

    def get_public_folder_guid(self):
        """
            Method to get the GUID associated with the 'All Public Folders'
            association in the CSDB

            Arguments:
                None

            Returns:
                public_folder_guid      (str)--     The GUID associated with the Public Folder
        """

        query = constants.GET_PUBLIC_FOLDER_GUID % self.tc_object.client.client_id
        query_result = self._utility.exec_commserv_query(query=query)
        public_folder_guid = query_result[0][0]
        return public_folder_guid

    def get_mailbox_assoc_policy(self, mailbox_smtp_address):
        """
            Method to get the policy IDs associated with the mailbox in the CSDB.

            Arguments:
                mailbox_smtp_address    (str)-- The SMTP Address of the mailbox

            Returns:
                policies                (dict)-- Dictionary of Archive, Retention and Config Policy
        """
        query = constants.GET_MAILBOX_ASSOC_POLICY % (
            mailbox_smtp_address, self.tc_object.client.client_id)
        query_result = self._utility.exec_commserv_query(query=query)
        self.tc_object.log.info(query_result)
        polic_dict = dict()
        for policyId, policyVal in query_result[1]:
            if policyId == '1':
                polic_dict['archive_policy'] = policyVal
            elif policyId == '2':
                polic_dict['cleanup_policy'] = policyVal
            elif policyId == '3':
                polic_dict['retention_policy'] = policyVal
        return polic_dict

    def check_deleted_flag_set(self, mailbox_smtp_address):
        """
            Checks whether the deleted flag is set for a mailbox in CSDB or not

            Arguments:
                mailbox_smtp_address        (str)--     The SMTP Address of the mailbox

            Returns
                deleted_flag_is_set         (Bool)--    True if the deleted flag is set
        """
        query = constants.CHECK_MAILBOX_DELETED_FLAG % (
            self.tc_object.client.client_id, mailbox_smtp_address)
        self.csdb.execute(query)
        row = self.csdb.fetch_one_row()
        return False if row == [''] else True

    def get_assoc_mailbox_count(self):
        """
            Method to get a count f the number of mailboxes associated with the Client
            Client ID is taken from TC_Object-> Client-> Client ID

            Arguments:
                None

            Returns:
                associated_mailbox_count        (int)--     Number of mailboxes associated with the client

        """

        query = constants.GET_ASSOC_MBX_COUNT % self.tc_object.client.client_id
        user_count_sql_server = self._utility.exec_commserv_query(query)
        assoc_count = int(user_count_sql_server[0][0])
        return assoc_count

    def get_mailbox_guid(self, mailbox_list):
        """
            Method to get the GUIDs for the list of mailboxes

            Arguments:
                mailbox_list        (list)--    List of mailboxes to get the GUID of

            Returns:
                mailbox_guid_dict   (dict)--    Dictionary of mailbox GUIDs
                Format:
                    mailbox_alias is the key
                    mailbox_guid would be the value

        """
        mailbox_guid_dict = dict()

        for mailbox in mailbox_list:
            _query = (
                "select userguid from APP_EmailConfigPolicyAssoc where "
                "subClientId = '%s'and aliasName = '%s'" %
                (self.tc_object.subclient.subclient_id, mailbox))

            self.csdb.execute(_query)
            mailbox_guid = self.csdb.fetch_one_row()
            mailbox_guid = mailbox_guid[0]

            mailbox_guid_dict[mailbox.lower()] = mailbox_guid

        return mailbox_guid_dict
