"""Main module for generating data for testcases"""

from AutomationUtils import config, constants
from .sharepointconstants import TEST_DATA_GENERATION_JSON
from .sharepointconstants import LIST_VERSIONS_UNWANTED_PROPERTIES, FILE_VERSIONS_UNWANTED_PROPERTIES
import os
import random
import string


class TestData:
    """Class to create testdata for SharePoint testcases"""

    def __init__(self, sp_object):
        """Initializes the test data object

            Args:

                sp_object (object)     --  instance of the SharePoint object

        """

        self.sp_object = sp_object
        self.testcase_id = str(sp_object.tc_object.id)
        self.log = self.sp_object.log
        self.json_value = self.read_data_from_json()
        self.share_point_data_flag = None
        try:
            self.json_value = getattr(self.json_value, ("t" + self.testcase_id))
        except Exception:
            self.json_value = None
            self.log.exception(f"Test data is not found for the test case {self.testcase_id}")
        self.testdata_metadata = {}

    @staticmethod
    def read_data_from_json():
        """Method to read data from JSON file

            Returns:

                sheet_number (int)  --  sheet number to read the test data

        """

        return config.get_config(
            reset=True, json_path=TEST_DATA_GENERATION_JSON)

    @staticmethod
    def generate_file(file_name, size=None):
        """Generates the file and populates data

            Args:

                file_name (str)      --   name of the file

                size (int)           --   size of file in bytes

        """
        if os.path.exists(file_name):
            os.remove(file_name)
        if not size:
            size = random.randint(50, 100)
        chars = ''.join([random.choice(string.ascii_letters) for i in range(size)]) + "\n"
        with open(file_name, 'w') as f:
            f.write(chars)
        return file_name

    @staticmethod
    def delete_file(file_name):
        """Deletes the specified file

            Args:

                file_name (str)      --      name of the file

        """
        if os.path.exists(file_name):
            os.remove(file_name)
        else:
            raise Exception("The file does not exist")

    @staticmethod
    def process_metadata(dictionary, list_item_versions=False, file_versions=False):
        """Returns only required fields in metadata and  removes all unwanted fields

            Args:

                dictionary (dict)             --      metadata dictionary

                list_item_versions (boolean)  --      whether it is a list item metadata with versions

        """
        processed_dictionary = {key: dictionary.get(key) for key in dictionary
                                if not isinstance(dictionary.get(key), dict)}
        if list_item_versions:
            return {x: processed_dictionary[x] for x in processed_dictionary if
                    x not in LIST_VERSIONS_UNWANTED_PROPERTIES}
        elif file_versions:
            return {x: processed_dictionary[x] for x in processed_dictionary if
                    x not in FILE_VERSIONS_UNWANTED_PROPERTIES}
        else:
            return processed_dictionary

    @staticmethod
    def update_file(file_name):
        """Updates the specified file

            Args:

                file_name (str)      --      name of the file

        """
        file = open(file_name, "a")
        size = random.randint(50, 100)
        chars = ''.join([random.choice(string.ascii_letters) for i in range(size)]) + "\n"
        file.write(chars)
        file.close()

    def create_list(self):
        """Creates SharePoint list"""
        try:
            self.log.info("Creating list")
            list = self.json_value.List
            list_title = list.TITLE
            if list.OP_TYPE.lower() == "create":
                self.log.info("Deleting the list if it exists")
                if self.sp_object.get_sp_list_metadata(list_title):
                    self.log.info("List exists")
                    self.sp_object.delete_sp_list(list_title)
                    self.log.info("Deleted the existing list")
                request_body = {
                    "AllowContentTypes": True,
                    "BaseTemplate": 100,
                    "ContentTypesEnabled": True,
                    "Description": "This is a list created for test automation",
                    "EnableVersioning": True,
                    "Title": list_title
                }
                list_metadata = self.sp_object.create_sp_list(request_body)
                self.testdata_metadata[self.sp_object.site_url]["List"] = {
                    "List Metadata": self.process_metadata(list_metadata),
                    "Title": list_title,
                    "Columns": [],
                    "List Items": {}
                }
                self.log.info("{0} list created successfully".format(self.testdata_metadata.get(
                    self.sp_object.site_url, {}).get("List", "").get("Title", "")))
        except Exception as exception:
            self.log.exception("Exception while creating list: %s", exception)
            raise exception

    def add_list_column(self, list_guid, title, field_id):
        """Adds column to SharePoint list

            Args:

                list_guid (str)      --      guid of the list

                title (str)          --      title of the column

                field_id (int)       --      field id of column i.e., type of the column

        """
        try:
            request_body = {
                "__metadata": {
                    "type": "SP.Field"
                },
                "Title": title,
                "FieldTypeKind": field_id
            }
            return self.sp_object.create_list_custom_column(list_guid, request_body)
        except Exception as exception:
            self.log.exception("Exception while adding list column: %s", exception)
            raise exception

    def create_custom_columns_in_list(self):
        """Creates custom columns for list"""
        try:
            self.log.info("Creating Custom Columns")
            list_columns = self.json_value.List.COLUMNS
            list_guid = self.testdata_metadata.get(self.sp_object.site_url, {}).get("List", {}).get("List Metadata",
                                                                                                    {}).get("Id", "")
            for column in list_columns:
                result = self.add_list_column(list_guid, column.Title, column.FieldTypeKind)
                self.testdata_metadata[self.sp_object.site_url]["List"]["Columns"].append(
                    [result.get("StaticName"), result.get("Title"), result.get("FieldTypeKind")])
            self.log.info("Custom columns are created successfully")
        except Exception as exception:
            self.log.exception("Exception while adding columns to the list: %s", exception)
            raise exception

    @staticmethod
    def generate_test_value_for_list_item_column(field_type_kind):
        """Generates test value for the list item based on column type

            Args:

                field_type_kind (int)       --      field id of column i.e., type of the column

        """
        size = random.randint(50, 100)
        if field_type_kind == 1:
            return size
        elif field_type_kind == 2:
            return ''.join([random.choice(string.ascii_letters) for i in range(size)])
        elif field_type_kind == 3:
            return ''.join([random.choice(string.ascii_letters) for i in range(size)]) + "\n" + \
                   ''.join([random.choice(string.ascii_letters) for i in range(size)])

    def update_list_item_metadata(self, list_title, list_item_id, oop_site=None):
        """Updates list item metadata dict

            Args:

                list_title (str)          --    title of the list

                list_item_id (int)        --    id of the list item

                oop_site (str)            --    out of place restore site

        """
        try:
            if oop_site:
                source_site = self.sp_object.site_url
                self.sp_object.site_url = oop_site
            list_item_metadata = self.process_metadata(
                self.sp_object.get_sp_list_item_metadata(list_title, str(list_item_id)))
            list_item_metadata["Versions Metadata"] = [self.process_metadata(i, list_item_versions=True) for i in
                                                       self.sp_object.get_sp_list_item_metadata(
                                                           list_title, str(list_item_id), "Versions").get("results", [])]
            if oop_site:
                self.sp_object.site_url = source_site
            if list_item_id in self.testdata_metadata[self.sp_object.site_url]["List"]["List Items"].keys():
                self.testdata_metadata[self.sp_object.site_url]["List"]["List Items"][list_item_id].append(
                    list_item_metadata)
            else:
                self.testdata_metadata[self.sp_object.site_url]["List"]["List Items"][list_item_metadata.get("Id")] = [
                    list_item_metadata]
            self.log.info("Fetched updated list item metadata")
        except Exception as exception:
            self.log.exception("Exception while updating list item metadata: %s", exception)
            raise exception

    def update_list_item(self, list_title, list_item_title, list_item_id=None):
        """Updates/Creates list item of the list

             Args:

                list_title (str)          --    title of the list

                list_item_title (str)     --    title of the list item

                list_item_id (int)        --    id of the list item

        """
        try:
            request_body = {
                "Title": list_item_title
            }
            columns = self.testdata_metadata.get(self.sp_object.site_url, {}).get("List").get("Columns")
            for column in columns:
                request_body[column[0]] = self.generate_test_value_for_list_item_column(column[2])
            if list_item_id:
                self.sp_object.update_sp_list_item(list_title, request_body, str(list_item_id))
                self.log.info(f"List Item with id {list_item_id} is updated successfully")
            else:
                result = self.process_metadata(self.sp_object.create_sp_list_item(list_title, request_body))
                list_item_id = result.get("Id")
                self.log.info(f"List Item with id {list_item_id} is created successfully")
            return list_item_id
        except Exception as exception:
            self.log.exception("Exception while adding list items in the list: %s", exception)
            raise exception

    def add_list_item(self, list_title, list_item, versions=True):
        """Adds a list item

             Args:

                list_title (str)          --    title of the list

                list_item (config)        --    config object of list item

                versions (boolean)        --    whether all versions enabled or not

        """
        try:
            list_id = self.update_list_item(list_title, list_item.Title)
            if versions:
                version_num = random.randint(2, 4)
                self.log.info(f"Creating {version_num} additional versions")
                for i in range(version_num):
                    self.update_list_item(list_title, list_item.Title, list_id)
            self.update_list_item_metadata(list_title, list_id)
        except Exception as exception:
            self.log.exception("Exception while adding list item in the list: %s", exception)
            raise exception

    def add_list_items(self, create_time, versions=True):
        """Adds list items to the list

            Args:

                create_time (str)         --    when to create the list item in the testcase

                versions (boolean)        --    whether all versions enabled or not

        """
        try:
            self.log.info("Adding list items with versions")
            list_items = self.json_value.List.LIST_ITEMS
            list_title = self.json_value.List.TITLE
            for list_item in list_items:
                if create_time == list_item.CREATE_TIME:
                    self.add_list_item(list_title, list_item, versions)
            self.log.info("List items are created successfully")
        except Exception as exception:
            self.log.exception("Exception while adding list items in the list: %s", exception)
            raise exception

    def update_list_items(self, versions=True):
        """Updates list items in the list

            Args:

                versions (boolean)        --    whether all versions enabled or not

        """
        try:
            self.log.info("Updating list items in list")
            list_items = self.json_value.List.LIST_ITEMS
            list_title = self.json_value.List.TITLE
            index = 1
            for list_item in list_items:
                list_item_id = index
                if list_item.OP_TYPE == "DELETE":
                    self.sp_object.delete_sp_list_item(list_title, list_item_id)
                    self.log.info(f"Deleted list item with id {list_item_id}")
                elif list_item.OP_TYPE == "EDIT":
                    self.update_list_item(list_title, list_item.Title, list_item_id)
                    self.update_list_item_metadata(list_title, list_item_id)
                elif list_item.OP_TYPE == "CREATE":
                    self.add_list_item(list_title, list_item, versions)
                index = index + 1
        except Exception as exception:
            self.log.exception("Exception while updating list: %s", exception)
            raise exception

    def create_folder(self):
        """Creates folder in a library"""
        try:
            self.log.info("Creating a folder")
            folder = self.json_value.Folder
            folder_name = folder.NAME
            site_name = "/".join(self.sp_object.site_url.split("/")[4:])
            folder_url = "/sites/" + site_name + "/shared documents/" + folder_name
            self.log.info("Delete the folder if exists")
            if self.sp_object.get_sp_folder_metadata(folder_url):
                self.log.info(f"{folder_name} folder exists")
                self.sp_object.delete_sp_file_or_folder(folder_url)
                self.log.info("Deleted the existing folder")
            request_body = {
                "ServerRelativeUrl": folder_url
            }
            self.sp_object.create_sp_folder(request_body)
            self.log.info("{0} folder created successfully".format(folder_url))
            self.testdata_metadata[self.sp_object.site_url]["Folder"] = {
                "Folder Name": folder_name,
                "Folder URL": folder_url
            }
            self.log.info("Created folder successfully")
        except Exception as exception:
            self.log.exception("Exception while creating a folder: %s", exception)
            raise exception

    def update_file_metadata(self, file_name, folder_path, oop_site=None):
        """Updates file metadata

            Args:

                file_name (str)        --      name of the file

                folder_path (str)      --      path of the folder

                oop_site (str)         --      out of place restore site

        """
        try:
            if oop_site:
                source_site = self.sp_object.site_url
                self.sp_object.site_url = oop_site
            file_metadata = self.process_metadata(self.sp_object.get_sp_file_metadata(file_name, folder_path))
            file_metadata["Versions Metadata"] = [self.process_metadata(i, file_versions=True) for i in
                                                  self.sp_object.get_sp_file_metadata(file_name, folder_path,
                                                                                      "Versions").get("results", [])]
            if oop_site:
                self.sp_object.site_url = source_site
            if file_name in self.testdata_metadata.get(self.sp_object.site_url, {}).get("Folder", {}).keys():
                self.testdata_metadata[self.sp_object.site_url]["Folder"][file_name].append(file_metadata)
            else:
                self.testdata_metadata[self.sp_object.site_url]["Folder"][file_name] = [file_metadata]
            self.log.info("Fetched updated file metadata")
        except Exception as exception:
            self.log.exception("Exception while getting updated file metadata: %s", exception)
            raise exception

    def update_sp_file(self, file_name, folder_path):
        """ Updates the provided file

            Args:

                file_name (str)        --      name of the file

                folder_path (str)      --      path of the folder

        """
        try:
            self.update_file(file_name)
            self.sp_object.upload_sp_binary_file(folder_path, file_name)
            self.log.info("{0} file updated successfully".format(file_name))
        except Exception as exception:
            self.log.exception("Exception while updating file: %s", exception)
            raise exception

    def create_file(self, file, folder_path, versions=True, large_files_size=None):
        """Creates file in a given library or folder

            Args:

                file (config object)         --     config object for file

                folder_path (str)            --     path of the folder

                versions (boolean)           --     whether all versions enabled or not

                large_files_size (int)       --     size of files in bytes

        """
        try:
            file_name = file.Name
            self.generate_file(file_name, large_files_size)
            if versions:
                version_num = random.randint(3, 4)
            else:
                version_num = 1
            self.log.info(f"Creating/Adding {version_num} versions to the file")
            for i in range(version_num):
                self.update_sp_file(file_name, folder_path)
            self.update_file_metadata(file_name=file_name, folder_path=folder_path)
            self.log.info("File with versions is uploaded in folder")
        except Exception as exception:
            self.log.exception("Exception while creating a folder: %s", exception)
            raise exception

    def create_files(self, create_time, versions=True, large_files_size=None):
        """Creates files in given library or folder

            Args:

                create_time (str)         --    when to create the file in the testcase

                versions (boolean)        --    whether versions enabled or not

                large_files_size (int)     --     size of files in bytes

        """
        try:
            self.log.info("Creating files in folder")
            files = self.json_value.Folder.FILES
            folder_name = self.json_value.Folder.NAME
            site_name = "/".join(self.sp_object.site_url.split("/")[4:])
            folder_url = "/sites/" + site_name + "/shared documents/" + folder_name
            for file in files:
                if create_time == file.CREATE_TIME:
                    self.create_file(file, folder_url, versions, large_files_size)
            self.log.info("Created files inside folder")
        except Exception as exception:
            self.log.exception("Exception while creating a folder: %s", exception)
            raise exception

    def update_files(self, versions=True):
        """Updates files in the folder

            Args:

                versions (boolean)           --     whether versions enabled or not

        """
        try:
            self.log.info("Updating files in folder")
            files = self.json_value.Folder.FILES
            folder_name = self.json_value.Folder.NAME
            site_name = "/".join(self.sp_object.site_url.split("/")[4:])
            folder_url = "/sites/" + site_name + "/shared documents/" + folder_name
            for file in files:
                if file.OP_TYPE == "DELETE":
                    file_relative_url = folder_url + "/" + file.Name
                    self.sp_object.delete_sp_file_or_folder(file_relative_url)
                    self.log.info(f"{file_relative_url} is deleted successfully")
                elif file.OP_TYPE == "EDIT":
                    self.create_file(file, folder_url, versions)
                elif file.OP_TYPE == "CREATE":
                    self.create_file(file, folder_url, versions)
        except Exception as exception:
            self.log.exception("Exception while upadting files: %s", exception)
            raise exception

    def create_site_structure_for_backup(self, site_url_list, folder=None, list=None, versions=True, large_files_size=None):
        """Creates the sample site structure for backup
            1. Create a folder under shared documents library
                - Create a folder
                - Upload a file with some random content
            2. Create a list

            Args:

                site_url_list (list)       --    list of sites in which testdata should be generated

                folder (boolean)           --    whether create test data for folder

                list (boolean)             --    whether create test data for list

                versions (boolean)         --    whether all versions enabled or not

                large_files_size (int)     --     size of files in bytes

        """
        try:
            for site_url in site_url_list:
                self.sp_object.site_url = site_url
                site_metadata = self.sp_object.get_site_properties(root_site=True)
                self.testdata_metadata[site_url] = {
                    'Title': site_metadata.get('Title', "")
                    }
                if folder:
                    self.create_folder()
                    self.create_files(create_time="create before backup", versions=versions,
                                      large_files_size=large_files_size)
                if list:
                    self.create_list()
                    self.create_custom_columns_in_list()
                    self.add_list_items(create_time="create before backup", versions=versions)
            self.share_point_data_flag = True
        except Exception as exception:
            self.log.exception("Exception while creating site structure on SharePoint Site: %s", str(exception))
            raise exception

    def create_initial_subsites(self, num_of_sites=2):
        """Creates subsite in a given site based on given input json
        """
        try:
            site_url_list = [self.sp_object.site_url]
            subsite_end_url_list = []
            subsite_title_list = []
            subsite_list = []
            for i in range(1, num_of_sites + 1):
                title = "Test Subsite - " + str(i)
                url = "subsite_" + str(i)
                subsite_list.append({
                    "Title": "Test Subsite - " + str(i),
                    "Url End": "subsite_" + str(i)
                })
                site_url_list.append(self.sp_object.site_url + "/" + url)
                subsite_end_url_list.append(url)
                subsite_title_list.append(title)
            self.sp_object.create_subsites(subsite_list)
            self.log.info(f"Sites used in this test case: {site_url_list}")
            return site_url_list, subsite_end_url_list, subsite_title_list
        except Exception as exception:
            self.log.exception("Exception while creating subsite on SharePoint Site: %s", str(exception))
            raise exception

    def modify_backup_content(self, folder=None, list=None, site_url=None):
        """Modifies SharePoint backup content

            Args:

                folder (boolean)           --    whether modify test data for folder

                list (boolean)             --    whether modify test data for list

                site_url (str)             --   url of site in which content is modified

        """
        try:
            if site_url:
                source_site = self.sp_object.site_url
                self.sp_object.site_url = site_url
            self.log.info(f"Modifying content for {self.sp_object.site_url}")
            if folder:
                self.update_files()
                self.log.info("Updated files inside folder")
            if list:
                self.update_list_items()
                self.log.info("List items are updated successfully")
            if site_url:
                self.sp_object.site_url = source_site
        except Exception as exception:
            self.log.exception("Exception while modifying backup content: %s", str(exception))
            raise exception

    def modify_content_before_restore(self, folder=None, list=None, site_url=None):
        """Modifies the content on SharePoint site

            Args:

                folder (boolean)           --    whether modify test data for folder

                list (boolean)             --    whether modify test data for list

                site_url (str)             --   url of site in which content is modified

        """
        try:
            if site_url:
                self.sp_object.site_url = site_url
            self.log.info(f"Modifying content for {self.sp_object.site_url} before restore")
            if folder:
                self.log.info("Adding versions to the file in folder")
                files = self.json_value.Folder.FILES
                folder_name = self.json_value.Folder.NAME
                site_name = "/".join(self.sp_object.site_url.split("/")[4:])
                folder_url = "/sites/" + site_name + "/shared documents/" + folder_name
                for file in files:
                    if file.OP_TYPE != "DELETE":
                        self.update_sp_file(file.Name, folder_url)
                self.log.info("Added versions all the files")
            if list:
                self.log.info("Editing list items")
                list_items = self.json_value.List.LIST_ITEMS
                list_title = self.json_value.List.TITLE
                index = 1
                for list_item in list_items:
                    list_item_id = index
                    if list_item.OP_TYPE != "DELETE":
                        self.update_list_item(list_title, list_item.Title, list_item_id)
                    index = index + 1
                self.log.info("Added versions to all list items")
        except Exception as exception:
            self.log.exception("Exception while modifying content on SharePoint site: %s", str(exception))
            raise exception

    def delete_backup_content(self, folder=None, list=None):
        """Deletes the backed up content

            Args:

                folder (boolean)           --    whether delete test data of folder

                list (boolean)             --    whether delete test data of list

        """
        try:
            if folder:
                self.log.info("Deleting folder from sharepoint site")
                folder_name = self.json_value.Folder.NAME
                site_name = "/".join(self.sp_object.site_url.split("/")[4:])
                folder_url = "/sites/" + site_name + "/shared documents/" + folder_name
                self.sp_object.delete_sp_file_or_folder(folder_url)
                self.log.info("{0} is deleted from sharepoint site".format(folder_url))
            if list:
                self.log.info("Deleting list from sharepoint site")
                list = self.json_value.List
                list_title = list.TITLE
                self.sp_object.delete_sp_list(list_title)
                self.log.info("List {0} is deleted successfully from sharepoint site".format(list_title))
            self.log.info("Deleted backup content successfully")
            self.share_point_data_flag = False
        except Exception as exception:
            self.log.exception("Exception while deleting backup content: %s", str(exception))
            raise exception

    def delete_disk_files(self):
        """Deletes that are created on system to upload into SharePoint"""
        try:
            self.log.info("Deleting files on disk")
            files = self.json_value.Folder.FILES
            for file in files:
                self.delete_file(file.Name)
                self.log.info(f"Deleted file {file.Name}")
        except Exception as exception:
            self.log.exception("Exception while deleting files on disk: %s", exception)

    def delete_backup_site_structure(self, site_url_list=None, folder=None, list=None):
        """Deletes the SharePoint site structure created for running testcase

            Args:

                site_url_list (str)        --   list of sites

                folder (boolean)           --    whether delete test data of folder

                list (boolean)             --    whether delete test data of list
        """
        try:
            if site_url_list is None:
                if self.sp_object.site_url_list:
                    site_url_list = self.sp_object.site_url_list
                else:
                    site_url_list = [self.sp_object.site_url]
            if self.share_point_data_flag and self.sp_object.tc_object.status != constants.FAILED:
                for site in site_url_list:
                    self.sp_object.site_url = site
                    self.delete_backup_content(folder, list)
                    self.log.info("SharePoint site structure created for running testcase is deleted")
            else:
                self.log.info("Testcase got failed in between or Test data is already deleted on SharePoint Site")
        except Exception as exception:
            self.log.exception("Exception while deleting backup site structure: %s", exception)
            raise exception
