# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
# See LICENSE.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
main file for validating all the options for restores

Class:
Restore - class defined for validating  all restores options

"""
from deepdiff import DeepDiff
from AutomationUtils.machine import Machine
from Application.Sharepoint import sharepointconstants as constants


class Restore(object):
    """Class for performing Restore related operations."""

    def __init__(self, sp_object):
        """Initializes the Restore object

            Args:

                sp_object (Object)  --  instance of SharePoint module

        """
        self.sp_object = sp_object
        self.tc_object = sp_object.tc_object
        self.log = self.tc_object.log
        self.app_name = self.__class__.__name__
        self.csdb = sp_object.csdb
        self.browse_response = None

    def get_path_for_restore(self, folder=None, list=None):
        """Gets paths for restore from browse response

            Args:

                folder (boolean)      --    whether get path for folder

                list (boolean)        --    whether get path for list

        """
        try:
            paths = []
            if not self.browse_response:
                self.log.exception("Browse response was not checked yet")
                raise Exception("Browse response was not checked yet")
            if folder:
                for url in self.browse_response:
                    if url != "MB":
                        site_url = url.split("\\")[1]
                        if url.split("\\")[-1] == self.tc_object.testdata.testdata_metadata.get(
                                self.sp_object.site_url, {}). \
                                get("Folder", {}).get("Folder Name", "") and site_url == self.sp_object.site_url:
                            folder_url = url
                            paths.append(folder_url)
                            break
            if list:
                for url in self.browse_response:
                    if url != "MB":
                        site_url = url.split("\\")[1]
                        if url.split("\\")[-1] == self.tc_object.testdata.testdata_metadata.get(
                                self.sp_object.site_url,
                                {}). \
                                get("List", {}).get("Title", "") and site_url == self.sp_object.site_url:
                            list_url = url
                            paths.append(list_url)
                            break
            return paths
        except Exception as exception:
            self.log.exception("Exception while getting path for restore job: %s", str(exception))
            raise exception

    def verify_browse_response(self, browse_paths, exclude_paths):
        """Verifies browse response

            Args:

                browse_paths (list)         --    list of paths that should be in browse response

                exclude_paths (list)        --    list of paths that should not be in browse response

        """
        try:
            if self.browse_response:
                if browse_paths:
                    for path in browse_paths:
                        path = "MB\\" + path
                        if path in self.browse_response:
                            self.log.info(f"Browse for restore has {path} entry as expected")
                        else:
                            self.log.exception(f"Browse for restore is not successful for entry {path}")
                            raise Exception(f"Browse for restore is not successful for entry {path}")
                if exclude_paths:
                    for path in exclude_paths:
                        path = "MB\\" + path
                        if path in self.browse_response:
                            self.log.exception(f"Browse for restore has {path} entry")
                            raise Exception(f"Browse for restore has {path} entry")
                        else:
                            self.log.info(f"Browse for restore does not have {path} entry as expected")
            else:
                self.log.exception("Browse response is not set")
                raise Exception("Browse response is not set")
        except Exception as exception:
            self.log.exception("Exception while verifying browse paths: %s", str(exception))
            raise exception

    def browse_restore_content_and_verify_browse_response(self, browse_args=None, **kwargs):
        """Browses content for restore and verifies the browse response

            Args:

                browse_args (dict)         -- dictionary of browse arguments
                Example:

                    from_time (str)         --  from time for browse
                                                    Example: from_time = "0"

                    to_time (str)           --  to time for browse
                                                    Example: to_time = "1606204288"
            Kwargs:

                browse_paths (list)         --  list of paths that should be present in browse response
                                                Example:
                                                [
                                                   "MB\\https://test.sharepoint.com/sites/TestSite\\Subsites\\sub_1",
                                                   "MB\\https://test.sharepoint.com/sites/TestSite\\Subsites\\sub_2"
                                                ]
                exclude_browse_paths (list)  --  list of paths that should not be present in browse response
                                                Example:
                                                [
                                                   "MB\\https://test.sharepoint.com/sites/TestSite",
                                                   "MB\\https://test.sharepoint.com/sites/TestSite2\\Subsites\\sub"
                                                ]


        """
        if browse_args is None:
            browse_args = {}
        try:
            self.log.info("Browsing content for restore")
            paths, dictionary = self.sp_object.cvoperations.subclient.browse(path=r"\MB", **browse_args)
            self.browse_response = paths
            if self.browse_response:
                self.verify_browse_response(kwargs.get("browse_paths", []), kwargs.get("exclude_browse_paths", []))
            else:
                self.log.error("No content available for restore")
                raise Exception("Content not available for restore")
        except Exception as exception:
            self.log.exception("Exception occurred during browse: %s", str(exception))
            raise exception

    def validate_browse_for_restore(self, include_subsite_end_url_list, exclude_subsite_end_url_list=None):
        """Validates browse for restore

            Args:

                include_subsite_end_url_list (list)    --   list of subsites end urls
                                                            Example: ["sub1", "sub2"]

                exclude_subsite_end_url_list (list)    --   list of subsites end urls to be excluded in browse response
                                                            Example: ["sub1", "sub2"]
        """
        try:
            self.sp_object.cvoperations.wait_time(60, f"Waiting before browse")
            subsite_include_browse_list = []
            subsite_exclude_browse_list = []
            for site in include_subsite_end_url_list:
                subsite_include_browse_list.append(self.sp_object.site_url + "\\Subsites\\" + site)
            if exclude_subsite_end_url_list:
                for site in exclude_subsite_end_url_list:
                    subsite_exclude_browse_list.append(self.sp_object.site_url + "\\Subsites\\" + site)
            self.browse_restore_content_and_verify_browse_response(
                browse_paths=subsite_include_browse_list,
                exclude_browse_paths=subsite_exclude_browse_list)
        except Exception as exception:
            self.log.exception("Exception while validating browse response: %s", str(exception))
            raise exception

    def compare_backup_and_restore_metadata(self,
                                            backup_metadata,
                                            restore_metadata,
                                            exclude_paths=None):
        """Compares backup metadata and restore metadata

            Args:

                backup_metadata (dict)      --      metadata of backed up content
                                                    Example:
                                                        {
                                                            "Id":1,
                                                            "Title":"Test Title",
                                                            "MajorVersion":3,
                                                            "MinorVersion":0,
                                                            "Name":"Test File"
                                                        }

                restore_metadata (dict)     --      metadata of restored content
                                                    Example:
                                                        {
                                                            "Id":1,
                                                            "Title":"Test Title",
                                                            "MajorVersion":3,
                                                            "MinorVersion":0,
                                                            "Name":"Test File"
                                                        }

                exclude_paths (set)        --      properties to exclude while validating
                                                   Example:
                                                       {
                                                            "root['ContentTag']",
                                                            "root['UniqueId']",
                                                            "root['ETag']"
                                                       }
        """
        try:
            diff = DeepDiff(backup_metadata, restore_metadata, exclude_paths=exclude_paths)
            if diff:
                if 'dictionary_item_removed' in diff.keys():
                    # By default title field is not enabled after restore. An MR is created for this. Will remove the
                    # following code once it is resolved
                    if "root['Title']" in diff.get('dictionary_item_removed')[0]:
                        list_title = self.tc_object.testdata.json_value.List.TITLE
                        item_title = self.sp_object.get_sp_list_item_title(list_title)
                        if backup_metadata.get("Title") == item_title:
                            self.log.info("Backup and restore metadata are validated")
                else:
                    self.log.error("Difference in metadata {0}".format(diff))
                    self.log.exception("Backup and restore metadata are not same")
                    raise Exception("Backup and restore metadata are not same")
            else:
                self.log.info("Backup and restore metadata are validated")
        except Exception as exception:
            self.log.exception("Exception while comparing metadata: %s", str(exception))
            raise exception

    def validate_restored_files(self, backup_metadata_index=-1, latest_version=False, oop_site=None):
        """Validates restored files in the folder

            Args:

                backup_metadata_index (int)      --     index of the backup metadata list to consider
                                                        it can be any number with in list length range
                                                        Default: -1 (takes latest backup metadata)

                latest_version (boolean)         --     whether to validate only latest version metadata or
                                                        all versions metadata
                                                        Default: False (validates all versions)

                oop_site (str)                   --     out of place restore site if it is oop restore validation

        """
        try:
            self.log.info("Collecting and validating metadata of files")
            files = self.tc_object.testdata.json_value.Folder.FILES
            folder_name = self.tc_object.testdata.json_value.Folder.NAME
            exclude_paths = {"root['ContentTag']", "root['UniqueId']", "root['ETag']",
                             "root['TimeCreated']", "root['TimeLastModified']"}
            if oop_site:
                exclude_paths.add("root['ServerRelativeUrl']")
                site_name = "/".join(oop_site.split("/")[4:])
            else:
                site_name = "/".join(self.sp_object.site_url.split("/")[4:])
            folder_url = "/sites/" + site_name + "/shared documents/" + folder_name
            if latest_version:
                exclude_paths.add("root['Versions Metadata']")
            for file in files:
                backup_metadata = \
                    self.tc_object.testdata.testdata_metadata.get(self.sp_object.site_url, {}).get("Folder", {}).get(
                        file.Name)[backup_metadata_index]
                if oop_site:
                    self.tc_object.testdata.update_file_metadata(file.Name, folder_url, oop_site)
                else:
                    self.tc_object.testdata.update_file_metadata(file.Name, folder_url)
                restore_metadata = \
                    self.tc_object.testdata.testdata_metadata.get(self.sp_object.site_url, {}).get("Folder", {}).get(
                        file.Name)[-1]
                if latest_version:
                    if len(restore_metadata["Versions Metadata"]) > 0:
                        self.log.exception("Didn't restore only latest version")
                        raise Exception("Didn't restore only latest version")
                self.compare_backup_and_restore_metadata(backup_metadata, restore_metadata,
                                                         exclude_paths=exclude_paths)
                self.log.info(f"{file.Name} is validated successfully")
        except Exception as exception:
            self.log.exception("Exception while validating file metadata: %s", str(exception))
            raise exception

    def validate_restored_list_items(self, backup_metadata_index=-1, latest_version=False, oop_site=None):
        """Validates restored list items in the list

             Args:

                backup_metadata_index (int)     --      index of the backup metadata list to consider
                                                        it can be any number with in list length range
                                                        Default: -1 (takes latest backup metadata)

                latest_version (boolean)         --     whether to validate only latest version metadata or
                                                        all versions metadata
                                                        Default: False (validates all versions)

                oop_site (str)                   --     out of place restore site if it is oop restore validation

        """
        try:
            self.log.info("Collecting and validating metadata of list items")
            list_items = self.tc_object.testdata.json_value.List.LIST_ITEMS
            list_title = self.tc_object.testdata.json_value.List.TITLE
            index = 1
            exclude_paths = {"root['ContentTypeId']", "root['GUID']", "root['ItemChildCount']",
                             "root['FolderChildCount']"}
            if latest_version:
                exclude_paths.add("root['Versions Metadata']")
                exclude_paths.add("root['OData__UIVersionString']")
            for list_item in list_items:
                list_item_id = index
                backup_metadata = \
                    self.tc_object.testdata.testdata_metadata.get(self.sp_object.site_url, {}).get("List", {}).get(
                        "List Items", {}).get(list_item_id, [])[backup_metadata_index]
                if oop_site:
                    self.tc_object.testdata.update_list_item_metadata(list_title, list_item_id, oop_site)
                else:
                    self.tc_object.testdata.update_list_item_metadata(list_title, list_item_id)
                restore_metadata = \
                    self.tc_object.testdata.testdata_metadata.get(self.sp_object.site_url, {}).get("List", {}).get(
                        "List Items", {}).get(list_item_id, [])[-1]
                if latest_version:
                    if len(restore_metadata["Versions Metadata"]) > 1 and \
                            restore_metadata['OData__UIVersionString'] != "1.0":
                        self.log.exception("Didn't restore only latest version")
                        raise Exception("Didn't restore only latest version")
                self.compare_backup_and_restore_metadata(backup_metadata, restore_metadata,
                                                         exclude_paths=exclude_paths)
                self.log.info(f"list item {list_item.Title} is validated successfully")
                index = index + 1
        except Exception as exception:
            self.log.exception("Exception while validating list item metadata: %s", str(exception))
            raise exception

    def restore_and_validate_sharepoint_content(self, restore_args=None, folder=None, list=None,
                                                backup_metadata_index=-1, site_url_list=None):
        """Restores the selected content

            Args:

                restore_args (dict)             --      dictionary of restore parameters
                                                        Example: restore_args = {
                                                                                    "overwrite": True,
                                                                                    "showDeletedItems": True
                                                                                }

                folder (boolean)                --      whether to restore folder if it is granular level restore

                list (boolean)                  --      whether to restore list if it is granular level restore

                backup_metadata_index (int)     --      index of the backup metadata to consider

                site_url_list (list)            --      if you want to restore and validate multiple sites

        """
        if restore_args is None:
            restore_args = {}
        try:
            if not restore_args.get("paths", []):
                restore_args["paths"] = self.get_path_for_restore(folder, list)
            self.log.info("Restoring folder and list")
            self.log.info("Path for restore:{0}".format(restore_args["paths"]))
            self.sp_object.cvoperations.run_restore(**restore_args)
            self.log.info("Collecting metadata from SharePoint Site and validating it")
            if site_url_list:
                sites_to_validate = site_url_list
            else:
                sites_to_validate = [self.sp_object.site_url]
            for site in sites_to_validate:
                self.sp_object.site_url = site
                self.log.info(f"Validating restore metadata for {site}")
                if folder:
                    self.validate_restored_files(backup_metadata_index=backup_metadata_index,
                                                 latest_version=restore_args.get("latest_version", False),
                                                 oop_site=restore_args.get("destination_site", ""))
                if list:
                    self.validate_restored_list_items(backup_metadata_index=backup_metadata_index,
                                                      latest_version=restore_args.get("latest_version", False),
                                                      oop_site=restore_args.get("destination_site", ""))
            self.tc_object.testdata.share_point_data_flag = True
        except Exception as exception:
            self.log.exception("Exception while triggering restore job: %s", str(exception))
            raise exception

    def process_disk_files(self, file_list, item=None):
        """Returns only file names from file paths of disk files

            Args:

                file_list (list)             --  list of disk files

                item (boolean)               --  whether it is a restored item or default created folder file list

        """
        try:
            dat_count = 0
            user_group_merged_files_count = 0
            files = []
            for file in file_list:
                file_name = file.split("\\")[-1]
                if file_name.endswith(".dat"):
                    dat_count = dat_count + 1
                elif "_UserGroup_" in file_name:
                    user_group_merged_files_count = user_group_merged_files_count + 1
                else:
                    files.append(file_name)
            if user_group_merged_files_count == 0 and (item == "root" or item == "cvtempbackup"):
                self.log.error(f"user group files not restored for {item}")
                raise Exception(f"user group files not restored for {item}")
            elif dat_count == 0 and item == "restored_item":
                self.log.error(f".dat files not restored for {item}")
                raise Exception(f".dat files not restored for {item}")
            else:
                self.log.info(dat_count, " .dat files restored")
            return files
        except Exception as exception:
            self.log.exception('An error occurred while processing disk files')
            raise exception

    def compare_restore_disk_native_files(self, restore_items, destination_client, destination_dir, num_of_backups):
        """Compare file properties for disk restores

            Args:

                restore_items (dict)        --  list of restored items

                destination_client (str)    --  client name where the disk restore happened

                destination_dir (str)       --  path where the disk restore happened

                num_of_backups (int)        -- number of backups ran before disk restore

        """
        try:
            machine_obj = Machine(destination_client, self.tc_object.commcell)
            if not machine_obj.check_directory_exists(destination_dir):
                raise Exception("The directory does not exist")
            disk_restore_files = constants.DISK_RESTORE_AS_NATIVE_FILES_VALIDATE_FILES
            restored_disk_files = []
            root_files_num = len(disk_restore_files.get("root")) + num_of_backups
            restored_items_count = 0
            disk_folders = machine_obj.get_folders_in_path(folder_path=destination_dir)
            for folder in disk_folders:
                if folder.split("\\")[-1] == "cvtempbackup":
                    cvtempbackup_files = machine_obj.get_files_in_path(folder_path=folder)
                    restored_disk_files.append(
                        ("cvtempbackup", self.process_disk_files(cvtempbackup_files, "cvtempbackup"), folder))
                elif folder.split("\\")[-1] in restore_items:
                    restored_items_count = restored_items_count + 1
                    item_files = machine_obj.get_files_in_path(folder_path=folder)
                    restored_disk_files.append(
                        ("restored_item", self.process_disk_files(item_files, "restored_item"), folder))
                else:
                    root_files = machine_obj.get_files_in_path(folder_path=folder)[:root_files_num]
                    restored_disk_files.append(("root", self.process_disk_files(root_files, "root"), folder))
            for item_files in restored_disk_files:
                self.log.info(f"Restored files: {item_files[1]}")
                self.log.info(f"Files to be restored: {disk_restore_files.get(item_files[0])}")
                if item_files[1] == disk_restore_files.get(item_files[0]):
                    self.log.info(f"Disk files for {item_files[2]} are validated")
                else:
                    self.log.exception(f"Disk files for {item_files[2]} are not validated")
                    raise Exception(f"Disk files for {item_files[2]} are not validated")
            else:
                if restored_items_count == len(restore_items):
                    self.log.info("All disk files are validated")
                    remote_machine = Machine(destination_client, self.sp_object.cvoperations.commcell)
                    remote_machine.remove_directory(destination_dir, 0)
                else:
                    self.log.exception("All disk files are not restored")
        except Exception as exception:
            self.log.exception('An error occurred while validating restore to disk as native files')
            raise exception

    def compare_restore_disk_original_files(self, before_backup_metadata, destination_client,
                                            destination_dir):
        """Compare file properties for disk restores

            Args:

                before_backup_metadata (dict)        --  file properties before backup

                destination_client (str)             --  client name where the disk restore happened

                destination_dir (str)                --  path where the disk restore happened

        """
        try:
            machine_obj = Machine(destination_client, self.tc_object.commcell)
            if not machine_obj.check_directory_exists(destination_dir):
                raise Exception("The directory does not exist")
            disk_restored_files_properties = machine_obj.scan_directory(path=destination_dir, filter_type="file")
            file_versions_metadata = before_backup_metadata.get("Versions Metadata")
            for i in range(len(file_versions_metadata)):
                if not file_versions_metadata[i].get("Size") == int(disk_restored_files_properties[i].get("size")):
                    self.log.exception(f"{disk_restored_files_properties[i]['path']} file is not validated properly")
                    raise Exception(f"{disk_restored_files_properties[i]['path']} file is not validated properly")
            else:
                self.log.info(f"{len(file_versions_metadata)} versions are validated")
                latest_version_file = disk_restored_files_properties[-1]["path"]
                version = latest_version_file.split("\\")[-2]
                file = latest_version_file.split("\\")[-1]
                if file == before_backup_metadata.get("Name") and version == before_backup_metadata.get(
                        "UIVersionLabel"):
                    self.log.info("Latest version is validated")
                    remote_machine = Machine(destination_client, self.sp_object.cvoperations.commcell)
                    remote_machine.remove_directory(destination_dir, 0)
                else:
                    self.log.exception("Latest version file is not validated")
                    raise Exception(f"{latest_version_file} file is not validated properly")
        except Exception as exception:
            self.log.exception('An error occurred while validating restore to disk as original files')
            raise exception
