# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
# See LICENSE.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
    Main Module for setting input variables and defining sharepoint specific functions
    This module is imported in any test case.
    You need to create an object of this module in the test case.

    SharePoint Online Pseudo Client: Class for defining sharepoint
    specific functions i.e., for making Graph and REST api calls
"""

from __future__ import unicode_literals
import random
import urllib.parse
import json
import requests
from office365.sharepoint.client_context import ClientContext
from office365.runtime.utilities.request_options import RequestOptions
from office365.runtime.utilities.http_method import HttpMethod
from office365.runtime.auth.authentication_context import AuthenticationContext
from O365 import Account
from O365.sharepoint import Sharepoint
from ..Office365.operations import SharepointCvOperation
from . import sharepointconstants


class SPGraphAPIHelper(Sharepoint):
    """Class for defining SharePoint graph api functions
    """

    def __init__(self, *, parent=None, con=None, **kwargs):
        """Initializes the SPGraphAPIHelper object

            Args:

                parent (Account object)     --  parent object

                con (Connection object)     --  connection to use if no parent specified

            Kwargs:

                protocol (Protocol object)  --  protocol to use if no parent specified

                main_resource (str)         --  use this resource instead of parent resource

        """
        if parent and con:
            raise ValueError('Need a parent or a connection but not both')
        self.con = parent.con if parent else con

        main_resource = kwargs.pop('main_resource',
                                   '')
        super().__init__(con=self.con,
                         protocol=parent.protocol if parent else kwargs.get('protocol'),
                         main_resource=main_resource)

    def search_sites(self, keyword):
        """Searches the SharePoint tenant for sites with the provided keyword

            Args:

                keyword (list[Site])        --   a keyword to search sites

        """
        if not keyword:
            raise ValueError('Must provide a valid keyword')

        url = self.build_url(
            self._endpoints.get('search').format(keyword=keyword))

        response = self.con.get(url)
        if not response:
            return []

        data = response.json()
        next_link = data.get('@odata.nextLink')
        return_dict = {}
        if next_link:
            return_dict['nextLink'] = next_link

        return_dict['site_info'] = [
            self.site_constructor(parent=self, **{self._cloud_data_key: site})
            for site in data.get('value', [])]
        return return_dict

    def get_user(self, keyword):
        """Searches the SharePoint tenant for the user with the provided keyword

            Args:

                keyword (list[Site])        --   a keyword to search users

        """
        if not keyword:
            raise ValueError('Must provide a valid keyword')
        self._endpoints['users'] = '/users/{keyword}'

        url = self.build_url(
            self._endpoints.get('users').format(keyword=keyword))

        response = self.con.get(url)
        if not response:
            return []

        data = response.json()
        if 'error' in data:
            error_code = response.json()['error']['code']

            if error_code != 0:
                error_string = response.json()['response']['errorString']
                raise Exception('Failed to get User Information\nError: "{0}"'.format(error_string))
        elif 'userPrincipalName' in data:
            return data


class SharePointOnline:
    """Class for defining SharePoint specific functions i.e., for making graph
    and REST api calls"""

    def __init__(self, tc_object):
        """creates object from other modules.

            Args:

                tc_object   --  instance of testcase class

            Returns:

                object      --  instance of SharePointOnline class

        """
        self.tc_object = tc_object
        self.log = self.tc_object.log
        self.log.info('logger initialized for SharePoint Client')
        self.app_name = self.__class__.__name__
        self.csdb = self.tc_object.csdb
        self.cvoperations = SharepointCvOperation(self)
        self.pseudo_client_name = None
        self.server_plan = None
        self.index_server = None
        self.access_nodes_list = None
        self.global_administrator = None
        self.global_administrator_password = None
        self.azure_app_id = None
        self.azure_app_key_id = None
        self.azure_directory_id = None
        self.azure_username = None
        self.azure_secret = None
        self.graph_client_id = None
        self.graph_client_secret = None
        self.graph_tenant_id = None
        self.site_collection_list = []
        self.sp_api_object = None
        self.site_url_list = []
        self.site_url = None
        self.office_365_plan = None
        self.api_client_id = None
        self.api_client_secret = None
        self.tenant_url = None
        self.user_username = None
        self.user_password = None
        self._ctx_auth = None

    @property
    def ctx_auth(self):
        """Returns the context authentication object"""
        if self._ctx_auth is None:
            self._ctx_auth = AuthenticationContext(self.site_url)
            self.generate_token()
        return self._ctx_auth

    def authenticate_microsoft_tenant(self):
        """This function authenticates the Microsoft tenant and
        creates an account object for running Microsoft graph APIs
        """
        try:
            credentials = (self.graph_client_id, self.graph_client_secret)
            account = Account(credentials, auth_flow_type='credentials',
                              tenant_id=self.graph_tenant_id)
            if account.authenticate():
                self.log.info('Client authenticated successfully for running graph API')
                return account
            self.log.error('Client authenticated failed for running graph API')
            raise Exception('Client authenticated failed for running graph API')
        except Exception as exception:
            self.log.exception("Exception while authenticating microsoft tenant: %s", str(exception))
            raise exception

    def get_sites_count(self):
        """Returns total count of site collections present in the tenant
        """
        try:
            # Authorise microsoft tenant account to run graph API
            account = self.authenticate_microsoft_tenant()
            # Call graph API and get response and count number of site collections
            api_site_count = 0
            self.sp_api_object = SPGraphAPIHelper(con=account.con, protocol=account.protocol)
            api_search_keyword = '*'
            while True:
                site_data = self.sp_api_object.search_sites(api_search_keyword)
                for site in site_data.get('site_info'):
                    if site.root:
                        self.site_collection_list.append(site)
                        api_site_count = api_site_count + 1
                if not site_data.get('nextLink'):
                    break
                api_search_keyword = site_data.get('nextLink').partition('=')[2]
            return api_site_count
        except Exception as exception:
            self.log.exception("Exception while getting count of sites: %s", str(exception))
            raise exception

    def get_group_based_sites_count(self):
        """Returns count of sites group wise including its subsites present in tenant
        """
        try:
            all_sites_count = self.get_sites_count()
            group_team_site_count = 0
            project_online_site_count = 0
            for site in self.site_collection_list:
                self.site_url = urllib.parse.unquote(site.web_url)
                site_properties = self.get_site_properties(root_site=True)
                if site_properties.get("WebTemplate") == "GROUP" or site_properties.get("WebTemplate") == "TEAMCHANNEL":
                    group_team_site_count = group_team_site_count + 1
                elif site_properties.get("WebTemplate") == "PWA":
                    project_online_site_count = project_online_site_count + 1
                sub_sites = site.get_subsites()
                if sub_sites:
                    for sub_site in sub_sites:
                        if sub_site:
                            if site_properties.get("WebTemplate") == "GROUP" or site_properties.get(
                                    "WebTemplate") == "TEAMCHANNEL":
                                group_team_site_count = group_team_site_count + 1
                            elif site_properties.get("WebTemplate") == "PWA":
                                project_online_site_count = project_online_site_count + 1
                            all_sites_count = all_sites_count + 1
            self.log.info(f"All Web Sites : {all_sites_count}")
            self.log.info(f"All Groups And Teams Sites : {group_team_site_count}")
            self.log.info(f"All Project Online Sites : {project_online_site_count}")
            return {
                "All Web Sites": all_sites_count,
                "All Groups And Teams Sites": group_team_site_count,
                "All Project Online Sites": project_online_site_count
            }
        except Exception as exception:
            self.log.exception("Exception while getting group based sites count: %s", str(exception))
            raise exception

    def validate_site_collections_count(self):
        """Validates site collections
            1. Count of site collections using Microsoft Graph API
            2. Count of site collections by querying CSDB
            Validates by comparing above two counts
        """
        try:
            api_site_count = self.get_sites_count()
            # Query CSDB to get site collections count, ItemType = 2
            query_string = sharepointconstants.CSDB_QUERY['site collection count']
            query = query_string.format(self.cvoperations.subclient.subclient_id,
                                        sharepointconstants.ITEM_TYPE['site collections'])
            self.csdb.execute(query)
            csdb_count = self.csdb.fetch_one_row()

            """
            The following urls are present in graph api_response and not in csdb
            https://cvdevtenant.sharepoint.com/sites/contentTypeHub

            The following urls are present in  csdb and not in graph api_response
            https://cvdevtenant.sharepoint.com/search
            https://cvdevtenant-my.sharepoint.com/

            Search template sites are not getting fetched by Graph API

            For now assuming the site count difference shouldn't exceed 10
            """

            self.log.info("CSDB Site Count  : {0}, Graph API Site Count : {1} ".format(csdb_count[0], api_site_count))
            if int(csdb_count[0]) - api_site_count > 10:
                self.log.error("Site Collections are not equal")
                raise Exception("Site Collections are not equal")
        except Exception as exception:
            self.log.exception("Exception while validating site collections count: %s", str(exception))
            raise exception

    def validate_site_data(self):
        """Validates the following site properties
            1. Display name
            2. SMTP Address
            3. Validates all webs
        """
        try:
            no_of_sites = len(self.site_collection_list)
            sites_validated_count = 0
            no_of_sites_to_be_validated = no_of_sites // 10
            if no_of_sites_to_be_validated > 20:
                no_of_sites_to_be_validated = 20
            self.log.info("Validating Site Data for random {0} sites".format(no_of_sites_to_be_validated))
            for i in range(no_of_sites_to_be_validated):
                site_index = random.randrange(0, no_of_sites)
                site = self.site_collection_list[site_index]
                smtp_address = urllib.parse.unquote(site.web_url).replace("\'", "\'\'")
                query_string = sharepointconstants.CSDB_QUERY['site collection properties']
                query = query_string.format(self.cvoperations.subclient.subclient_id,
                                            sharepointconstants.ITEM_TYPE['site collections'],
                                            smtp_address)
                self.csdb.execute(query)
                csdb_site_data = self.csdb.fetch_one_row()
                self.log.error("Query Result : {0}".format(csdb_site_data))
                if len(csdb_site_data) > 0:
                    if csdb_site_data[0] != '' and csdb_site_data[1].split('/')[-1] == site.name:
                        sites_validated_count = sites_validated_count + 1
                        self.log.info("{0} site collection is present in csdb".format(smtp_address))
                        self.validate_webs_for_the_site(site)
                    else:
                        self.log.error("Unable to fetch site information for {0} from csdb".format(smtp_address))
                else:
                    """
                    Few site collections are not present in csdb
                    For now just displaying the url if it not present in csdb
                    """
                    self.log.error("{0} site collection is not present in csdb".format(smtp_address))
            self.log.info(("{0} sites and their web data are validated correctly out of {1} "
                           "sites").format(sites_validated_count, no_of_sites_to_be_validated))
            if no_of_sites_to_be_validated - sites_validated_count > 5:
                self.log.exception("Site Validation is not done properly")
                raise Exception("Site Validation is not done properly")
        except Exception as exception:
            self.log.exception("Exception while validating site data: %s", str(exception))
            raise exception

    def validate_site_collections(self):
        """Validates site collections by
            1. Validating site collections count
            2. Validating site data
        """
        self.validate_site_collections_count()
        self.log.info("Site Collections count is validated successfully")
        self.validate_site_data()
        self.log.info("Site Data is validated successfully")

    def validate_webs_for_the_site(self, site):
        """Validates webs i.e., all sub_sites  of a site collection in share point tenant
            1. Get sub_sites i.e, webs of the site collection
            2. Query csdb to validate sub_sites. The following properties are validated
                -> SMTP Address
                -> Display name
                -> ItemType = 1

            Args:

                site (Site)   --  object of site collection

        """
        try:
            sub_sites = site.get_subsites()
            if sub_sites:
                for sub_site in sub_sites:
                    if sub_site:
                        smtp_address = urllib.parse.unquote(sub_site.web_url).replace("\'", "\'\'")
                        query_string = sharepointconstants.CSDB_QUERY['web properties']
                        query = query_string.format(self.cvoperations.subclient.subclient_id,
                                                    sharepointconstants.ITEM_TYPE['site collections'], smtp_address)
                        self.csdb.execute(query)
                        csdb_web_data = self.csdb.fetch_one_row()
                        self.log.info("Query Result : {0}".format(csdb_web_data))
                        if len(csdb_web_data) > 0 and csdb_web_data[0] != '':
                            if csdb_web_data[1].split('/')[-1] == sub_site.name and int(csdb_web_data[2]) == 1:
                                self.log.info("{0} web is present in csdb".format(sub_site.web_url))
                            else:
                                self.log.error("{0} web properties are not correct".format(sub_site.web_url))
                        else:
                            self.log.error("{0} web is not present in csdb".format(sub_site.web_url))
                self.log.info("Validated webs successfully for the site {0}".format(site.web_url))
            else:
                self.log.info("No webs present for the site {0}".format(site.web_url))
        except Exception as exception:
            self.log.exception("Exception while validating subsites of the site: %s", str(exception))
            raise exception

    def validate_properties(self, property_list_1, property_list_2):
        """Validates whether properties in provided lists are equal or not

            Args:

               property_list_1 (list)   --  list of tuples of properties and their values
                                            Example:
                                                property_list_1 = [('property1', value1),('property2', value2)]

               property_list_2 (list)   --  list of values of properties
                                            Example:
                                                property_list_2 = [value1, value2]

           Raises:

                Exception:
                    if subsite properties are not equal


        """
        try:
            for i in range(len(property_list_1)):
                if property_list_1[i][1] == property_list_2[i]:
                    self.log.info(f"{property_list_1[i][0]} is validated")
                else:
                    self.log.error(f"Values for {property_list_1[i][0]} property are not equal")
                    raise Exception(f"values for {property_list_1[i][0]} property are not equal")
            else:
                self.log.info("All properties are validated")
        except Exception as exception:
            self.log.exception("Exception while validating property: %s", str(exception))
            raise exception

    def validate_subsite_properties(self, subsite_metadata):
        """Validates properties of subsite in CSDB

            Query csdb to validate sub_site. The following properties are validated
               -> SMTP Address
               -> Display name
               -> Type of Association
               -> flags value
               -> Status for deleted sites

           Args:

               subsite_metadata (dict)   --  metadata of subsite

           Raises:

                Exception:
                    if subsite properties are not correct

        """
        try:
            subsite_url = self.site_url + '/' + subsite_metadata.get('Url End', "")
            query_string = sharepointconstants.CSDB_QUERY['web properties']
            query = query_string.format(self.cvoperations.subclient.subclient_id, sharepointconstants.ITEM_TYPE['web'],
                                        subsite_url)
            self.csdb.execute(query)
            csdb_web_data = self.csdb.fetch_one_row()
            self.log.info("Query Result : {0}".format(csdb_web_data))
            if len(csdb_web_data) > 0 and csdb_web_data[0] != '':
                input_web_data = [("Title", subsite_metadata.get('Title', "")), ("URL", subsite_url),
                                  ("Item Type", '1'),
                                  ("Association Type", 'true'), ("Associated Flags Value",
                                                                 subsite_metadata.get('Associated Flags Value', "")),
                                  ('Office 365 Plan Id', str(subsite_metadata.get("Office 365 Plan Id", 0)))]
                if subsite_metadata.get('Operation') == 'DELETED':
                    input_web_data.append(('Status', '1'))
                self.log.info(f"Validating list : {input_web_data}")
                self.validate_properties(input_web_data, csdb_web_data)
                self.log.info(f"{subsite_url} site is validated in csdb")
            else:
                self.log.error(f"{subsite_url} record is not present in csdb")
        except Exception as exception:
            self.log.exception("Exception while validating subsites: %s", str(exception))
            raise exception

    def generate_token(self):
        """Generates token for making SharePoint REST API calls

            Returns:

                True (boolean)         --  if token is generated

            Raises:

                Exception:
                    if token is not generated even after 5 attempts

        """
        try:
            if not self.site_url:
                if self.site_url_list:
                    self.site_url = self.site_url_list[0]
                else:
                    raise Exception("Site url is empty. Please provide site url or site url list")
            self.log.info("Generating token for making SharePoint REST API calls")
            count = 0
            while count < 5:
                if self.ctx_auth.acquire_token_for_app(self.api_client_id,
                                                       self.api_client_secret):
                    self.log.info("Token is generated successfully")
                    return True
                else:
                    count = count + 1
                    self.log.info("Token is not generated successfully")
                    self.cvoperations.wait_time(120, "Sleeping for 120 seconds before retrying")
            if count > 5:
                raise Exception(f"Exception in generating token for making SharePoint REST API calls : "
                                f"{self.ctx_auth.provider.error}")
        except Exception as exception:
            self.log.exception("Exception while generating token: %s", str(exception))
            raise exception

    def make_patch_request(self, full_url, request_body=None):
        """Makes SharePoint PATCH request using Python

            Args:

                full_url (str)             --  URL to make patch request

                request_body (dict)        --  body of the request

            Returns:

                response (dict)            --  response of the patch request

        """
        try:
            context = ClientContext(full_url, self.ctx_auth)
            options = RequestOptions(full_url)
            options.set_header('Accept', 'application/json; odata=verbose')
            options.set_header('Content-Type', 'application/json')
            options.set_header('If-Match', '*')
            options.method = "PATCH"
            context.authenticate_request(options)
            if request_body is None:
                request_body = {}
            response = requests.patch(url=full_url, data=json.dumps(request_body), headers=options.headers)
            return response
        except Exception as exception:
            self.log.exception("Exception while making patch request: %s", str(exception))
            raise exception

    def make_post_request(self, full_url, request_body=None, headers=None):
        """Makes SharePoint POST request using Python

            Args:

                full_url (str)             --  URL to make post request

                headers(dict)              -- additional headers to make post request

                request_body (dict)        --  body of the request

            Returns:

                response (dict)            --  response of the post request

        """
        try:
            context = ClientContext(full_url, self.ctx_auth)
            options = RequestOptions(full_url)
            if headers:
                options.set_headers(headers)
            else:
                options.set_header('Accept', 'application/json; odata=verbose')
                options.set_header('Content-Type', 'application/json')
            options.method = HttpMethod.Post
            context.authenticate_request(options)
            if request_body is None:
                request_body = {}
            response = requests.post(url=full_url, data=json.dumps(request_body), headers=options.headers)
            return response
        except Exception as exception:
            self.log.exception("Exception while making post request: %s", str(exception))
            raise exception

    def make_get_request(self, full_url):
        """Makes SharePoint GET request using Python

            Args:
                    full_url (str)             --  URL to make get request

            Returns:
                    response (dict)            --  response of the get request

        """
        try:
            context = ClientContext(full_url, self.ctx_auth)
            options = RequestOptions(full_url)
            options.set_header('Accept', 'application/json; odata=verbose')
            options.set_header('Content-Type', 'application/json')
            options.method = HttpMethod.Get
            context.authenticate_request(options)
            response = requests.get(url=full_url, headers=options.headers)
            return response
        except Exception as exception:
            self.log.exception("Exception while making get request: %s", str(exception))
            raise exception

    def make_delete_request(self, full_url):
        """Makes SharePoint DELETE request using Python

             Args:
                    full_url (str)             --  URL to make delete request

            Returns:
                    response (dict)            --  response of the delete request

        """
        try:
            context = ClientContext(full_url, self.ctx_auth)
            options = RequestOptions(full_url)
            options.set_header('If-Match', '*')
            options.method = HttpMethod.Delete
            context.authenticate_request(options)
            response = requests.delete(url=full_url, headers=options.headers)
            return response
        except Exception as exception:
            self.log.exception("Exception while making delete request: %s", str(exception))
            raise exception

    def get_request_to_create_subsite(self, title, url_end, web_template=None):
        """Attempts to create a subsite in SharePoint Site

             Args:

                 title (str)               --   title of the subsite

                 url_end (str)             --   url end of the subsite

                 web_template (str)        --   web template of the subsite

            Returns:

                 request_body (dict)       --  body of the request

                                               request_body : {
                                                        'parameters': {
                                                        '__metadata':  {'type': 'SP.WebInfoCreationInformation'},
                                                        'Url': 'RestSubWeb1',
                                                        'Title': 'RestSubWeb1',
                                                        'Description': 'REST created web',
                                                        'Language': 1033,
                                                        'WebTemplate': 'STS',
                                                        'UseUniquePermissions': False}
                                                    }

            Raises:
                    Exception:
                        if response is not success
        """
        try:
            if not web_template:
                web_template = 'STS'
            self.log.info("Creating a sub site")
            request_body = \
                {
                    'parameters':
                        {
                            '__metadata':
                                {
                                    'type': 'SP.WebInfoCreationInformation'
                                },
                            'Url': url_end,
                            'Title': title,
                            'Description': 'It is a subsite',
                            'Language': 1033,
                            'WebTemplate': web_template,
                            'UseUniquePermissions': False
                        }
                }
            return request_body
        except Exception as exception:
            self.log.exception("Exception while getting request to create subsite: %s", str(exception))
            raise exception

    def create_subsite(self, title, url_end, web_template=None):
        """Attempts to create a subsite in SharePoint Site

             Args:

                 title (str)               --   title of the subsite

                 url_end (str)             --   url end of the subsite

                 web_template (str)        --   web template of the subsite

            Returns:

                response (dict)             --  metadata of site created

            Raises:

                Exception:
                    if response is not success

        """
        try:
            request_body = self.get_request_to_create_subsite(title, url_end, web_template)
            subsite_end_url = request_body.get('parameters').get('Url')
            if self.get_site_properties(subsite_end_url):
                self.delete_subsite(subsite_end_url)
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['CREATE_SUBSITE'].format(
                self.site_url)
            self.log.info(f"Request Body : {request_body}")
            subsite_url = self.site_url + '/' + request_body.get('parameters').get('Url')
            self.log.info(f"Subsite URL : {subsite_url}")

            headers = {
                'Accept': 'application/json; odata=verbose',
                'Content-Type': 'application/json; odata=verbose'
            }
            response = self.make_post_request(full_url, request_body, headers)
            if response.status_code == 200 or response.status_code == 201:
                self.log.info(f"{subsite_url} site created successfully")
                return response.json().get("d", {})
            else:
                raise Exception(f"Failed to create site with status code {response.status_code}")
        except Exception as exception:
            self.log.exception("Exception while creating a subsite: %s", str(exception))
            raise exception

    def get_site_properties(self, subsite_end_url=None, root_site=None, additional_uri=None):
        """Returns the properties of the SharePoint subsite

               Args:
                      subsite_end_url (str)    --   end part of site url if it is a subsite
                                                    Example: 'subsite' in https:test.sharepoint.com/sites/site/subsite

                      root_site (boolean)      --   if it is root site or not

                      additional_uri(str)      --   additional uri in API endpoint
                                                    Example: Title

              Returns:

                      properties_dict (dict)   --   dictionary of site properties

              Raises:

                      Exception:
                          if response is not success

          """
        try:
            if root_site:
                site_url = self.site_url
            else:
                site_url = self.site_url + '/' + subsite_end_url
            if additional_uri:
                full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['WEB_SPECIFIC_METADATA'].format(
                    site_url, additional_uri)
            else:
                full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['WEB'].format(
                    site_url)
            response = self.make_get_request(full_url)
            if response.status_code == 200 or response.status_code == 201:
                return response.json().get("d", {})
            elif response.status_code == 404:
                self.log.info(f"{full_url} site is not present in the SharePoint tenant")
                return {}
        except Exception as exception:
            self.log.exception("Exception while getting properties of subsite: %s", str(exception))
            raise exception

    def update_subsite_level_properties(self, prop_dict, subsite_end_url):
        """Updates properties of a SharePoint subsite

            Args:

                prop_dict (dict)         --  dictionary of properties
                                                     Example: {
                                                                "ServerRelativeUrl": "subsite",
                                                                 "Title": "Subsite Title"
                                                              }

                subsite_end_url(str)    --  end url of subsite

           Raises:

               Exception:
                   if response is not success

           """
        try:
            request_body = {
                "__metadata": {
                    "type": "SP.Web"
                }
            }
            request_body.update(prop_dict)
            subsite_url = self.site_url + '/' + subsite_end_url
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['WEB'].format(subsite_url)
            self.log.info(f"API URL : {full_url}")
            headers = {
                'Accept': 'application/json; odata=verbose',
                'Content-Type': 'application/json; odata=verbose',
                'X-HTTP-Method': 'MERGE'
            }
            response = self.make_post_request(full_url, request_body, headers)
            if response.status_code == 200 or response.status_code == 204:
                self.log.info(f"Site level properties of site {subsite_url} are updated successfully")
            else:
                raise Exception(f"Failed to update site level properties with status code {response.status_code}")
        except Exception as exception:
            self.log.exception("Exception while updating subsite level properties: %s", str(exception))
            raise exception

    def create_subsites(self, subsite_list):
        """Creates subsites and returns metadata of them

            Args:

                site_list (list)           --    list of sites dictionary to create subsites

        """
        try:
            subsites_metadata = {}
            for site in subsite_list:
                title = site.get("Title")
                url_end = site.get("Url End")
                response = self.create_subsite(title, url_end)
                subsites_metadata[response.get('ServerRelativeUrl')] = {
                    'Url End': response.get('ServerRelativeUrl').split("/")[-1],
                    'Title': response.get('Title', ""),
                    'Operation': "ADDED"
                }
            return subsites_metadata
        except Exception as exception:
            self.log.exception("Exception while creating subsites: %s", str(exception))
            raise exception

    def delete_subsites(self, subsite_end_url_list, additional_uri=None):
        """Delete subsites specified in site list

            Args:

                subsite_end_url_list (list)           --    list of end urls of subsites

                additional_uri(str)                   --    additional uri in API endpoint
                                                            Example: Title

        """
        try:
            for subsite_end_url in subsite_end_url_list:
                if self.get_site_properties(subsite_end_url=subsite_end_url, additional_uri=additional_uri):
                    self.log.info("Site is present in SharePoint site")
                    self.delete_subsite(subsite_end_url)
        except Exception as exception:
            self.log.exception("Exception while deleting subsites: %s", str(exception))
            raise exception

    def delete_subsite(self, subsite_end_url):
        """Deletes a SharePoint subsite

           Args:

              subsite_end_url      --  end part of subsite url
                                       Example: 'subsite' in https:test.sharepoint.com/sites/site/subsite

          Returns:

              boolean              --  if the site is deleted successfully

          Raises:

              Exception:
                  if response is not success

          """
        try:
            subsite_url = self.site_url + '/' + subsite_end_url
            self.log.info(f"Subsite URL : {subsite_url}")
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['WEB'].format(
                subsite_url)
            response = self.make_delete_request(full_url)
            if response.status_code == 200 or response.status_code == 201:
                self.log.info(f"{subsite_url}\nsite deleted successfully")
            else:
                raise Exception(f"Failed to delete site with status code {response.status_code}")
        except Exception as exception:
            self.log.exception("Exception while deleting subsite: %s", str(exception))
            raise exception

    def upload_sp_binary_file(self, folder_path, file_name):
        """Attempts to upload a binary file in SharePoint Site

             Args:

                folder_path (str)           --  path of the folder where file has to be uploaded

                file_name (str)             --  name of the file

            Returns:

                response (dict)             --  metadata of file created

            Raises:

                Exception:
                    if response is not success

        """
        try:
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['CREATE_FILE'].format(
                self.site_url, folder_path, file_name)
            context = ClientContext(full_url, self.ctx_auth)
            options = RequestOptions(full_url)
            options.set_header('Accept', 'application/json; odata=verbose')
            options.set_header('Content-Type', 'application/octet-stream')
            options.method = HttpMethod.Post
            with open(file_name, 'rb') as outfile:
                context.authenticate_request(options)
                response = requests.post(url=full_url, data=outfile, headers=options.headers)
                if response.status_code == 200 or response.status_code == 201:
                    return response.json().get("d", {})
                raise Exception(
                    f"Response is not success with status code {response.status_code} and unable to create file")
        except Exception as exception:
            self.log.exception("Exception while uploading a file to SharePoint Site: %s", str(exception))
            raise exception

    def create_sp_folder(self, request_body):
        """Attempts to create a folder in SharePoint Site

            Args:

               request_body(dict)       --  body of the request
                                                 request_body : {
                                                    "ServerRelativeUrl": folder_url
                                                 }

            Returns:

                response(dict)          --  metadata of folder created

            Raises:

                Exception:
                    if response is not success

        """
        try:
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['CREATE_FOLDER'].format(self.site_url)
            response = self.make_post_request(full_url, request_body)
            if response.status_code == 200 or response.status_code == 201:
                return response.json().get("d", {})
            raise Exception(
                f"Response is not success with status code {response.status_code} and unable to create folder")
        except Exception as exception:
            self.log.exception("Exception while creating folder in SharePoint Site: %s", str(exception))
            raise exception

    def rename_sp_file(self, folder_path, file_name, updated_file_name):
        """Renames file in SharePoint Site

            Args:

                folder_path (str)           --  path of folder in which the file is present

                file_name (str)             --  name of the file

                updated_file_name (str)     --  updated name of the file

            Raises:

                Exception:
                    if response is not success

        """
        try:
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['RENAME_FILE'].format(
                self.site_url, folder_path, file_name, updated_file_name)
            response = self.make_post_request(full_url)
            if not (response.status_code == 200 or response.status_code == 201):
                raise Exception(
                    f"Response is not success with status code {response.status_code} and unable to update file properties")
        except Exception as exception:
            self.log.exception("Exception while updating file metadata in SharePoint Site: %s", str(exception))
            raise exception

    def create_sp_list(self, request_body):
        """Attempts to create a list in SharePoint site

             Args:

                request_body (dict)       --  body of the request
                                                 request_body : {
                                                    "AllowContentTypes": True,
                                                    "BaseTemplate": 100,
                                                    "ContentTypesEnabled": True,
                                                    "Description": "This is a list created for test automation",
                                                    "Title": list_title
                                                 }

            Returns:

                response (dict)          --  metadata of list created

            Raises:

                Exception:
                    if response is not success

        """
        try:
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['CREATE_LIST'].format(self.site_url)
            response = self.make_post_request(full_url, request_body)
            if response.status_code == 200 or response.status_code == 201:
                return response.json().get("d", {})
            raise Exception(
                f"Response is not success with status code {response.status_code} and unable to create list")
        except Exception as exception:
            self.log.exception("Exception while creating a list in SharePoint Site: %s", str(exception))
            raise exception

    def create_list_custom_column(self, list_guid, request_body):
        """Attempts to create a column in the list

             Args:

                list_guid (str)           --  guid of the list

                request_body (dict)       --  body of the request
                                                 request_body : {
                                                    "AllowContentTypes": True,
                                                    "BaseTemplate": 100,
                                                    "ContentTypesEnabled": True,
                                                    "Description": "This is a list created for test automation",
                                                    "Title": list_title
                                                 }

            Returns:

                response (dict)          --  metadata of list created

            Raises:

                Exception:
                    if response is not success

        """
        try:
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['LIST_FIELDS'].format(self.site_url,
                                                                                                    list_guid)
            headers = {
                'Accept': 'application/json; odata=verbose',
                'Content-Type': 'application/json; odata=verbose'
            }
            response = self.make_post_request(full_url, request_body, headers)
            if response.status_code == 200 or response.status_code == 201:
                return response.json().get("d", {})
            raise Exception(
                f"Response is not success with status code {response.status_code} and unable to create column in list")
        except Exception as exception:
            self.log.exception("Exception while creating a column for the list: %s", str(exception))
            raise exception

    def create_sp_list_item(self, list_title, request_body):
        """Creates a list item in the specified list

            Args:

                list_title (str)          --  title of the list

                request_body (dict)       --  body of the request
                                                 request_body : {
                                                        "Title": item_title
                                                    }

            Returns:

                response (dict)          --  metadata of list item created

            Raises:

                Exception:
                    if response is not success

        """
        try:
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['LIST_ITEMS'].format(self.site_url,
                                                                                                   list_title)
            response = self.make_post_request(full_url, request_body)
            if response.status_code == 200 or response.status_code == 201:
                return response.json().get("d", {})
            raise Exception(
                f"Response is not success with status code {response.status_code} and unable to create list item")
        except Exception as exception:
            self.log.exception("Exception while creating a list item in SharePoint Site: %s", str(exception))
            raise exception

    def update_sp_list_item(self, list_title, request_body, list_item_id=None):
        """Updates list item of the specified list in SharePoint Site

            Args:

                list_title (str)          --  title of the list

                request_body (dict)       --  body of the request
                                                 request_body : {
                                                        "Title": item_title
                                                    }

                list_item_id (str)        --  id of list item

            Raises:

                Exception:
                    if response is not success
        """
        try:
            if not list_item_id:
                list_item_id = "1"
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['LIST_ITEM'].format(
                self.site_url, list_title, list_item_id)
            response = self.make_patch_request(full_url, request_body)
            if not (response.status_code == 200 or response.status_code == 204):
                raise Exception(f"Response is not success with status code {response.status_code} "
                                f"and unable to update list_item properties")
        except Exception as exception:
            self.log.exception("Exception while updating a list in SharePoint Site: %s", str(exception))
            raise exception

    def get_sp_list_metadata(self, list_title):
        """Returns metadata of  list in SharePoint Site

              Args:

                  list_title (str)          --  title of the list

              Raises:

                  Exception:
                      if response is not success

        """
        try:
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['LIST'].format(
                self.site_url, list_title)
            response = self.make_get_request(full_url)
            if response.status_code == 200 or response.status_code == 201:
                return response.json().get("d", {})
            elif response.status_code == 404:
                return False
            raise Exception(f"Response is not success with status code {response.status_code} and "
                            f"unable to fetch list metadata")
        except Exception as exception:
            self.log.exception("Exception while retrieving metadata of a list in SharePoint Site: %s",
                               str(exception))
            raise exception

    def get_sp_file_metadata(self, file_name, folder_path, additional_uri=None):
        """Returns the metadata of specified file

            Args:

                file_name (str)             --  name of the file

                folder_path (str)           --  path of folder in which the file is present

                additional_uri(str)         --  additional uri in API endpoint
                                                Example: Versions


            Returns:

                response (dict)             --  metadata of specified file

            Raises:

                Exception:
                    if response is not success
        """
        try:
            if additional_uri:
                full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['GET_SPECIFIC_FILE_METADATA'].format(
                    self.site_url, folder_path, file_name, additional_uri)
            else:
                full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['GET_FILE_METADATA'].format(
                    self.site_url, folder_path, file_name)
            response = self.make_get_request(full_url)
            if response.status_code == 200 or response.status_code == 201:
                return response.json().get("d", {})
            raise Exception(f"Response is not success with status code {response.status_code} and"
                            f" unable to fetch file metadata")
        except Exception as exception:
            self.log.exception("Exception while retrieving metadata of a file in SharePoint Site: %s", str(exception))
            raise exception

    def get_sp_folder_metadata(self, folder_path):
        """Returns the metadata of specified folder

            Args:

                folder_path (str)           --  path of folder in which the file is present

            Returns:

                response (dict)             --  metadata of specified folder

            Raises:

                Exception:
                    if response is not success
        """
        try:
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['FOLDER'].format(
                self.site_url, folder_path)
            response = self.make_get_request(full_url)
            if response.status_code == 200 or response.status_code == 201:
                return response.json().get("d", {})
            elif response.status_code == 404:
                return False
            raise Exception(f"Response is not success with status code {response.status_code} and "
                            f"unable to fetch folder metadata")
        except Exception as exception:
            self.log.exception("Exception while retrieving metadata of a folder in SharePoint Site: %s",
                               str(exception))
            raise exception

    def get_sp_list_item_metadata(self, list_title, list_item_id=None, additional_uri=None):
        """Returns the metadata of first list item of specified list

            Args:

                list_title(str)             --  title of the list

                list_item_id (str)          --  id of list item

                additional_uri(str)         --  additional uri in API endpoint
                                                Example: Versions

            Returns:

                response (dict)             --  metadata of first list item of specified list

            Raises:

                Exception:
                    if response is not success

        """
        try:
            if not list_item_id:
                list_item_id = "1"
            if additional_uri:
                full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['LIST_ITEM_SPECIFIC_METADATA'].format(
                    self.site_url, list_title, list_item_id, additional_uri)
            else:
                full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['LIST_ITEM'].format(
                    self.site_url, list_title, list_item_id)
            response = self.make_get_request(full_url)
            if response.status_code == 200 or response.status_code == 201:
                return response.json().get("d", {})
            raise Exception(f"Response is not success with status code {response.status_code} and"
                            f" unable to fetch list item metadata")
        except Exception as exception:
            self.log.exception("Exception while retrieving metadata of a list item in SharePoint Site: %s",
                               str(exception))
            raise exception

    def get_sp_list_item_title(self, list_title, list_item_id=None):
        """Returns the title of first list item of specified list in SharePoint Site

            Args:

                list_title(str)             --  title of the list

            Returns:

                title (str)                 --  title of first list item of the list (for default case)
                                                title of specified list item of the list

            Raises:

                Exception:
                    if response is not success

        """
        try:
            if not list_item_id:
                list_item_id = "1"
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['LIST_ITEM_TITLE'].format(
                self.site_url, list_title, list_item_id)
            response = self.make_get_request(full_url)
            if response.status_code == 200 or response.status_code == 201:
                return response.json().get("d", {}).get("Title", "")
            raise Exception(f"Response is not success with status code {response.status_code} and"
                            f" unable to fetch list item title")
        except Exception as exception:
            self.log.exception("Exception while retrieving title of first list item of a list in SharePoint Site: %s",
                               str(exception))
            raise exception

    def delete_sp_file_or_folder(self, path):
        """Deletes the file or folder present at specified path in SharePoint Site

            Args:

                path (str)                  --  path of file/folder to be deleted

            Raises:

                Exception:
                    if response is not success

        """
        try:
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['FOLDER'].format(
                self.site_url, path)
            response = self.make_delete_request(full_url)
            if not response.status_code == 200:
                raise Exception(f"Response is not success with status code {response.status_code} and"
                                f" unable to delete file/folder")
        except Exception as exception:
            self.log.exception("Exception while deleting file/folder in SharePoint Site: %s", str(exception))
            raise exception

    def delete_sp_list(self, list_title):
        """Deletes the specified list from SharePoint Site

            Args:

                list_title (str)            --  title of the list to be deleted

            Raises:

                Exception:
                    if response is not success

        """
        try:
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['LIST'].format(
                self.site_url, list_title)
            response = self.make_delete_request(full_url)
            if not response.status_code == 200:
                raise Exception(f"Response is not success with status code {response.status_code} and"
                                f" unable to delete list")
        except Exception as exception:
            self.log.exception("Exception while deleting list in SharePoint Site: %s", str(exception))
            raise exception

    def delete_sp_list_item(self, list_title, list_item_id):
        """Deletes the specified list item from SharePoint list

            Args:

                list_title (str)            --  title of the list to be deleted

                list_item_id(str)           --  id of the list item

            Raises:

                Exception:
                    if response is not success

        """
        try:
            full_url = sharepointconstants.SHAREPOINT_REST_APIS_DICT_TEMPLATE['LIST_ITEM'].format(
                self.site_url, list_title, list_item_id)
            response = self.make_delete_request(full_url)
            if not response.status_code == 200:
                raise Exception(f"Response is not success with status code {response.status_code} and"
                                f"unable to delete list item")
        except Exception as exception:
            self.log.exception("Exception while deleting list item from the list in SharePoint Site: %s",
                               str(exception))
            raise exception
